@@ -1,4 +1,4 @@
-/* $Id: display.c,v 1.11 2000/06/06 05:39:13 aito Exp $ */
+/* $Id: display.c,v 1.1.1.1.4.8 2001/01/22 12:46:55 okabe Exp $ */
 #include <signal.h>
 #include "fm.h"
 
@@ -15,6 +15,8 @@
 #define EFFECT_FORM_END           effect_form_end()
 #define EFFECT_ACTIVE_START	  effect_active_start()
 #define EFFECT_ACTIVE_END	  effect_active_end()
+#define EFFECT_VISITED_START      effect_visited_start()
+#define EFFECT_VISITED_END        effect_visited_end()
 
 /* color: *     0  black *      1  red *        2  green *      3  yellow
  * *    4  blue *       5  magenta *    6  cyan *       7  white */
@@ -23,11 +25,13 @@
 #define EFFECT_IMAGE_START_C        setfcolor(image_color)
 #define EFFECT_FORM_START_C         setfcolor(form_color)
 #define EFFECT_ACTIVE_START_C      (setfcolor(active_color), underline())
+#define EFFECT_VISITED_START_C      setfcolor(visited_color)
 
 #define EFFECT_IMAGE_END_C          setfcolor(basic_color)
 #define EFFECT_ANCHOR_END_C         setfcolor(basic_color)
 #define EFFECT_FORM_END_C           setfcolor(basic_color)
 #define EFFECT_ACTIVE_END_C        (setfcolor(basic_color), underlineend())
+#define EFFECT_VISITED_END_C        setfcolor(basic_color)
 
 #define EFFECT_ANCHOR_START_M       underline()
 #define EFFECT_ANCHOR_END_M         underlineend()
@@ -39,6 +43,8 @@
 #define EFFECT_ACTIVE_END_NC        underlineend()
 #define EFFECT_ACTIVE_START_M       bold()
 #define EFFECT_ACTIVE_END_M         boldend()
+#define EFFECT_VISITED_START_M      /**/
+#define EFFECT_VISITED_END_M        /**/
 
 #define define_effect(name_start,name_end,color_start,color_end,mono_start,mono_end) \
 static void name_start { if (useColor) { color_start; } else { mono_start; }}\
@@ -47,33 +53,60 @@ static void name_end { if (useColor) { color_end; } else { mono_end; }}
 define_effect(EFFECT_ANCHOR_START, EFFECT_ANCHOR_END, EFFECT_ANCHOR_START_C, EFFECT_ANCHOR_END_C, EFFECT_ANCHOR_START_M, EFFECT_ANCHOR_END_M)
 define_effect(EFFECT_IMAGE_START, EFFECT_IMAGE_END, EFFECT_IMAGE_START_C, EFFECT_IMAGE_END_C, EFFECT_IMAGE_START_M, EFFECT_IMAGE_END_M)
 define_effect(EFFECT_FORM_START, EFFECT_FORM_END, EFFECT_FORM_START_C, EFFECT_FORM_END_C, EFFECT_FORM_START_M, EFFECT_FORM_END_M)
-    static void EFFECT_ACTIVE_START
-    {
-	if (useColor)
+static void EFFECT_ACTIVE_START
 {
-if (useActiveColor)
-    EFFECT_ACTIVE_START_C;
-else {
-    EFFECT_ACTIVE_START_NC;
-}
-}
-	else
-	     EFFECT_ACTIVE_START_M;
+    if (useColor) {
+	if (useActiveColor) {
+#ifdef __EMX__
+	    if(!getenv("WINDOWID"))
+		setfcolor(active_color);
+	    else
+#endif
+	    {
+		EFFECT_ACTIVE_START_C;
+	    }
+	} else {
+	    EFFECT_ACTIVE_START_NC;
+	}
+    } else {
+	EFFECT_ACTIVE_START_M;
     }
+}
 
 static void EFFECT_ACTIVE_END
 {
     if (useColor) {
-	if (useActiveColor)
+	if (useActiveColor) {
 	    EFFECT_ACTIVE_END_C;
-	else {
+	} else {
 	    EFFECT_ACTIVE_END_NC;
 	}
-    }
-    else
+    } else {
 	 EFFECT_ACTIVE_END_M;
+    }
 }
 
+static void EFFECT_VISITED_START
+{
+    if (useVisitedColor) {
+	if (useColor) {
+	    EFFECT_VISITED_START_C;
+	} else {
+	    EFFECT_VISITED_START_M;
+	}
+    }
+}
+
+static void EFFECT_VISITED_END
+{
+    if (useVisitedColor) {
+	if (useColor) {
+	    EFFECT_VISITED_END_C;
+	} else {
+	    EFFECT_VISITED_END_M;
+	}
+    }
+}
 
 #else				/* not COLOR */
 
@@ -85,6 +118,8 @@ static void EFFECT_ACTIVE_END
 #define EFFECT_FORM_END           standend()
 #define EFFECT_ACTIVE_START       bold()
 #define EFFECT_ACTIVE_END         boldend()
+#define EFFECT_VISITED_START      /**/
+#define EFFECT_VISITED_END        /**/
 
 #endif				/* not COLOR */
 
@@ -154,10 +189,13 @@ static int ccolumn = -1;
 
 static int ulmode = 0, somode = 0, bomode = 0;
 static int anch_mode = 0, emph_mode = 0, imag_mode = 0, form_mode = 0,
- active_mode = 0;
+ active_mode = 0, visited_mode = 0;
 #ifndef KANJI_SYMBOLS
 static int graph_mode = 0;
 #endif				/* not KANJI_SYMBOLS */
+#ifdef ANSI_COLOR
+static Linecolor color_mode = 0;
+#endif
 
 #ifdef BUFINFO
 static Buffer *save_current_buf = NULL;
@@ -168,27 +206,12 @@ static int in_check_url = FALSE;
 void
 displayBuffer(Buffer * buf, int mode)
 {
-    Str msg = Strnew();
-    Anchor *aa;
-    AnchorList *formitem;
+    Str msg;
+    Anchor *aa = NULL;
 
     if (in_check_url)
 	return;
-    if (buf->sourcefile != NULL && buf->topLine == NULL) {	/* clear_buffer 
-								 */
-	formitem = buf->formitem;
-	reloadBuffer(buf);
-	in_check_url = TRUE;
-	if (buf->check_url & CHK_URL)
-	    chkURL();
-#ifdef USE_NNTP
-	if (buf->check_url & CHK_NMID)
-	    chkNMID();
-#endif
-	in_check_url = FALSE;
-	gotoLine(buf, buf->currentLineNum);
-	arrangeCursor(buf);
-	formResetBuffer(buf, formitem);
+    if (buf->topLine == NULL && readBufferCache(buf) == 0) {	/* clear_buffer */
 	mode = B_FORCE_REDRAW;
     }
 
@@ -223,42 +246,33 @@ displayBuffer(Buffer * buf, int mode)
 	cline = buf->topLine;
 	ccolumn = buf->currentColumn;
     }
-    if (buf->firstLine == NULL) {
-	clear();
-	standout();
-	msg = Sprintf("Viewing <%s>\tNo Line", cleanup_str(buf->buffername));
-	message(msg->ptr, buf->cursorX, buf->cursorY);
-	standend();
-	refresh();
-	return;
-    }
-    move(buf->cursorY, buf->cursorX);
     if (buf->topLine == NULL)
 	buf->topLine = buf->firstLine;
-    standout();
+
 #ifdef MOUSE
-    if (!use_mouse)
-	msg = Sprintf("Viewing <%s> ", cleanup_str(buf->buffername));
-    else
+    if (use_mouse)
 #if LANG == JA
-	msg = Sprintf("㢬Viewing <%s> ", cleanup_str(buf->buffername));
+	msg = Strnew_charp("㢬");
 #else				/* LANG != JA */
-	msg = Sprintf("<=UpDn Viewing <%s> ", cleanup_str(buf->buffername));
+	msg = Strnew_charp("<=UpDn ");
 #endif				/* LANG != JA */
-#else				/* not MOUSE */
-    msg = Sprintf("Viewing <%s> ", cleanup_str(buf->buffername));
+    else
 #endif				/* not MOUSE */
-    if (displayLink) {
+	msg = Strnew();
+    Strcat_charp(msg, "Viewing <");
+    Strcat_charp(msg, buf->buffername);
+    if (displayLink)
 	aa = retrieveCurrentAnchor(buf);
-	if (aa) {
-	    ParsedURL url;
-	    Str s;
-	    int l;
-	    parseURL2(aa->url, &url, baseURL(buf));
-	    s = parsedURL2Str(&url);
-	    l = buf->width - msg->length;
-	    if (s->length > l && l >= 4) {
-		Strcat(msg, Strsubstr(s, 0, (l - 2) / 2));
+    if (aa) {
+	ParsedURL url;
+	Str s;
+	int l;
+	parseURL2(aa->url, &url, baseURL(buf));
+	s = parsedURL2Str(&url);
+	l = buf->width - 2;
+	if (s->length > l) {
+	    if (l >= 4) {
+		msg = Strsubstr(s, 0, (l - 2) / 2);
 #if LANG == JA
 		Strcat_charp(msg, "");
 #else				/* LANG != JA */
@@ -266,12 +280,35 @@ displayBuffer(Buffer * buf, int mode)
 #endif				/* LANG != JA */
 		l = buf->width - msg->length;
 		Strcat(msg, Strsubstr(s, s->length - l, l));
+	    } else {
+		msg = s;
 	    }
-	    else {
-		Strcat(msg, s);
+	} else {
+	    l -= s->length;
+	    if (msg->length > l) {
+#ifdef JP_CHARSET
+		char *bn = msg->ptr;
+		int i, j;
+		for (i = 0; bn[i]; i += j) {
+		    j = get_mclen(get_mctype(&bn[i]));
+		    if (i + j > l)
+			break;
+		}
+		l = i;
+#endif
+		Strtruncate(msg, l);
 	    }
+	    Strcat_charp(msg, "> ");
+	    Strcat(msg, s);
 	}
+    } else {
+	Strcat_charp(msg, ">");
     }
+    if (buf->firstLine == NULL) {
+	Strcat_charp(msg, "\tNo Line");
+	clear();
+    }
+    standout();
     message(msg->ptr, buf->cursorX, buf->cursorY);
     standend();
     refresh();
@@ -317,13 +354,23 @@ redrawNLine(Buffer * buf, int n)
 	clrtobotx();
 }
 
+#define addKanji(pc,pr) (addChar((pc)[0],(pr)[0]),addChar((pc)[1],(pr)[1]))
+
 Line *
 redrawLine(Buffer * buf, Line * l, int i)
 {
-    int j, pos, rcol;
+    int j, pos, rcol, ncol, delta;
     int column = buf->currentColumn;
     char *p;
     Lineprop *pr;
+#ifdef ANSI_COLOR
+    Linecolor *pc;
+#endif
+#ifdef COLOR
+    Anchor *a;
+    ParsedURL url;
+    int k, vpos = -1;
+#endif
 
     if (l == NULL) {
 	if (buf->pagerSource) {
@@ -335,7 +382,9 @@ redrawLine(Buffer * buf, Line * l, int i)
 	    return NULL;
     }
     move(i, 0);
-    if (l->len == 0 || COLPOS(l, l->len - 1) < column) {
+    if (l->width < 0)
+	l->width = COLPOS(l, l->len);
+    if (l->len == 0 || l->width - 1 < column) {
 	clrtoeolx();
 	return l->next;
     }
@@ -343,29 +392,62 @@ redrawLine(Buffer * buf, Line * l, int i)
     pos = columnPos(l, column);
     p = &(l->lineBuf[pos]);
     pr = &(l->propBuf[pos]);
+#ifdef ANSI_COLOR
+    if (useColor && l->colorBuf)
+	pc = &(l->colorBuf[pos]);
+    else
+	pc = NULL;
+#endif
     rcol = COLPOS(l, pos);
 
-    for (j = 0; rcol - column < COLS && pos + j < l->len; j++) {
+#ifndef JP_CHARSET
+    delta = 1;
+#endif
+    for (j = 0; rcol - column < COLS && pos + j < l->len; j += delta) {
+#ifdef COLOR
+	if (useVisitedColor && vpos <= pos + j && !(pr[j] & PE_VISITED)) {
+	    a = retrieveAnchor(buf->href, l->linenumber, pos + j);
+	    if (a) {
+		parseURL2(a->url, &url, baseURL(buf));
+		if (getHashHist(URLHist, parsedURL2Str(&url)->ptr)) {
+		    for (k = a->start.pos; k < a->end.pos; k++)
+			pr[k - pos] |= PE_VISITED;
+		}
+		vpos = a->end.pos;
+	    }
+	}
+#endif
 #ifdef JP_CHARSET
-	if (rcol - column == COLS - 1 && CharType(pr[j]) == PC_KANJI1)
+	if (CharType(pr[j]) == PC_KANJI1)
+	    delta = 2;
+	else
+	    delta = 1;
+#endif
+	ncol = COLPOS(l, pos + j + delta);
+	if (ncol - column > COLS)
 	    break;
-#endif				/* JP_CHARSET */
-	if (p[j] == '\t') {
-	    do {
-		if (rcol >= column)
-		    addChar(' ', 0);
-	    } while ((++rcol % Tabstop) != 0);
-	}
-	else {
-#ifdef JP_CHARSET
-	    if (rcol == column && CharType(pr[j]) == PC_KANJI2)
+#ifdef ANSI_COLOR
+	if (pc)
+	    do_color(pc[j]);
+#endif
+	if (rcol < column) {
+	    for (rcol = column; rcol < ncol; rcol++)
 		addChar(' ', 0);
-	    else
-#endif				/* JP_CHARSET */
-	    if (rcol >= column)
-		addChar(p[j], pr[j]);
-	    rcol++;
+	    continue;
 	}
+	if (p[j] == '\t') {
+	    for (; rcol < ncol; rcol++)
+		addChar(' ', 0);
+	}
+#ifdef JP_CHARSET
+	else if (delta == 2) {
+	    addKanji(&p[j], &pr[j]);
+	}
+#endif
+	else {
+	    addChar(p[j], pr[j]);
+	}
+	rcol = ncol;
     }
     if (somode) {
 	somode = FALSE;
@@ -396,7 +478,10 @@ redrawLine(Buffer * buf, Line * l, int i)
 	form_mode = FALSE;
 	EFFECT_FORM_END;
     }
-
+    if (visited_mode) {
+	visited_mode = FALSE;
+	EFFECT_VISITED_END;
+    }
     if (active_mode) {
 	active_mode = FALSE;
 	EFFECT_ACTIVE_END;
@@ -407,6 +492,10 @@ redrawLine(Buffer * buf, Line * l, int i)
 	graphend();
     }
 #endif				/* not KANJI_SYMBOLS */
+#ifdef ANSI_COLOR
+    if (color_mode)
+	do_color(0);
+#endif
     if (rcol - column < COLS)
 	clrtoeolx();
     return l->next;
@@ -415,49 +504,87 @@ redrawLine(Buffer * buf, Line * l, int i)
 int
 redrawLineRegion(Buffer * buf, Line * l, int i, int bpos, int epos)
 {
-    int j, pos, rcol;
+    int j, pos, rcol, ncol, delta;
     int column = buf->currentColumn;
     char *p;
     Lineprop *pr;
+#ifdef ANSI_COLOR
+    Linecolor *pc;
+#endif
     int bcol, ecol;
+#ifdef COLOR
+    Anchor *a;
+    ParsedURL url;
+    int k, vpos = -1;
+#endif
 
     if (l == NULL)
 	return 0;
     pos = columnPos(l, column);
     p = &(l->lineBuf[pos]);
     pr = &(l->propBuf[pos]);
+#ifdef ANSI_COLOR
+    if (useColor && l->colorBuf)
+	pc = &(l->colorBuf[pos]);
+    else
+	pc = NULL;
+#endif
     rcol = COLPOS(l, pos);
     bcol = bpos - pos;
     ecol = epos - pos;
 
-    for (j = 0; rcol - column < COLS && pos + j < l->len; j++) {
-#ifdef JP_CHARSET
-	if (rcol - column == COLS - 1 && CharType(pr[j]) == PC_KANJI1)
-	    break;
-	if (rcol == column && CharType(pr[j]) == PC_KANJI2) {
-	    if (j >= bcol && j < ecol) {
-		move(i, 0);
-		addChar(' ', 0);
-	    }
-	    rcol++;
-	}
-	else
-#endif				/* JP_CHARSET */
-	if (p[j] == '\t') {
-	    do {
-		if (rcol >= column && j >= bcol && j < ecol) {
-		    move(i, rcol - column);
-		    addChar(' ', 0);
+#ifndef JP_CHARSET
+    delta = 1;
+#endif
+    for (j = 0; rcol - column < COLS && pos + j < l->len; j += delta) {
+#ifdef COLOR
+	if (useVisitedColor && vpos <= pos + j && !(pr[j] & PE_VISITED)) {
+	    a = retrieveAnchor(buf->href, l->linenumber, pos + j);
+	    if (a) {
+		parseURL2(a->url, &url, baseURL(buf));
+		if (getHashHist(URLHist, parsedURL2Str(&url)->ptr)) {
+		    for (k = a->start.pos; k < a->end.pos; k++)
+			pr[k - pos] |= PE_VISITED;
 		}
-	    } while ((++rcol % Tabstop) != 0);
+		vpos = a->end.pos;
+	    }
 	}
-	else {
-	    if (rcol >= column && j >= bcol && j < ecol) {
-		move(i, rcol - column);
+#endif
+#ifdef JP_CHARSET
+	if (CharType(pr[j]) == PC_KANJI1)
+	    delta = 2;
+	else
+	    delta = 1;
+#endif
+	ncol = COLPOS(l, pos + j + delta);
+	if (ncol - column > COLS)
+	    break;
+#ifdef ANSI_COLOR
+	if (pc)
+	    do_color(pc[j]);
+#endif
+	if (j >= bcol && j < ecol) {
+	    if (rcol < column) {
+		move(i, 0);
+		for (rcol = column; rcol < ncol; rcol++)
+		    addChar(' ', 0);
+		continue;
+	    }
+	    move(i, rcol - column);
+	    if (p[j] == '\t') {
+		for (; rcol < ncol; rcol++)
+		    addChar(' ', 0);
+	    }
+#ifdef JP_CHARSET
+	    else if (delta == 2) {
+		addKanji(&p[j], &pr[j]);
+	    }
+#endif
+	    else {
 		addChar(p[j], pr[j]);
 	    }
-	    rcol++;
 	}
+	rcol = ncol;
     }
     if (somode) {
 	somode = FALSE;
@@ -488,6 +615,10 @@ redrawLineRegion(Buffer * buf, Line * l, int i, int bpos, int epos)
 	form_mode = FALSE;
 	EFFECT_FORM_END;
     }
+    if (visited_mode) {
+	visited_mode = FALSE;
+	EFFECT_VISITED_END;
+    }
     if (active_mode) {
 	active_mode = FALSE;
 	EFFECT_ACTIVE_END;
@@ -498,34 +629,39 @@ redrawLineRegion(Buffer * buf, Line * l, int i, int bpos, int epos)
 	graphend();
     }
 #endif				/* not KANJI_SYMBOLS */
+#ifdef ANSI_COLOR
+    if (color_mode)
+	do_color(0);
+#endif
     return rcol - column;
 }
 
 #define do_effect1(effect,modeflag,action_start,action_end) \
 if (m & effect) { \
-	if (!modeflag) { \
-		action_start; \
-		modeflag = TRUE; \
-	} \
+    if (!modeflag) { \
+	action_start; \
+	modeflag = TRUE; \
+    } \
 }
 
 #define do_effect2(effect,modeflag,action_start,action_end) \
 if (modeflag) { \
-	action_end; \
-	modeflag = FALSE; \
+    action_end; \
+    modeflag = FALSE; \
 }
 
 void
 do_effects(Lineprop m)
 {
     /* effect end */
-    do_effect2(PE_UNDER, ulmode, underline(), underlineend())
-	do_effect2(PE_STAND, somode, standout(), standend())
-	do_effect2(PE_BOLD, bomode, bold(), boldend())
-	do_effect2(PE_EMPH, emph_mode, bold(), boldend())
-	do_effect2(PE_ANCHOR, anch_mode, EFFECT_ANCHOR_START, EFFECT_ANCHOR_END);
+    do_effect2(PE_UNDER, ulmode, underline(), underlineend());
+    do_effect2(PE_STAND, somode, standout(), standend());
+    do_effect2(PE_BOLD, bomode, bold(), boldend());
+    do_effect2(PE_EMPH, emph_mode, bold(), boldend());
+    do_effect2(PE_ANCHOR, anch_mode, EFFECT_ANCHOR_START, EFFECT_ANCHOR_END);
     do_effect2(PE_IMAGE, imag_mode, EFFECT_IMAGE_START, EFFECT_IMAGE_END);
     do_effect2(PE_FORM, form_mode, EFFECT_FORM_START, EFFECT_FORM_END);
+    do_effect2(PE_VISITED, visited_mode, EFFECT_VISITED_START, EFFECT_VISITED_END);
     do_effect2(PE_ACTIVE, active_mode, EFFECT_ACTIVE_START, EFFECT_ACTIVE_END);
 #ifndef KANJI_SYMBOLS
     if (graph_mode) {
@@ -535,16 +671,17 @@ do_effects(Lineprop m)
 #endif				/* not KANJI_SYMBOLS */
 
     /* effect start */
-    do_effect1(PE_UNDER, ulmode, underline(), underlineend())
-	do_effect1(PE_STAND, somode, standout(), standend())
-	do_effect1(PE_BOLD, bomode, bold(), boldend())
-	do_effect1(PE_EMPH, emph_mode, bold(), boldend())
-	do_effect1(PE_ANCHOR, anch_mode, EFFECT_ANCHOR_START, EFFECT_ANCHOR_END);
+    do_effect1(PE_UNDER, ulmode, underline(), underlineend());
+    do_effect1(PE_STAND, somode, standout(), standend());
+    do_effect1(PE_BOLD, bomode, bold(), boldend());
+    do_effect1(PE_EMPH, emph_mode, bold(), boldend());
+    do_effect1(PE_ANCHOR, anch_mode, EFFECT_ANCHOR_START, EFFECT_ANCHOR_END);
     do_effect1(PE_IMAGE, imag_mode, EFFECT_IMAGE_START, EFFECT_IMAGE_END);
     do_effect1(PE_FORM, form_mode, EFFECT_FORM_START, EFFECT_FORM_END);
+    do_effect1(PE_VISITED, visited_mode, EFFECT_VISITED_START, EFFECT_VISITED_END);
     do_effect1(PE_ACTIVE, active_mode, EFFECT_ACTIVE_START, EFFECT_ACTIVE_END);
 #ifndef KANJI_SYMBOLS
-    if (m & PC_ISRULE) {
+    if (m & PC_RULE) {
 	if (!graph_mode && graph_ok()) {
 	    graphstart();
 	    graph_mode = TRUE;
@@ -553,6 +690,24 @@ do_effects(Lineprop m)
 #endif				/* not KANJI_SYMBOLS */
 }
 
+#ifdef ANSI_COLOR
+void
+do_color(Linecolor c)
+{
+    if (c & 0x8)
+	setfcolor(c & 0x7);
+    else if (color_mode & 0x8)
+	setfcolor(basic_color);
+#ifdef BG_COLOR
+    if (c & 0x80)
+	setbcolor((c >> 4) & 0x7);
+    else if (color_mode & 0x80)
+	setbcolor(bg_color);
+#endif
+    color_mode = c;
+}
+#endif
+
 void
 addChar(char c, Lineprop mode)
 {
@@ -560,49 +715,69 @@ addChar(char c, Lineprop mode)
 
 #ifdef JP_CHARSET
     if (CharType(mode) != PC_KANJI2)
-	do_effects(m);
-    if (c & 0x80 && CharType(mode) == PC_ASCII) {
-	addch(' ');
-    }
-    else
-#else				/* not JP_CHARSET */
+#endif				/* JP_CHARSET */
     do_effects(m);
-#endif				/* not JP_CHARSET */
 #ifndef KANJI_SYMBOLS
-    if (graph_mode)
-	addch(g_rule[PC_RULE_VAL(m)]);
-    else
+    if (m & PC_RULE) {
+	if (graph_mode)
+	    addch(g_rule[c & 0xF]);
+	else
+	    addch(alt_rule[c & 0xF]);
+    } else
 #endif				/* not KANJI_SYMBOLS */
-    if (c == '\t') {
-	addch(c);
+    if (IS_UNPRINTABLE_ASCII(c, mode)) {
+	addstr(Sprintf("\\%3o", (unsigned char)c)->ptr);
     }
-    else if (!(c & ~0x1f) && c != CTRL_J) {	/* Control code */
-	addch('^');
-	addch(c + '@');
+    else if (c == '\t') {
+	addch(c);
     }
     else if (c == DEL_CODE)
 	addstr("^?");
-#if LANG != EN
+    else if (IS_UNPRINTABLE_CONTROL(c, mode)) {	/* Control code */
+	addch('^');
+	addch(c + '@');
+    }
     else if (c != '\n')
 	addch(c);
-#else
-    else if ((unsigned char) c == 0xa0)		/* NO-BREAK SPACE */
-	addch(' ');
-    else if (c != '\n') {
-	if ((unsigned char) c < 0x80 || (unsigned char) c > 0x9F) {
-	    addch(c);
-	}
-	else {
-	    /* else: Eliminate unprintables according to * iso-8859-*.
-	     * Particularly 0x96 messes up T.Dickey's * (xfree-)xterm */
-	    addch(' ');
-	}
-    }
-#endif
     else	/* \n */
 	addch(' ');
 }
 
+GeneralList *message_list = NULL;
+
+void
+record_err_message(char *s)
+{
+    if (fmInitialized) {
+	if (!message_list)
+	    message_list = newGeneralList();
+	if (message_list->nitem >= LINES)
+	    popValue(message_list);
+	pushValue(message_list, allocStr(s, 0));
+    }
+}
+
+/* 
+ * List of error messages
+ */
+Buffer *
+message_list_panel(void)
+{
+    Str tmp = Strnew_size(LINES * COLS);
+    ListItem *p;
+
+    Strcat_charp(tmp,
+		 "<html><head><title>List of error messages</title></head><body>"
+		 "<h1>List of error messages</h1><table cellpadding=0>\n");
+    if (message_list)
+      for (p = message_list->last ; p ; p = p->prev)
+	Strcat_m_charp(tmp, "<tr><td><pre>", htmlquote_str(p->ptr), "</pre></td></tr>\n", NULL);
+    else
+      Strcat_charp(tmp, "<tr><td>(no message recorded)</td></tr>\n");
+    Strcat_charp(tmp, "</table></body></html>");
+    return loadHTMLString(tmp);
+}
+
 void
 message(char *s, int return_x, int return_y)
 {
@@ -689,7 +864,7 @@ cursorUpDown(Buffer * buf, int n)
 
     if (buf->firstLine == NULL)
 	return;
-    if ((buf->currentLine = lineSkip(buf, cl, n, FALSE)) == cl)
+    if ((buf->currentLine = currentLineSkip(buf, cl, n, FALSE)) == cl)
 	return;
     arrangeLine(buf);
 }
@@ -703,6 +878,8 @@ cursorRight(Buffer * buf)
 
     if (buf->firstLine == NULL)
 	return;
+    if (buf->pos == l->len)
+	return;
     i = buf->pos;
     p = l->propBuf;
 #ifdef JP_CHARSET
@@ -710,15 +887,17 @@ cursorRight(Buffer * buf)
 	delta = 2;
 #endif				/* JP_CHARSET */
     if (i + delta < l->len) {
-      buf->pos = i + delta;
+	buf->pos = i + delta;
     }
     else if (l->len == 0) {
         buf->pos = 0;
     }
     else {
         buf->pos = l->len -1;
+#ifdef JP_CHARSET
         if (CharType(p[buf->pos]) == PC_KANJI2)
             buf->pos--;
+#endif                          /* JP_CHARSET */
     }
     cpos = COLPOS(l, buf->pos);
     buf->visualpos = cpos - buf->currentColumn;
@@ -832,8 +1011,16 @@ arrangeLine(Buffer * buf)
 	buf->pos = i;
     }
     else if (Currentbuf->currentLine->len > i) {
+	int delta = 1;
+#ifdef JP_CHARSET
+	if (Currentbuf->currentLine->len > i + 1 &&
+	    CharType(buf->currentLine->propBuf[i + 1]) == PC_KANJI2)
+	    delta = 2;
+#endif
 	buf->cursorX = 0;
-	buf->pos = i + 1;
+	buf->pos = i;
+	if (COLPOS(buf->currentLine, i + delta) <= buf->currentColumn)
+	    buf->pos += delta;
     }
     else {
 	buf->cursorX = 0;
@@ -870,3 +1057,8 @@ cursorXY(Buffer * buf, int x, int y)
 	    cursorLeft(buf);
     }
 }
+
+/* Local Variables:    */
+/* c-basic-offset: 4   */
+/* tab-width: 8        */
+/* End:                */
