@@ -1,25 +1,20 @@
-
-/* $Id: etc.c,v 1.21 2000/06/06 05:39:13 aito Exp $ */
+/* $Id: etc.c,v 1.8.2.31 2001/01/22 12:46:55 okabe Exp $ */
 #include "fm.h"
 #include <pwd.h>
 #include "myctype.h"
 #include "html.h"
 #include "local.h"
 #include "hash.h"
+#include "terms.h"
 
 #ifdef GETCWD
 #include <unistd.h>
 #include <sys/param.h>
 #endif				/* GETCWD */
 
-#ifdef __EMX__
-#include <process.h>
-#endif				/* __EMX__ */
-
 #include <sys/types.h>
 #include <time.h>
 #include <sys/wait.h>
-#include <signal.h>
 
 #ifdef	__WATT32__
 #define	read(a,b,c)	read_s(a,b,c)
@@ -75,25 +70,10 @@ arg_is(char *str, char *tag)
     return (*str == '=');
 }
 
-
-int
-bpcmp(BufferPoint p1, BufferPoint p2)
-{
-    if (p1.line < p2.line)
-	return -1;
-    if (p1.line > p2.line)
-	return 1;
-    if (p1.pos < p2.pos)
-	return -1;
-    if (p1.pos > p2.pos)
-	return 1;
-    return 0;
-}
-
 int
 columnSkip(Buffer * buf, int offset)
 {
-    int i, len, maxColumn;
+    int i, maxColumn;
     int column = buf->currentColumn + offset;
     int nlines = LASTLINE + 1;
     Line *l;
@@ -102,9 +82,10 @@ columnSkip(Buffer * buf, int offset)
     for (i = 0, l = buf->topLine;
 	 i < nlines && l != NULL;
 	 i++, l = l->next) {
-	len = COLPOS(l, l->len - 1);
-	if (len > maxColumn)
-	    maxColumn = len;
+	if (l->width < 0)
+	    l->width = COLPOS(l, l->len);
+	if (l->width - 1 > maxColumn)
+	    maxColumn = l->width - 1;
     }
     maxColumn -= COLS - 1;
     if (column < maxColumn)
@@ -138,17 +119,30 @@ columnPos(Line * line, int column)
 Line *
 lineSkip(Buffer * buf, Line * line, int offset, int last)
 {
-    int i, n;
-    Line *l = line, *line_p;
-    int linelen = 0;
+    int i;
+    Line *l;
 
-    if (buf->pagerSource && strcmp(buf->buffername, CPIPEBUFFERNAME)) {
+    l = currentLineSkip(buf, line, offset, last);
+    for (i = (LASTLINE - 1) - (buf->lastLine->linenumber - l->linenumber);
+	i > 0 && l->prev != NULL;
+	i--, l = l->prev);
+    return l;
+}
+
+Line *
+currentLineSkip(Buffer * buf, Line * line, int offset, int last)
+{
+    int i, n;
+    Line *l = line;
+
+    if (buf->pagerSource && !(buf->bufferprop & BP_CLOSE)) {
 	n = line->linenumber + offset + LASTLINE;
+	if (buf->lastLine->linenumber < n)
+	    getNextPage(buf, n - buf->lastLine->linenumber);
 	while ((last || (buf->lastLine->linenumber < n)) &&
-	       ((line_p = getNextPage(buf, 1)) != NULL)) {
-	    linelen += line_p->len;
-	    showProgress(&linelen, &(buf->trbyte));
-	}
+	       (getNextPage(buf, 1) != NULL));
+	if (last)
+	    l = buf->lastLine;
     }
 
     if (offset == 0)
@@ -178,14 +172,22 @@ get_cmd(Hash_si * hash,
     if (status)
 	*status = 0;
     (*s)++;
-    while ((IS_ALNUM(**s) || **s == '_' || **s == '/') &&
+    /* first character */
+    if (IS_ALNUM(**s) || **s == '_' || **s == '/')
+	*(p++) = tolower(*((*s)++));
+    else
+	return defaultcmd;
+    if (p[-1] == '/')
+	SKIP_BLANKS(*s);
+    while ((IS_ALNUM(**s) || **s == '_') &&
 	   p - cmdstr < MAX_CMD_LEN) {
 	*(p++) = tolower(*((*s)++));
     }
     if (p - cmdstr == MAX_CMD_LEN) {
 	/* buffer overflow: perhaps caused by bad HTML source */
 	*s = save + 1;
-	*status = -1;
+	if (status)
+	    *status = -1;
 	return defaultcmd;
     }
     *p = '\0';
@@ -216,7 +218,7 @@ int
 gethtmlcmd(char **s, int *status)
 {
     extern Hash_si tagtable;
-    return get_cmd(&tagtable, s, NULL, '>', HTML_UNKNOWN, 1, status);
+    return get_cmd(&tagtable, s, NULL, '>', HTML_UNKNOWN, TRUE, status);
 }
 
 static char *
@@ -275,183 +277,342 @@ getAnchor(char *arg, char **arg_return)
     return allocStr(arg, p - arg);
 }
 
+#ifdef ANSI_COLOR
+static int
+parse_ansi_color(char **str, Lineprop *effect, Linecolor *color)
+{
+    char *p = *str, *q;
+    Lineprop e = *effect;
+    Linecolor c = *color;
+    int i;
+
+    if (*p != ESC_CODE || *(p+1) != '[')
+	return 0;
+    p += 2;
+    for (q = p; IS_DIGIT(*q) || *q == ';'; q++)
+	;
+    if (*q != 'm')
+	return 0;
+    *str = q + 1;
+    while (1) {
+	if (*p == 'm') {
+	    e = PE_NORMAL;
+	    c = 0;
+	    break;
+	}
+	if (IS_DIGIT(*p)) {
+	    q = p;
+	    for (p++; IS_DIGIT(*p); p++)
+		;
+	    i = atoi(allocStr(q, p - q));
+	    switch (i) {
+	    case 0:
+		e = PE_NORMAL;
+		c = 0;
+		break;
+	    case 1:
+	    case 5:
+		e = PE_BOLD;
+		break;
+	    case 4:
+		e = PE_UNDER;
+		break;
+	    case 7:
+		e = PE_STAND;
+		break;
+	    case 100:	/* for EWS4800 kterm */
+		c = 0;
+		break;
+	    case 39:
+		c &= 0xf0;
+		break;
+	    case 49:
+		c &= 0x0f;
+		break;
+	    default:
+		if (i >= 30 && i <= 37)
+		    c = (c & 0xf0) | (i - 30) | 0x08;
+		else if (i >= 40 && i <= 47)
+		    c = (c & 0x0f) | ((i - 40) << 4) | 0x80;
+		break;
+	    }
+	    if (*p == 'm')
+		break;
+	} else {
+	    e = PE_NORMAL;
+	    c = 0;
+	    break;
+	}
+	p++;	/* *p == ';' */
+    }
+    *effect = e;
+    *color = c;
+    return 1;
+}
+#endif
+
 /* 
  * Check character type
  */
 
 Str
-checkType(char *istr, Lineprop * oprop, int len)
+checkType(Str s, Lineprop * oprop,
+#ifdef ANSI_COLOR
+	  Linecolor * ocolor, int * check_color,
+#endif
+	  int len)
 {
-    Lineprop mode = PC_ASCII, prev_mode;
-    Lineprop effect = PE_NORMAL, leffect = PE_NORMAL;
+    Lineprop mode;
+    Lineprop effect = PE_NORMAL;
     Lineprop *prop = oprop;
-    char *str = istr;
-    Str s = Strnew();
+    char *str = s->ptr, *endp = &s->ptr[s->length], *bs = NULL;
+#ifdef ANSI_COLOR
+    Lineprop ceffect = PE_NORMAL;
+    Linecolor cmode = 0;
+    Linecolor *color = NULL;
+    char *es = NULL;
+#endif
+    int do_copy = FALSE;
+    int size = (len < s->length) ? len : s->length;
 
-    while (*str != '\0') {
+#ifdef ANSI_COLOR
+    if (check_color)
+	*check_color = FALSE;
+#endif
+    if (ShowEffect) {
+	bs = memchr(str, '\b', s->length);
+#ifdef ANSI_COLOR
+	if (ocolor) {
+	    es = memchr(str, ESC_CODE, s->length);
+	    if (es)
+		color = ocolor;
+	}
+#endif
+	if (s->length > size || (bs != NULL)
+#ifdef ANSI_COLOR
+	    || (es != NULL)
+#endif
+	    ) {
+	    s = Strnew_size(size);
+	    do_copy = TRUE;
+	}
+    }
+
+    while (str < endp) {
 	if (prop - oprop >= len)
 	    break;
-	leffect = effect;
-	if (ShowEffect
-#ifdef JP_CHARSET
-	    && mode != PC_KANJI1
-#endif				/* JP_CHARSET */
-	    ) {
-	    if (!strncmp(str, "_\b", 2)) {
-		str += 2;
-		effect = PE_UNDER;
-		continue;
-	    }
-	    else if (!strncmp(str, "__\b\b", 4)) {
+	if (bs != NULL) {
+	    if (str == bs - 2 && !strncmp(str, "__\b\b", 4)) {
 		str += 4;
 		effect = PE_UNDER;
+		if (str < endp)
+		    bs = memchr(str, '\b', endp - str);
 		continue;
 	    }
-	    else if (!strncmp(str, "\b_", 2)) {
+	    else if (str == bs - 1 && *str == '_') {
+		str += 2;
+		effect = PE_UNDER;
+		if (str < endp)
+		    bs = memchr(str, '\b', endp - str);
+		continue;
+	    }
+	    else if (str == bs) {
+		if (*(str + 1) == '_') {
 #ifdef JP_CHARSET
-		if (s->length > 1 && CharType(*(prop - 2)) == PC_KANJI1) {
-		    str += 2;
-		    *(prop - 1) |= PE_UNDER;
-		    *(prop - 2) |= PE_UNDER;
-		}
-		else
+		    if (s->length > 1 && CharType(*(prop - 2)) == PC_KANJI1) {
+			str += 2;
+			*(prop - 1) |= PE_UNDER;
+			*(prop - 2) |= PE_UNDER;
+		    }
+		    else
 #endif				/* JP_CHARSET */
-		if (s->length > 0) {
-		    str += 2;
-		    *(prop - 1) |= PE_UNDER;
+		    if (s->length > 0) {
+			str += 2;
+			*(prop - 1) |= PE_UNDER;
+		    }
+		    else {
+			str++;
+		    }
 		}
-		else {
-		    str++;
-		}
-		continue;
-	    }
-	    else if (!strncmp(str, "\b\b__", 4)) {
+		else if (!strncmp(str + 1, "\b__", 3)) {
 #ifdef JP_CHARSET
-		if (s->length > 1 && CharType(*(prop - 2)) == PC_KANJI1) {
-		    str += 4;
-		    *(prop - 1) |= PE_UNDER;
-		    *(prop - 2) |= PE_UNDER;
-		}
-		else
-#endif				/* JP_CHARSET */
-		if (s->length > 0) {
-		    str += 3;
-		    *(prop - 1) |= PE_UNDER;
-		}
-		else {
-		    str += 2;
-		}
-		continue;
-	    }
-	    else if (!strncmp(str, "\b\b", 2)) {
-#ifdef JP_CHARSET
-		if (s->length > 1 && CharType(*(prop - 2)) == PC_KANJI1) {
-		    if (!strncmp(str - 2, str + 2, 2)) {
-			*(prop - 1) |= PE_BOLD;
-			*(prop - 2) |= PE_BOLD;
+		    if (s->length > 1 && CharType(*(prop - 2)) == PC_KANJI1) {
 			str += 4;
+			*(prop - 1) |= PE_UNDER;
+			*(prop - 2) |= PE_UNDER;
 		    }
-		    else {
-			Strshrink(s, 2);
-			prop -= 2;
-			str += 2;
-		    }
-		}
-		else
+		    else
 #endif				/* JP_CHARSET */
-		if (s->length > 0) {
-		    if (*(str - 1) == *(str + 2)) {
-			*(prop - 1) |= PE_BOLD;
+		    if (s->length > 0) {
 			str += 3;
+			*(prop - 1) |= PE_UNDER;
 		    }
 		    else {
-			Strshrink(s, 1);
-			prop--;
 			str += 2;
 		    }
 		}
-		else {
-		    str += 2;
-		}
-		continue;
-	    }
-	    else if (*str == '\b') {
+		else if (*(str + 1) == '\b') {
 #ifdef JP_CHARSET
-		if (s->length > 1 && CharType(*(prop - 2)) == PC_KANJI1) {
-		    if (!strncmp(str - 2, str + 1, 2)) {
-			*(prop - 1) |= PE_BOLD;
-			*(prop - 2) |= PE_BOLD;
-			str += 3;
+		    if (s->length > 1 && CharType(*(prop - 2)) == PC_KANJI1) {
+			if (str + 4 <= endp &&
+			    !strncmp(str - 2, str + 2, 2)) {
+			    *(prop - 1) |= PE_BOLD;
+			    *(prop - 2) |= PE_BOLD;
+			    str += 4;
+			}
+			else {
+			    Strshrink(s, 2);
+			    prop -= 2;
+			    str += 2;
+			}
 		    }
-		    else {
-			Strshrink(s, 2);
-			prop -= 2;
-			str++;
-		    }
-		}
-		else
+		    else
 #endif				/* JP_CHARSET */
-		if (s->length > 0) {
-		    if (*(str - 1) == *(str + 1)) {
-			*(prop - 1) |= PE_BOLD;
-			str += 2;
+		    if (s->length > 0) {
+			if (str + 3 <= endp &&
+			    *(str - 1) == *(str + 2)) {
+			    *(prop - 1) |= PE_BOLD;
+			    str += 3;
+			}
+			else {
+			    Strshrink(s, 1);
+			    prop--;
+			    str += 2;
+			}
 		    }
 		    else {
-			Strshrink(s, 1);
-			prop--;
-			str++;
+			str += 2;
 		    }
 		}
 		else {
-		    str++;
+#ifdef JP_CHARSET
+		    if (s->length > 1 && CharType(*(prop - 2)) == PC_KANJI1) {
+			if (str + 3 <= endp &&
+			    !strncmp(str - 2, str + 1, 2)) {
+			    *(prop - 1) |= PE_BOLD;
+			    *(prop - 2) |= PE_BOLD;
+			    str += 3;
+			}
+			else {
+			    Strshrink(s, 2);
+			    prop -= 2;
+			    str++;
+			}
+		    }
+		    else
+#endif				/* JP_CHARSET */
+		    if (s->length > 0) {
+			if (str + 2 <= endp &&
+			    *(str - 1) == *(str + 1)) {
+			    *(prop - 1) |= PE_BOLD;
+			    str += 2;
+			}
+			else {
+			    Strshrink(s, 1);
+			    prop--;
+			    str++;
+			}
+		    }
+		    else {
+			str++;
+		    }
 		}
+		if (str < endp)
+		    bs = memchr(str, '\b', endp - str);
 		continue;
 	    }
+#ifdef ANSI_COLOR
+	    else if (str > bs)
+		bs = memchr(str, '\b', endp - str);
+#endif
 	}
+#ifdef ANSI_COLOR
+	if (es != NULL) {
+	    if (str == es) {
+		int ok = parse_ansi_color(&str, &ceffect, &cmode);
+		if (str < endp)
+		    es = memchr(str, ESC_CODE, endp - str);
+		if (ok) {
+		    if (cmode)
+			*check_color = TRUE;
+		    continue;
+		}
+	    }
+	    else if (str > es)
+		es = memchr(str, ESC_CODE, endp - str);
+	}
+#endif
 
-	prev_mode = mode;
-	mode = get_ctype((unsigned char) *str, prev_mode);
+	mode = get_mctype(str);
+#ifdef ANSI_COLOR
+	effect |= ceffect;
+#endif
 #ifdef JP_CHARSET
-	if (prev_mode == PC_KANJI1 && mode != PC_KANJI2) {
-	    *(prop - 1) &= ~P_CHARTYPE;
-	    *(prop - 1) |= PC_ASCII;
+	if (mode == PC_KANJI) {
+	    prop[0] = (effect | PC_KANJI1);
+	    prop[1] = (effect | PC_KANJI2);
+	    if (do_copy) {
+		Strcat_char(s, str[0]);
+		Strcat_char(s, str[1]);
+	    }
+	    prop += 2;
+	    str += 2;
+#ifdef ANSI_COLOR
+	    if (color) {
+		color[0] = cmode;
+		color[1] = cmode;
+		color += 2;
+	    }
+#endif
 	}
-	if (mode == PC_KANJI2)
-	    *(prop++) = (leffect | mode);
 	else
 #endif				/* JP_CHARSET */
-	    *(prop++) = (effect | mode);
-	Strcat_char(s, *(str++));
-#ifdef JP_CHARSET
-	if (mode != PC_KANJI1)
-#endif				/* JP_CHARSET */
-	    effect = PE_NORMAL;
+	{
+	    *prop = (effect | mode);
+	    if (do_copy)
+		Strcat_char(s, *str);
+	    prop++;
+	    str++;
+#ifdef ANSI_COLOR
+	    if (color) {
+		*color = cmode;
+		color++;
+	    }
+#endif
+	}
+	effect = PE_NORMAL;
     }
     return s;
 }
 
 int
-calcPosition(char *l, int pos, int bpos, int mode)
+calcPosition(char *l, Lineprop *pr, int len, int pos, int bpos, int mode)
 {
-    static short realColumn[LINELEN];
+    static short realColumn[LINELEN + 1];
     static char *prevl = NULL;
     int i, j;
 
-    if (l == NULL || *l == '\0')
+    if (l == NULL || len == 0)
 	return bpos;
     if (l == prevl && mode == CP_AUTO) {
-        if (pos <= strlen(l))
-	return realColumn[pos];
+        if (pos <= len)
+	    return realColumn[pos];
     }
     prevl = l;
     j = bpos;
     for (i = 0;; i++) {
 	realColumn[i] = j;
-        if (l[i] == '\0')
+        if (i == len)
             break;
-	if (l[i] == '\t')
+	if (l[i] == '\t' && pr[i] == PC_CTRL)
 	    j = (j + Tabstop) / Tabstop * Tabstop;
-	else if ((!(l[i] & ~0x1f) && l[i] != CTRL_J) || l[i] == DEL_CODE)
+#ifndef KANJI_SYMBOLS
+	else if (pr[i] & PC_RULE)
+	    j++;
+#endif
+	else if (IS_UNPRINTABLE_ASCII(l[i], pr[i]))
+	    j = j + 4;
+	else if (IS_UNPRINTABLE_CONTROL(l[i], pr[i]))
 	    j = j + 2;
 	else
 	    j++;
@@ -502,72 +663,6 @@ bzero(void *ptr, int len)
 }
 #endif				/* NOBCOPY */
 
-/* fgets() substitute which terminates \n, \r or \n\r */
-char *
-myfgets(char *ptr, int len, FILE * stream)
-{
-    int read_len = 0;
-    char c, *p = ptr;
-
-    if (feof(stream))
-	return NULL;
-    while (c = getc(stream), !feof(stream)) {
-	*(p++) = c;
-	read_len++;
-	if (read_len == len - 1)
-	    break;
-	if (c == '\n')
-	    break;
-	if (c == '\r') {
-	    c = getc(stream);
-	    if (!feof(stream) && c != '\n')
-		ungetc(c, stream);
-	    *(p - 1) = '\n';
-	    break;
-	}
-    }
-    ptr[read_len] = '\0';
-    return ptr;
-}
-
-/* Strfgets() substitute which terminates \n, \r or \n\r */
-Str
-Strmyfgets(FILE * stream)
-{
-    char c;
-    Str p = Strnew();
-
-    if (feof(stream))
-	return p;
-    while (c = getc(stream), !feof(stream) && !ferror(stream)) {
-	if (c == '\0') {
-	    Strcat_charp(p, "^@");
-	    continue;
-	}
-	Strcat_char(p, c);
-	if (c == '\n')
-	    break;
-	if (c == '\r') {
-	    c = getc(stream);
-	    if (!feof(stream) && c != '\n')
-		ungetc(c, stream);
-	    p->ptr[p->length - 1] = '\n';
-	    break;
-	}
-    }
-    return p;
-}
-
-#ifdef JP_CHARSET
-Str
-Strkconv(Str s, int fcode, int tcode)
-{
-    if (fcode == tcode)
-	return s;
-    return conv(s->ptr, fcode, tcode);
-}
-#endif				/* JP_CHARSET */
-
 char *
 mybasename(char *s)
 {
@@ -583,6 +678,25 @@ mybasename(char *s)
     return allocStr(p, 0);
 }
 
+char *
+mydirname(char *s)
+{
+    char *p = s;
+    while (*p)
+	p++;
+    if (s != p)
+	p--;
+    while (s != p && *p == '/')
+	p--;
+    while (s != p && *p != '/')
+	p--;
+    if (*p != '/')
+	return ".";
+    while (s != p && *p == '/')
+	p--;
+    return allocStr(s, strlen(s) - strlen(p) + 1);
+}
+
 #ifndef STRERROR
 char *
 strerror(int errno)
@@ -752,10 +866,10 @@ read_token(Str buf, char **instr, int *status, int pre, int append)
     char *p;
     int prev_status;
 
-    if (**instr == '\0')
-	return 0;
     if (!append)
 	Strclear(buf);
+    if (**instr == '\0')
+	return 0;
     for (p = *instr; *p; p++) {
 	prev_status = *status;
 	next_status(*p, status);
@@ -772,24 +886,7 @@ read_token(Str buf, char **instr, int *status, int pre, int append)
 		p++;
 		goto proc_end;
 	    }
-	    if (*p == ' ' || *p == '\t') {
-		if (pre)
-		    Strcat_char(buf, *p);
-		else
-		    Strcat_char(buf, ' ');
-	    }
-	    else if (*p == '\n' || *p == '\r') {
-		if (pre) {
-		    if ((p[0] == '\n' && p[1] == '\r') ||
-			(p[0] == '\r' && p[1] == '\n'))
-			p++;
-		    Strcat_char(buf, '\n');
-		}
-		else
-		    Strcat_char(buf, ' ');
-	    }
-	    else
-		Strcat_char(buf, *p);
+	    Strcat_char(buf, (!pre && IS_SPACE(*p)) ? ' ' : *p);
 	    if (ST_IS_REAL_TAG(prev_status)) {
 		*instr = p + 1;
 		if (buf->length < 2 ||
@@ -840,6 +937,46 @@ read_token(Str buf, char **instr, int *status, int pre, int append)
     return 1;
 }
 
+Str
+correct_irrtag(int status)
+{
+    char c;
+    Str tmp = Strnew();
+
+    while (status != R_ST_NORMAL) {
+	switch (status) {
+	case R_ST_CMNT:		/* required "-->" */
+	case R_ST_NCMNT1:	/* required "->" */
+	    c = '-';
+	    break;
+	case R_ST_NCMNT2:
+	case R_ST_NCMNT3:
+	case R_ST_IRRTAG:
+	case R_ST_CMNT1:
+	case R_ST_CMNT2:
+	case R_ST_TAG:
+	case R_ST_TAG0:
+	case R_ST_EQL:		/* required ">" */
+	    c = '>';
+	    break;
+	case R_ST_QUOTE:
+	    c = '\'';
+	    break;
+	case R_ST_DQUOTE:
+	    c = '"';
+	    break;
+	case R_ST_AMP:
+	    c = ';';
+	    break;
+	default:
+	    return tmp;
+	}
+	next_status(c, &status);
+	Strcat_char(tmp, c);
+    }
+    return tmp;
+}
+
 /* authentication */
 struct auth_cookie *
 find_auth(char *host, char *realm)
@@ -904,59 +1041,6 @@ last_modified(Buffer * buf)
     return "unknown";
 }
 
-int
-is_kanji1(unsigned char ch, char code)
-{
-    switch (code) {
-    case 'E':
-	if (ch >= 0xa1 && ch <= 0xfe)
-	    return 1;
-	else
-	    return 0;
-    case 'S':
-	if ((ch >= 0x81 && ch <= 0x9f) || (ch >= 0xe0 && ch <= 0xfc))
-	    return 1;
-	else
-	    return 0;
-    }
-    return 0;
-}
-
-int
-is_kanji2(unsigned char ch, char code)
-{
-    switch (code) {
-    case 'E':
-	if (ch >= 0xa1 && ch <= 0xfe)
-	    return 1;
-	else
-	    return 0;
-    case 'S':
-	if ((ch >= 0x40 && ch <= 0x7e) || (ch >= 0x80 && ch <= 0xfc))
-	    return 1;
-	else
-	    return 0;
-    }
-    return 0;
-}
-
-Lineprop
-get_ctype(unsigned char ch, Lineprop prev_mode)
-{
-#ifdef JP_CHARSET
-    if (prev_mode == PC_KANJI1 && is_kanji2(ch, InnerCode))
-	return PC_KANJI2;
-
-    if (is_kanji1(ch, InnerCode))
-	return PC_KANJI1;
-    else
-#endif				/* JP_CHARSET */
-    if (!(ch & ~0x1f))
-	return PC_CTRL;
-    else
-	return PC_ASCII;
-}
-
 static char roman_num1[] =
 {
     'i', 'x', 'c', 'm', '*',
@@ -1039,282 +1123,6 @@ romanAlphabet(int n)
     return r;
 }
 
-#ifdef USE_SSL
-SSLStream
-newSSLStream(SSL * handle, int sock)
-{
-    SSLStream ss;
-    if (sock < 0)
-	return NULL;
-    ss = New(struct ssl_stream);
-    ss->pos = ss->epos = 0;
-    ss->handle = handle;
-    ss->sock = sock;
-    ss->status = -1;
-    ss->bufsize = SSL_BUF_SIZE;
-    ss->buf = NewAtom_N(char, ss->bufsize);
-    return ss;
-}
-
-void
-SSclose(SSLStream ss)
-{
-    if (!ss)
-	return;
-
-    close(ss->sock);
-    if (ss->handle)
-	SSL_free(ss->handle);
-}
-
-static int
-SSgetc(SSLStream ss)
-{
-    if (ss->epos <= ss->pos) {
-	ss->pos = ss->epos = 0;
-	if (ss->handle) {
-#ifdef USE_SSL_VERIFY
-	    for (;;) {
-		ss->status = SSL_read(ss->handle, ss->buf, ss->bufsize);
-		if (ss->status > 0)
-		    break;
-		switch (SSL_get_error(ss->handle, ss->status)) {
-		case SSL_ERROR_WANT_READ:
-		    continue;
-		default:
-		    break;
-		}
-		break;
-	    }
-#else				/* if !defined(USE_SSL_VERIFY) */
-	    ss->status = SSL_read(ss->handle, ss->buf, ss->bufsize);
-#endif				/* !defined(USE_SSL_VERIFY) */
-	}
-	else
-	    ss->status = read(ss->sock, ss->buf, ss->bufsize);
-	if (ss->status <= 0) {
-	    return '\0';
-	}
-	ss->epos += ss->status;
-    }
-    return ss->buf[ss->pos++];
-}
-
-static int
-SSundogetc(SSLStream ss)
-{
-    if (ss && ss->pos > 0) {
-	ss->pos--;
-	return 0;
-    }
-    return -1;
-}
-
-Str
-StrSSgets(SSLStream ss)
-{
-    char c = '\0';
-    Str p = Strnew();
-
-    if (!ss)
-	return p;
-
-    while (c = SSgetc(ss), ss->status > 0) {
-	Strcat_char(p, c);
-	if (c == '\n')
-	    break;
-    }
-    return p;
-}
-
-Str
-StrmySSgets(SSLStream ss)
-{
-    char c = '\0';
-    Str p = Strnew();
-
-    if (!ss)
-	return p;
-
-    while (c = SSgetc(ss), ss->status > 0) {
-	if (c == '\0') {
-	    Strcat_charp(p, "^@");
-	    continue;
-	}
-	Strcat_char(p, c);
-	if (c == '\n')
-	    break;
-	if (c == '\r') {
-	    c = SSgetc(ss);
-	    if (ss->status > 0 && c != '\n')
-		SSundogetc(ss);
-	    p->ptr[p->length - 1] = '\n';
-	    break;
-	}
-    }
-    return p;
-}
-#endif				/* USE_SSL */
-
-static Str
-sgets(char **str)
-{
-    int i;
-    Str ptr = Strnew();
-    char c;
-
-    if (*str == NULL)
-	return ptr;
-
-    for (i = 0; (*str)[i]; i++) {
-	c = (*str)[i];
-	Strcat_char(ptr, c);
-	if (c == '\n') {
-	    (*str) += i + 1;
-	    return ptr;
-	}
-    }
-    (*str) += i;
-    return ptr;
-}
-
-Str
-StrUFgets(URLFile * f)
-{
-    Str s = NULL;
-
-    switch (f->stream_type) {
-    case SMT_STRING:
-	s = sgets(&f->stream.is);
-	if (!f->stream.is || f->stream.is[0] == '\0')
-	    f->iseof = TRUE;
-	break;
-#ifdef USE_SSL
-    case SMT_SSL:
-	s = StrSSgets(f->stream.ss);
-	if (!f->stream.ss || f->stream.ss->status <= 0)
-	    f->iseof = TRUE;
-	break;
-#endif				/* USE_SSL */
-    case SMT_FILE:
-    default:
-	s = Strfgets(f->stream.f);
-	if (!f->stream.f || feof(f->stream.f) || ferror(f->stream.f))
-	    f->iseof = TRUE;
-	break;
-    }
-
-    return s;
-}
-
-Str
-StrmyUFgets(URLFile * f)
-{
-    Str s = NULL;
-
-    switch (f->stream_type) {
-    case SMT_STRING:
-	s = sgets(&f->stream.is);
-	if (!f->stream.is || f->stream.is[0] == '\0')
-	    f->iseof = TRUE;
-	break;
-#ifdef USE_SSL
-    case SMT_SSL:
-	s = StrmySSgets(f->stream.ss);
-	if (!f->stream.ss || f->stream.ss->status <= 0)
-	    f->iseof = TRUE;
-	break;
-#endif				/* USE_SSL */
-    case SMT_FILE:
-    default:
-	s = Strmyfgets(f->stream.f);
-	if (!f->stream.f || feof(f->stream.f) || ferror(f->stream.f))
-	    f->iseof = TRUE;
-	break;
-    }
-
-    return s;
-}
-
-static int _prev_c = '\0';
-int
-UFgetc(URLFile * f)
-{
-    switch (f->stream_type) {
-    case SMT_STRING:
-	if (f->stream.is)
-	    _prev_c = *(f->stream.is++);
-	if (!f->stream.is || f->stream.is[0] == '\0')
-	    f->iseof = TRUE;
-	break;
-#ifdef USE_SSL
-    case SMT_SSL:
-	if (f->stream.ss)
-	    _prev_c = SSgetc(f->stream.ss);
-	if (!f->stream.ss || f->stream.ss->status <= 0)
-	    f->iseof = TRUE;
-	break;
-#endif				/* USE_SSL */
-    case SMT_FILE:
-    default:
-	if (f->stream.f)
-	    _prev_c = getc(f->stream.f);
-	if (!f->stream.f || feof(f->stream.f) || ferror(f->stream.f))
-	    f->iseof = TRUE;
-	break;
-    }
-    return _prev_c;
-}
-
-void
-UFundogetc(URLFile * f)
-{
-    switch (f->stream_type) {
-    case SMT_STRING:
-	if (f->stream.is)
-	    f->stream.is--;
-	break;
-#ifdef USE_SSL
-    case SMT_SSL:
-	if (f->stream.ss)
-	    SSundogetc(f->stream.ss);
-	break;
-#endif				/* USE_SSL */
-    case SMT_FILE:
-    default:
-	if (f->stream.f)
-	    ungetc(_prev_c, f->stream.f);
-	break;
-    }
-}
-
-void
-UFclose0(URLFile * f, void (*closep) ())
-{
-    if (!f->stream.f)
-	return;
-    switch (f->stream_type) {
-    case SMT_STRING:
-	break;
-#ifdef USE_SSL
-    case SMT_SSL:
-	SSclose(f->stream.ss);
-	break;
-#endif				/* USE_SSL */
-    case SMT_FILE:
-    default:
-	if (closep)
-	    closep(f->stream.f);
-	break;
-    }
-}
-
-void
-UFclose(URLFile * f)
-{
-    UFclose0(f, f->close);
-}
-
 Str
 quoteShell(char *string)
 {
@@ -1332,10 +1140,19 @@ quoteShell(char *string)
 void
 mySystem(char *command, int background)
 {
+#ifdef __EMX__ /* jsawa */
+    if (background){
+	Str cmd = Strnew_charp("start /f ");
+	Strcat_charp(cmd, command);
+	system(cmd->ptr);
+    }else
+	system(command);
+#else
     Str cmd = Strnew_charp(command);
     if (background)
-	Strcat_char(cmd, '&');
+	cmd = Sprintf("(%s) >/dev/null 2>&1 &", cmd->ptr);
     system(cmd->ptr);
+#endif
 }
 
 char *
@@ -1372,6 +1189,26 @@ expandName(char *name)
     return extpath->ptr;
 }
 
+static char *tmpf_base[MAX_TMPF_TYPE] =
+{
+    "tmp", "src", "frame", "cache" 
+};
+static unsigned int tmpf_seq[MAX_TMPF_TYPE];
+    
+Str
+tmpfname(int type, char *ext)
+{
+    Str tmpf;
+    tmpf = Sprintf("%s/w3m%s%d-%d%s",
+		   rc_dir,
+		   tmpf_base[type],
+		   (int) getpid(),
+		   tmpf_seq[type]++,
+		   (ext)? ext : ""
+		   );
+    return tmpf;
+}
+
 #ifdef USE_COOKIE
 static char *monthtbl[] =
 {
@@ -1389,7 +1226,7 @@ get_day(char **s)
     if (!**s)
 	return -1;
 
-    while (**s && isdigit((unsigned char) **s))
+    while (**s && IS_DIGIT(**s))
 	Strcat_char(tmp, *((*s)++));
 
     day = atoi(tmp->ptr);
@@ -1411,13 +1248,13 @@ get_month(char **s)
     if (!**s)
 	return -1;
 
-    while (**s && isdigit((unsigned char) **s))
+    while (**s && IS_DIGIT(**s))
 	Strcat_char(tmp, *((*s)++));
     if (tmp->length > 0) {
 	mon = atoi(tmp->ptr);
     }
     else {
-	while (**s && isalpha((unsigned char) **s))
+	while (**s && IS_ALPHA(**s))
 	    Strcat_char(tmp, *((*s)++));
 	for (mon = 1; mon <= 12; mon++) {
 	    if (strncmp(tmp->ptr, monthtbl[mon - 1], 3) == 0)
@@ -1441,7 +1278,7 @@ get_year(char **s)
     if (!**s)
 	return -1;
 
-    while (**s && isdigit((unsigned char) **s))
+    while (**s && IS_DIGIT(**s))
 	Strcat_char(tmp, *((*s)++));
     if (tmp->length != 2 && tmp->length != 4) {
 	*s = ss;
@@ -1467,7 +1304,7 @@ get_time(char **s, int *hour, int *min, int *sec)
     if (!**s)
 	return -1;
 
-    while (**s && isdigit((unsigned char) **s))
+    while (**s && IS_DIGIT(**s))
 	Strcat_char(tmp, *((*s)++));
     if (**s != ':') {
 	*s = ss;
@@ -1477,7 +1314,7 @@ get_time(char **s, int *hour, int *min, int *sec)
 
     (*s)++;
     Strclear(tmp);
-    while (**s && isdigit((unsigned char) **s))
+    while (**s && IS_DIGIT(**s))
 	Strcat_char(tmp, *((*s)++));
     if (**s != ':') {
 	*s = ss;
@@ -1487,7 +1324,7 @@ get_time(char **s, int *hour, int *min, int *sec)
 
     (*s)++;
     Strclear(tmp);
-    while (**s && isdigit((unsigned char) **s))
+    while (**s && IS_DIGIT(**s))
 	Strcat_char(tmp, *((*s)++));
     *sec = atoi(tmp->ptr);
 
@@ -1515,27 +1352,27 @@ mymktime(char *timestr)
     fprintf(stderr, "mktime: %s\n", timestr);
 #endif				/* DEBUG */
 
-    while (*s && isalpha((unsigned char) *s))
+    while (*s && IS_ALPHA(*s))
 	s++;
-    while (*s && !isalnum((unsigned char) *s))
+    while (*s && !IS_ALNUM(*s))
 	s++;
 
-    if (isdigit(*s)) {
+    if (IS_DIGIT(*s)) {
 	/* RFC 1123 or RFC 850 format */
 	if ((day = get_day(&s)) == -1)
 	    return -1;
 
-	while (*s && !isalnum((unsigned char) *s))
+	while (*s && !IS_ALNUM(*s))
 	    s++;
 	if ((mon = get_month(&s)) == -1)
 	    return -1;
 
-	while (*s && !isdigit((unsigned char) *s))
+	while (*s && !IS_DIGIT(*s))
 	    s++;
 	if ((year = get_year(&s)) == -1)
 	    return -1;
 
-	while (*s && !isdigit((unsigned char) *s))
+	while (*s && !IS_DIGIT(*s))
 	    s++;
 	if (!*s) {
 	    hour = 0;
@@ -1548,22 +1385,22 @@ mymktime(char *timestr)
     }
     else {
 	/* ANSI C asctime() format. */
-	while (*s && !isalnum((unsigned char) *s))
+	while (*s && !IS_ALNUM(*s))
 	    s++;
 	if ((mon = get_month(&s)) == -1)
 	    return -1;
 
-	while (*s && !isdigit((unsigned char) *s))
+	while (*s && !IS_DIGIT(*s))
 	    s++;
 	if ((day = get_day(&s)) == -1)
 	    return -1;
 
-	while (*s && !isdigit((unsigned char) *s))
+	while (*s && !IS_DIGIT(*s))
 	    s++;
 	if (get_time(&s, &hour, &min, &sec) == -1)
 	    return -1;
 
-	while (*s && !isdigit((unsigned char) *s))
+	while (*s && !IS_DIGIT(*s))
 	    s++;
 	if ((year = get_year(&s)) == -1)
 	    return -1;
@@ -1603,6 +1440,9 @@ FQDN(char *host)
     if (host == NULL)
 	return NULL;
 
+    if (strcasecmp(host, "localhost") == 0)
+	return host;
+
     for (p = host; *p && *p != '.'; p++);
 
     if (*p == '.')
