@@ -1,4 +1,4 @@
-/* $Id: fm.h,v 1.24 2000/06/06 05:39:13 aito Exp $ */
+/* $Id: fm.h,v 1.7.2.33 2001/01/22 12:47:06 okabe Exp $ */
 /* 
  * w3m: WWW wo Miru utility
  * 
@@ -16,11 +16,16 @@
 #include <unistd.h>
 #include <sys/types.h>
 #include "config.h"
+#include "history.h"
 
 #ifdef MENU
 #define MENU_SELECT
 #endif				/* MENU */
 
+#ifndef COLOR
+#undef ANSI_COLOR
+#endif
+
 #include "ctrlcode.h"
 #include "html.h"
 #include "gc.h"
@@ -28,10 +33,12 @@
 #include "form.h"
 #include "frame.h"
 #include "parsetag.h"
+#include "parsetagx.h"
 #include "func.h"
 #include "menu.h"
 #include "textlist.h"
 #include "funcname1.h"
+#include "terms.h"
 
 #ifdef NOBCOPY
 void bcopy(void *, void *, int);
@@ -59,7 +66,9 @@ void bzero(void *, int);
 #define PAGER_MAX_LINE	10000	/* Maximum line kept as pager */
 #define FNLEN 80
 
-#define PIXEL_PER_CHAR  8	/* arbitrary */
+#define DEFAULT_PIXEL_PER_CHAR  8.0	/* arbitrary */
+#define MINIMUM_PIXEL_PER_CHAR  4.0
+#define MAXIMUM_PIXEL_PER_CHAR  32.0
 
 #ifdef FALSE
 #undef FALSE
@@ -72,7 +81,9 @@ void bzero(void *, int);
 #define FALSE 0
 #define TRUE   1
 
-#define INTERNAL_CODE    'E'	/* use EUC-JP internally; do not change */
+#ifdef USE_COOKIE
+#define PERHAPS 2
+#endif
 
 #define SHELLBUFFERNAME	"*Shellout*"
 #define PIPEBUFFERNAME	"*stream*"
@@ -85,30 +96,30 @@ void bzero(void *, int);
 /* 
  * Line Property
  */
-#ifdef JP_CHARSET
 /* Character type */
-#define P_CHARTYPE	0xe000
-#define PC_ASCII	0x0000
-#define PC_KANJI1	0x2000
-#define PC_KANJI2	0x4000
-#define PC_CTRL		0x8000
-
-#else				/* ISO-8859-1 charset (not JP_CHARSET) */
-
-#define P_CHARTYPE	0x2000
 #define PC_ASCII	0x0000
 #define PC_CTRL		0x2000
+
+#ifdef JP_CHARSET
+#define PC_KANJI1	0x4000
+#define PC_KANJI2	0x8000
+#define PC_KANJI	(PC_KANJI1|PC_KANJI2)
+#define P_CHARTYPE	(PC_ASCII|PC_CTRL|PC_KANJI)
+#else				/* ISO-8859-1 charset (not JP_CHARSET) */
+#define P_CHARTYPE	(PC_ASCII|PC_CTRL)
 #endif				/* not JP_CHARSET */
+#if 0
+#define GET_PCTYPE(c)   ((GET_MYCTYPE(c)&MYCTYPE_CNTRL)<<13)
+#else
+#define GET_PCTYPE(c)   ((GET_MYCTYPE(c)&MYCTYPE_CNTRL)?PC_CTRL:PC_ASCII)
+#endif
 
 #ifndef KANJI_SYMBOLS
-#define PC_RULE         0x0f00
-#define PC_ISRULE       0x1000
-#define PC_RULE_FLAG(x) (((x)&0xf)<<8)
-#define PC_RULE_VAL(x)  (((x)&PC_RULE)>>8)
+#define PC_RULE         0x1000
 #endif				/* not KANJI_SYMBOLS */
 
 /* Effect ( standout/underline ) */
-#define P_EFFECT	0xfe
+#define P_EFFECT	0x01fe
 #define PE_NORMAL	0x00
 #define PE_UNDER	0x02
 #define PE_STAND	0x04
@@ -118,6 +129,7 @@ void bzero(void *, int);
 #define PE_IMAGE        0x20
 #define PE_FORM         0x40
 #define PE_ACTIVE	0x80
+#define PE_VISITED	0x0100
 
 /* Mark */
 #define PM_MARK		0x01
@@ -126,12 +138,18 @@ void bzero(void *, int);
 #ifdef KANJI_SYMBOLS
 #define CharEffect(c)	((c)&P_EFFECT)
 #else				/* not KANJI_SYMBOLS */
-#define CharEffect(c)	((c)&(P_EFFECT|PC_ISRULE|PC_RULE))
+#define CharEffect(c)	((c)&(P_EFFECT|PC_RULE))
 #endif				/* not KANJI_SYMBOLS */
 #define SetCharType(v,c)	((v)=(((v)&~P_CHARTYPE)|(c)))
 
 
-#define COLPOS(l,c)	calcPosition(l->lineBuf,c,0,CP_AUTO)
+#define COLPOS(l,c)	calcPosition(l->lineBuf,l->propBuf,l->len,c,0,CP_AUTO)
+#define IS_UNPRINTABLE_CONTROL(c,m) (CharType(m)==PC_CTRL&&(c)!=CTRL_I&&(c)!=CTRL_J)
+#ifdef JP_CHARSET
+#define IS_UNPRINTABLE_ASCII(c,m) (!IS_ASCII(c)&&CharType(m)==PC_ASCII)
+#else
+#define IS_UNPRINTABLE_ASCII(c,m) (!IS_LATIN1(c))
+#endif
 
 /* Flags for displayBuffer() */
 #define B_NORMAL	0
@@ -147,6 +165,7 @@ void bzero(void *, int);
 #define BP_INTERNAL	0x8
 #define BP_NO_URL	0x10
 #define BP_REDIRECTED   0x20
+#define BP_CLOSE        0x40
 
 /* Link Buffer */
 #define LB_NOLINK	-1
@@ -185,12 +204,14 @@ extern int REV_LB[];
 #define CPL_OFF		0x1
 #define CPL_ON		0x2
 #define CPL_ALWAYS	0x4
+#define CPL_URL		0x8
 
 /* Flags for inputLine() */
 #define IN_STRING	0x10
 #define IN_FILENAME	0x20
 #define IN_PASSWORD	0x40
 #define IN_COMMAND	0x80
+#define IN_URL		0x100
 
 /* 
  * Macros.
@@ -216,24 +237,41 @@ extern int REV_LB[];
 #define IS_ENDL(c) ((c)=='\0'||(c)=='\r'||(c)=='\n')
 #define IS_ENDT(c) (IS_ENDL(c)||(c)==';')
 
+#define bpcmp(a,b) \
+  (((a).line - (b).line) ? ((a).line - (b).line) : ((a).pos - (b).pos))
+
+#define RELATIVE_WIDTH(w)   (((w)>=0)?(int)((w)/pixel_per_char):(w))
+#define REAL_WIDTH(w,limit) (((w)>=0)?(int)((w)/pixel_per_char):-(w)*(limit)/100)
+
+#define EOL(l) (&(l)->ptr[(l)->length])
+#define IS_EOL(p,l) ((p)==&(l)->ptr[(l)->length])
+
 /* 
  * Types.
  */
 
 typedef unsigned short Lineprop;
+#ifdef ANSI_COLOR
+typedef unsigned char Linecolor;
+#endif
 
 typedef struct _MapList {
     Str name;
     TextList *urls;
+    TextList *alts;
     struct _MapList *next;
 } MapList;
 
 typedef struct _Line {
     char *lineBuf;
     Lineprop *propBuf;
+#ifdef ANSI_COLOR
+    Linecolor *colorBuf;
+#endif
     struct _Line *next;
     struct _Line *prev;
     short len;
+    short width;
     long linenumber;		/* on buffer */
     long real_linenumber;	/* on file */
     unsigned short usrflags;
@@ -281,15 +319,16 @@ typedef struct _Buffer {
     short width;
     short height;
     char *type;
+    char *real_type;
+    char encoding;
     int allLine;
     short bufferprop;
     short currentColumn;
-    short currentLineNum;
     short cursorX;
     short cursorY;
     short pos;
     short visualpos;
-    FILE *pagerSource;
+    InputStream pagerSource;
     AnchorList *href;
     AnchorList *name;
     AnchorList *img;
@@ -305,6 +344,7 @@ typedef struct _Buffer {
     struct frameset *frameset;
     struct frameset_queue *frameQ;
     int *clone;
+    int linelen;
     int trbyte;
     char check_url;
 #ifdef JP_CHARSET
@@ -312,6 +352,11 @@ typedef struct _Buffer {
 #endif				/* JP_CHARSET */
     TextList *document_header;
     FormItemList *form_submit;
+    char *savecache;
+    char *edit;
+#ifdef USE_SSL
+    char *ssl_certificate;
+#endif
 } Buffer;
 
 #define NO_BUFFER ((Buffer*)1)
@@ -345,7 +390,7 @@ struct readbuffer {
     Str line;
     Lineprop cprop;
     short pos;
-    char prevchar;
+    int prevchar;
     long flag;
     long flag_stack[RB_STACK_SIZE];
     int flag_sp;
@@ -383,16 +428,16 @@ struct readbuffer {
 #define RB_PRE_INT	0x100
 #define RB_PREMODE	(RB_PRE | RB_PRE_INT)
 #define RB_SPECIAL	(RB_PRE|RB_XMPMODE|RB_LSTMODE|RB_PLAIN|RB_NOBR|RB_PRE_INT)
+#define RB_PLAINMODE	(RB_XMPMODE|RB_LSTMODE|RB_PLAIN)
 
 #define RB_IN_DT	0x200
 #define RB_INTXTA	0x400
-#ifdef MENU
 #define RB_INSELECT	0x800
-#endif				/* MENU */
 #define RB_IGNORE	0x1000
 #define RB_INSEL	0x2000
 #define RB_IGNORE_P	0x4000
 #define RB_TITLE	0x8000
+#define RB_NFLUSHED	0x10000
 
 #ifdef FORMAT_NICE
 #define RB_FILL		0x200000
@@ -435,18 +480,21 @@ struct readbuffer {
 
 /* flags for loadGeneralFile */
 #define RG_NOCACHE   1
+#define RG_FRAME     2
 
 struct html_feed_environ {
     struct readbuffer *obuf;
-    TextList *buf;
+    TextLineList *buf;
     FILE *f;
     Str tagbuf;
     int limit;
+    int maxlimit;
     struct environment *envs;
     int nenv;
     int envc;
     int envc_real;
     char *title;
+    int blank_lines;
 };
 
 struct auth_cookie {
@@ -481,15 +529,23 @@ struct cookie {
 #define COO_DOMAIN	4
 #define COO_PATH	8
 #define COO_DISCARD	16
-#endif				/* USE_COOKIE */
+#define COO_OVERRIDE	32			/* user chose to override security checks */
 
-typedef struct _Hist {
-    int offset;
-    int position;
-    int length;
-    int size;
-    char **line;
-} Hist;
+#define COO_OVERRIDE_OK	32			/* flag to specify that an error is overridable */
+						/* version 0 refers to the original cookie_spec.html */
+						/* version 1 refers to RFC 2109 */
+						/* version 1' refers to the Internet draft to obsolete RFC 2109 */
+#define COO_EINTERNAL	(1)			/* unknown error; probably forgot to convert "return 1" in cookie.c */
+#define COO_ETAIL	(2 | COO_OVERRIDE_OK)	/* tail match failed (version 0) */
+#define COO_ESPECIAL	(3)			/* special domain check failed (version 0) */
+#define COO_EPATH	(4)			/* Path attribute mismatch (version 1 case 1) */
+#define COO_ENODOT	(5 | COO_OVERRIDE_OK)	/* no embedded dots in Domain (version 1 case 2.1) */
+#define COO_ENOTV1DOM	(6 | COO_OVERRIDE_OK)	/* Domain does not start with a dot (version 1 case 2.2) */
+#define COO_EDOM	(7 | COO_OVERRIDE_OK)	/* domain-match failed (version 1 case 3) */
+#define COO_EBADHOST	(8 | COO_OVERRIDE_OK)	/* dot in matched host name in FQDN (version 1 case 4) */
+#define COO_EPORT	(9)			/* Port match failed (version 1' case 5) */
+#define COO_EMAX	COO_EPORT
+#endif				/* USE_COOKIE */
 
 /* modes for align() */
 
@@ -497,6 +553,10 @@ typedef struct _Hist {
 #define ALIGN_LEFT   1
 #define ALIGN_RIGHT  2
 
+#define VALIGN_MIDDLE 0
+#define VALIGN_TOP    1
+#define VALIGN_BOTTOM 2
+
 typedef struct http_request {
     char command;
     char flag;
@@ -516,6 +576,14 @@ typedef struct http_request {
 #define HTST_NORMAL		0
 #define HTST_CONNECT		1
 
+#define TMPF_DFL	0
+#define TMPF_SRC	1
+#define TMPF_FRAME	2
+#define TMPF_CACHE	3
+#define MAX_TMPF_TYPE	4
+
+#define set_no_proxy(domains) (NO_proxy_domains=make_domain_list(domains))
+
 /* 
  * Globals.
  */
@@ -531,7 +599,7 @@ global int Tabstop init(8);
 global int ShowEffect init(TRUE);
 global int PagerMax init(PAGER_MAX_LINE);
 #ifdef JP_CHARSET
-global char InnerCode init(INTERNAL_CODE);
+global char InnerCode init(CODE_INNER_EUC);	/* use EUC-JP internally; do not change */
 #endif				/* JP_CHARSET */
 
 global char SearchHeader init(FALSE);
@@ -546,13 +614,21 @@ extern char GlobalKeymap[];
 extern char EscKeymap[];
 extern char EscBKeymap[];
 extern char EscDKeymap[];
+#ifdef __EMX__
+extern char	PcKeymap[];
+#endif
 extern FuncList w3mFuncList[];
+extern KeyList w3mKeyList;
 
 global char *HTTP_proxy init(NULL);
+#ifdef USE_GOPHER
 global char *GOPHER_proxy init(NULL);
+#endif				/* USE_GOPHER */
 global char *FTP_proxy init(NULL);
 global ParsedURL HTTP_proxy_parsed;
+#ifdef USE_GOPHER
 global ParsedURL GOPHER_proxy_parsed;
+#endif				/* USE_GOPHER */
 global ParsedURL FTP_proxy_parsed;
 global char *NO_proxy init(NULL);
 global int NOproxy_netaddr init(TRUE);
@@ -569,8 +645,14 @@ global char *personal_document_root init(NULL);
 global char *cgi_bin init(NULL);
 global char *index_file init(NULL);
 
+global char *CurrentDir;
 global Buffer *Currentbuf;
 global Buffer *Firstbuf;
+global int CurrentKey;
+global char *CurrentKeyData;
+#ifdef MENU
+global char *CurrentMenuData;
+#endif
 extern char *ullevel[];
 
 extern char *version;
@@ -590,6 +672,8 @@ global int bg_color init(8);	/* don't change */
 #endif				/* BG_COLOR */
 global int useActiveColor init(FALSE);
 global int active_color init(6);	/* cyan */
+global int useVisitedColor init(FALSE);
+global int visited_color init(5);	/* magenta  */
 #endif				/* COLOR */
 global int confirm_on_quit init(TRUE);
 global int displayLink init(FALSE);
@@ -607,9 +691,12 @@ global char *UserAgent init(NULL);
 global int NoSendReferer init(FALSE);
 global char *AcceptLang init(NULL);
 global int WrapDefault init(FALSE);
+global int IgnoreCase init(TRUE);
 global int WrapSearch init(FALSE);
 global int squeezeBlankLine init(FALSE);
 global char *BookmarkFile init(NULL);
+global char *pauth init(NULL);
+global Str proxy_auth_cookie init(NULL);
 global int UseExternalDirBuffer init(TRUE);
 #ifdef __EMX__
 global char *DirBufferCommand init("file:///$LIB/dirlist.cmd");
@@ -620,30 +707,41 @@ global char *config_file init(NULL);
 global int ignore_null_img_alt init(TRUE);
 
 global struct auth_cookie *Auth_cookie init(NULL);
+global char *Local_cookie init(NULL);
 #ifdef USE_COOKIE
 global struct cookie *First_cookie init(NULL);
 #endif				/* USE_COOKIE */
 
-global struct mailcap *UserMailcap;
-global struct mailcap *SysMailcap;
-global struct table2 *UserMimeTypes;
-global struct table2 *SysMimeTypes;
+global struct mailcap **UserMailcap;
+global struct table2 **UserMimeTypes;
+global TextList *mailcap_list;
+global TextList *mimetypes_list;
+global char *mailcap_files init(USER_MAILCAP ", " SYS_MAILCAP);
+global char *mimetypes_files init(USER_MIMETYPES ", " SYS_MIMETYPES);
 
 global TextList *fileToDelete;
 
+extern Hist *LoadHist;
+extern Hist *SaveHist;
+extern Hist *URLHist;
+extern Hist *ShellHist;
+extern Hist *TextHist;
 #ifdef USE_HISTORY
 global int URLHistSize init(100);
-global int URLSaveHist init(TRUE);
+global int SaveURLHist init(TRUE);
 #endif				/* USE_HISTORY */
 global int multicolList init(FALSE);
 
 #ifdef JP_CHARSET
 extern char DisplayCode;
+extern char DocumentCode;
 #endif				/* JP_CHARSET */
 #ifndef KANJI_SYMBOLS
 global int no_graphic_char init(FALSE);
+extern char alt_rule[];
 #endif				/* not KANJI_SYMBOLS */
-extern char *rc_dir;
+global char *rc_dir;
+global int rc_dir_is_tmp init(FALSE);
 
 #ifdef MOUSE
 global int use_mouse init(TRUE);
@@ -655,6 +753,11 @@ global int reverse_mouse init(FALSE);
 global int default_use_cookie init(TRUE);
 global int use_cookie init(TRUE);
 global int accept_cookie init(FALSE);
+global int accept_bad_cookie init(FALSE);
+global char *cookie_reject_domains init(NULL);
+global char *cookie_accept_domains init(NULL);
+global TextList *Cookie_reject_domains;
+global TextList *Cookie_accept_domains;
 #endif				/* USE_COOKIE */
 
 #ifdef VIEW_UNSEENOBJECTS
@@ -670,9 +773,36 @@ global char *ssl_ca_file init(NULL);
 global int ssl_path_modified init(FALSE);
 #endif				/* defined(USE_SSL) &&
 				 * defined(USE_SSL_VERIFY) */
+#ifdef USE_SSL
+global char *ssl_forbid_method init(NULL);
+#endif
 
 global int is_redisplay init(FALSE);
 global int clear_buffer init(TRUE);
+global double pixel_per_char init(DEFAULT_PIXEL_PER_CHAR);
+global int use_lessopen init(FALSE);
+
+#ifdef JP_CHARSET
+#define is_kanji(s)    (IS_KANJI1((s)[0])&&IS_KANJI2((s)[1]))
+#define get_mctype(s)  (is_kanji(s)?PC_KANJI:GET_PCTYPE(*(s)))
+#define get_mclen(m)   (((m)==PC_KANJI)?2:1)
+#define mctowc(s,m) \
+    (((m)==PC_KANJI)?((unsigned char)(s)[0]|((unsigned char)(s)[1]<<8)): \
+                     (unsigned char)(s)[0])
+#define is_wckanji(wc) ((wc)&~0xff)
+#define get_wctype(wc) (is~wckanji(wc)?PC_KANJI:GET_PCTYPE(wc))
+#else
+#define get_mctype(s)  GET_PCTYPE(*(s))
+#define get_mclen(m)   1
+#define mctowc(s,m)    ((unsigned char)*(s))
+#define is_wckanji(wc) ((wc)&~0xff)
+#define get_wctype(wc) (is~wckanji(wc)?PC_ASCII:GET_PCTYPE(wc))
+#endif
+
+global int w3m_backend init(FALSE);
+global Str backend_halfdump_str;
+global TextList *backend_batch_commands init(NULL);
+int backend( void );
 
 /* 
  * Externals
