@@ -1,16 +1,17 @@
 
-/* $Id: form.c,v 1.8 2000/06/06 05:39:13 aito Exp $ */
+/* $Id: form.c,v 1.2.2.13 2001/01/25 18:54:48 okabe Exp $ */
 /* 
  * HTML forms
  */
 #include "fm.h"
 #include "parsetag.h"
+#include "parsetagx.h"
 #include "myctype.h"
 #include "local.h"
 
 #ifdef __EMX__
-/* lstat is identical to stat, only the link itself is statted, not * the
- * file that is obtained by tracing the links.  But on OS/2 * systems,
+/* lstat is identical to stat, only the link itself is statted, not the
+ * file that is obtained by tracing the links.	But on OS/2 systems,
  * there is no differences. */
 #define lstat stat
 #endif				/* __EMX__ */
@@ -85,10 +86,11 @@ newFormList(char *action, char *method, char *charset, char *enctype, char *targ
  * add <input> element to form_list
  */
 struct form_item_list *
-formList_addInput(struct form_list *fl, struct parsed_tagarg *arg)
+formList_addInput(struct form_list *fl, struct parsed_tag *tag)
 {
-    struct parsed_tagarg *t;
     struct form_item_list *item;
+    char *p;
+    int i;
 
     /* if not in <form>..</form> environment, just ignore <input> tag */
     if (fl == NULL)
@@ -100,45 +102,32 @@ formList_addInput(struct form_list *fl, struct parsed_tagarg *arg)
     item->rows = 0;
     item->checked = 0;
     item->accept = 0;
-    for (t = arg; t != NULL; t = t->next) {
-	if (!strcasecmp(t->arg, "type") && t->value) {
-	    item->type = formtype(t->value);
-	    if (item->size < 0 &&
-		(item->type == FORM_INPUT_TEXT ||
-		 item->type == FORM_INPUT_FILE ||
-		 item->type == FORM_INPUT_PASSWORD))
-		item->size = FORM_I_TEXT_DEFAULT_SIZE;
-	}
-	else if (!strcasecmp(t->arg, "name") && t->value) {
-	    item->name = Strnew_charp(t->value);
-	}
-	else if (!strcasecmp(t->arg, "value") && t->value) {
-	    item->value = Strnew_charp(t->value);
-	}
-	else if (!strcasecmp(t->arg, "checked")) {
-	    item->checked = 1;
-	}
-	else if (!strcasecmp(t->arg, "accept")) {
-	    item->accept = 1;
-	}
-	else if (!strcasecmp(t->arg, "size") && t->value) {
-	    item->size = atoi(t->value);
-	}
-	else if (!strcasecmp(t->arg, "maxlength") && t->value) {
-	    item->maxlength = atoi(t->value);
-	}
-	else if (!strcasecmp(t->arg, "textareanumber") && t->value) {
-	    item->value = textarea_str[atoi(t->value)];
-#ifdef MENU_SELECT
-	}
-	else if (!strcasecmp(t->arg, "selectnumber") && t->value) {
-	    item->select_option = select_option[atoi(t->value)].first;
-#endif				/* MENU_SELECT */
-	}
-	else if (!strcasecmp(t->arg, "rows") && t->value) {
-	    item->rows = atoi(t->value);
-	}
+    item->name = NULL;
+    item->value = NULL;
+    if (parsedtag_get_value(tag, ATTR_TYPE, &p)) {
+	item->type = formtype(p);
+	if (item->size < 0 &&
+	    (item->type == FORM_INPUT_TEXT ||
+	     item->type == FORM_INPUT_FILE ||
+	     item->type == FORM_INPUT_PASSWORD))
+	    item->size = FORM_I_TEXT_DEFAULT_SIZE;
     }
+    if (parsedtag_get_value(tag, ATTR_NAME, &p))
+	item->name = Strnew_charp(p);
+    if (parsedtag_get_value(tag, ATTR_VALUE, &p))
+	item->value = Strnew_charp(p);
+    item->checked = parsedtag_exists(tag, ATTR_CHECKED);
+    item->accept = parsedtag_exists(tag, ATTR_ACCEPT);
+    parsedtag_get_value(tag, ATTR_SIZE, &item->size);
+    parsedtag_get_value(tag, ATTR_MAXLENGTH, &item->maxlength);
+    if (parsedtag_get_value(tag, ATTR_TEXTAREANUMBER, &i))
+	item->value = textarea_str[i];
+#ifdef MENU_SELECT
+    if (parsedtag_get_value(tag, ATTR_SELECTNUMBER, &i))
+	item->select_option = select_option[i].first;
+#endif				/* MENU_SELECT */
+    if (parsedtag_get_value(tag, ATTR_ROWS, &p))
+	item->rows = atoi(p);
     if (item->type == FORM_UNKNOWN) {
 	/* type attribute is missing. Ignore the tag. */
 	return NULL;
@@ -149,6 +138,10 @@ formList_addInput(struct form_list *fl, struct parsed_tagarg *arg)
 	item->label = chooseSelectOption(item->select_option, CHOOSE_OPTION);
     }
 #endif				/* MENU_SELECT */
+    if (item->type == FORM_INPUT_FILE && item->value && item->value->length) {
+	/* security hole ! */
+	return NULL;
+    }
     item->parent = fl;
     item->next = NULL;
     if (fl->item == NULL) {
@@ -256,7 +249,7 @@ formResetBuffer(Buffer * buf, AnchorList * formitem)
 #ifdef MENU_SELECT
 	    f1->select_option = f2->select_option;
 	    f1->label = f2->label;
-#endif
+#endif				/* MENU_SELECT */
 	    break;
 	default:
 	    continue;
@@ -268,10 +261,11 @@ formResetBuffer(Buffer * buf, AnchorList * formitem)
 void
 formUpdateBuffer(Anchor * a, Buffer * buf, FormItemList * form)
 {
-    int i, j, c_type = PC_ASCII;
+    int i, j, k;
     Buffer save;
     char *p;
-    int spos, epos;
+    int spos, epos, c_len;
+    Lineprop c_type;
 
     copyBuffer(&save, buf);
     gotoLine(buf, a->start.line);
@@ -311,46 +305,41 @@ formUpdateBuffer(Anchor * a, Buffer * buf, FormItemList * form)
 	else
 #endif				/* MENU_SELECT */
 	    p = form->value->ptr;
-	j = 0;
-	for (i = spos + 1; i < epos; i++) {
-	    if (p[j] == '\0')
+	i = spos + 1;
+	for (j = 0; p[j];) {
+	    if (p[j] == '\r') {
+		j++;
+		continue;
+	    }
+	    c_type = get_mctype(&p[j]);
+	    c_len = get_mclen(c_type);
+	    k = i + c_len;
+	    if (k > epos)
 		break;
 #ifdef JP_CHARSET
-	    if (c_type == PC_KANJI1)
-		c_type = PC_KANJI2;
-	    else if (p[j] & 0x80)
-		c_type = PC_KANJI1;
-#endif				/* JP_CHARSET */
-	    else if (IS_CNTRL(p[j]))
-		c_type = PC_CTRL;
-	    else
-		c_type = PC_ASCII;
-	    SetCharType(buf->currentLine->propBuf[i], c_type);
-
-	    if (form->type == FORM_INPUT_PASSWORD)
-		buf->currentLine->lineBuf[i] = '*';
-	    else if (c_type == PC_CTRL) {
-		if (p[j] != '\r')
-		    buf->currentLine->lineBuf[i] = ' ';
-		else
-		    i--;
+	    if (c_type == PC_KANJI && form->type != FORM_INPUT_PASSWORD) {
+		SetCharType(buf->currentLine->propBuf[i], PC_KANJI1);
+		SetCharType(buf->currentLine->propBuf[i+1], PC_KANJI2);
 	    }
 	    else
-		buf->currentLine->lineBuf[i] = p[j];
+#endif				/* JP_CHARSET */
+	    SetCharType(buf->currentLine->propBuf[i], PC_ASCII);
 
-	    j++;
+	    for (; i < k; i++, j++) {
+		if (form->type == FORM_INPUT_PASSWORD)
+		    buf->currentLine->lineBuf[i] = '*';
+		else if (c_type == PC_CTRL ||
+			 IS_UNPRINTABLE_ASCII(p[j], c_type))
+		    buf->currentLine->lineBuf[i] = ' ';
+		else
+		    buf->currentLine->lineBuf[i] = p[j];
+	    }
 	}
 	for (; i < epos; i++) {
 	    buf->currentLine->lineBuf[i] = ' ';
 	    SetCharType(buf->currentLine->propBuf[i], PC_ASCII);
 	}
 	i--;
-#ifdef JP_CHARSET
-	if (CharType(buf->currentLine->propBuf[i]) == PC_KANJI1) {
-	    SetCharType(buf->currentLine->propBuf[i], PC_ASCII);
-	    buf->currentLine->lineBuf[i] = ' ';
-	}
-#endif				/* JP_CHARSET */
 	break;
     }
     copyBuffer(buf, &save);
@@ -360,24 +349,22 @@ formUpdateBuffer(Anchor * a, Buffer * buf, FormItemList * form)
 Str
 textfieldrep(Str s, int width)
 {
-    int c_type = PC_ASCII;
+    Lineprop c_type;
     Str n = Strnew_size(width + 2);
-    int i, j;
+    int i, j, k, c_len;
 
-    for (i = 0, j = 0; j < width && i < s->length; i++, j++) {
-#ifdef JP_CHARSET
-	if (c_type == PC_KANJI1)
-	    c_type = PC_KANJI2;
-	else if (s->ptr[i] & 0x80)
-	    c_type = PC_KANJI1;
-	else
-#endif				/* JP_CHARSET */
-	    c_type = PC_ASCII;
+    j = 0;
+    for (i = 0; i < s->length; i += c_len) {
+	c_type = get_mctype(&s->ptr[i]);
+	c_len = get_mclen(c_type);
+	if (s->ptr[i] == '\r') {
+	    continue;
+	}
+	k = j + c_len;
+	if (k > width)
+	    break;
 	if (IS_CNTRL(s->ptr[i])) {
-	    if (s->ptr[i] != '\r')
-		Strcat_char(n, ' ');
-	    else
-		j--;
+	    Strcat_char(n, ' ');
 	}
 	else if (s->ptr[i] == '&')
 	    Strcat_charp(n, "&amp;");
@@ -386,14 +373,9 @@ textfieldrep(Str s, int width)
 	else if (s->ptr[i] == '>')
 	    Strcat_charp(n, "&gt;");
 	else
-	    Strcat_char(n, s->ptr[i]);
+	    Strcat_charp_n(n, &s->ptr[i], c_len);
+	j = k;
     }
-#ifdef JP_CHARSET
-    if (c_type == PC_KANJI1) {
-	Strshrink(n, 1);
-	j--;
-    }
-#endif				/* JP_CHARSET */
     for (; j < width; j++)
 	Strcat_char(n, ' ');
     return n;
@@ -430,7 +412,7 @@ form_fputs_decode(Str s, FILE * f)
 	}
     }
 #ifdef JP_CHARSET
-    fputs(Strkconv(z, InnerCode, DisplayCode)->ptr, f);
+    fputs(conv_str(z, InnerCode, DisplayCode)->ptr, f);
 #else				/* not JP_CHARSET */
     fputs(z->ptr, f);
 #endif				/* not JP_CHARSET */
@@ -440,13 +422,13 @@ form_fputs_decode(Str s, FILE * f)
 void
 input_textarea(FormItemList * fi)
 {
-    Str tmpname = Strnew_charp(rcFile(Sprintf("w3mtmp%d", getpid())->ptr));
+    Str tmpname = tmpfname(TMPF_DFL, NULL);
     Str tmp;
     FILE *f;
 
     f = fopen(tmpname->ptr, "w");
     if (f == NULL) {
-	disp_message("Can't open temporary file", FALSE);
+	disp_err_message("Can't open temporary file", FALSE);
 	return;
     }
     if (fi->value)
@@ -462,7 +444,7 @@ input_textarea(FormItemList * fi)
 
     f = fopen(tmpname->ptr, "r");
     if (f == NULL) {
-	disp_message("Can't open temporary file", FALSE);
+	disp_err_message("Can't open temporary file", FALSE);
 	return;
     }
     fi->value = Strnew();
@@ -477,7 +459,7 @@ input_textarea(FormItemList * fi)
 	    Strcat_charp(tmp, "\r\n");
 	}
 #ifdef JP_CHARSET
-	Strcat(fi->value, Strkconv(tmp, DisplayCode, InnerCode));
+	Strcat(fi->value, conv_str(tmp, DisplayCode, InnerCode));
 #else				/* not JP_CHARSET */
 	Strcat(fi->value, tmp);
 #endif				/* not JP_CHARSET */
@@ -505,11 +487,8 @@ addSelectOption(FormSelectOption * fso, Str value, Str label, int chk)
 {
     FormSelectOptionItem *o;
     o = New(FormSelectOptionItem);
-    if (value == NULL) {
-	if (label->length == 0)
-	    label = Strnew_charp("???");
+    if (value == NULL)
 	value = label;
-    }
     o->value = value;
     Strremovefirstspaces(label);
     Strremovetrailingspaces(label);
@@ -528,6 +507,8 @@ Str
 chooseSelectOption(FormSelectOptionItem * item, int choose_type)
 {
     Str chosen;
+    if (item == NULL)
+	return Strnew_size(0);
     if (choose_type == CHOOSE_OPTION)
 	chosen = item->label;
     else
@@ -547,7 +528,7 @@ chooseSelectOption(FormSelectOptionItem * item, int choose_type)
 void
 formChooseOptionByMenu(struct form_item_list *fi, int x, int y)
 {
-    int i, n, selected, init_select = 0;
+    int i, n, selected = -1, init_select = 0;
     FormSelectOptionItem *opt;
     char **label;
 
