@@ -1,6 +1,6 @@
-/* $Id: frame.c,v 1.16 2000/06/06 05:39:13 aito Exp $ */
+/* $Id: frame.c,v 1.4.2.18 2001/01/22 12:47:07 okabe Exp $ */
 #include "fm.h"
-#include "parsetag.h"
+#include "parsetagx.h"
 #include "myctype.h"
 #include <signal.h>
 #include <setjmp.h>
@@ -12,7 +12,6 @@
 #endif				/* not KANJI_SYMBOLS */
 
 static JMP_BUF AbortLoading;
-static int frameFileSeq;
 struct frameset *renderFrameSet = NULL;
 
 static MySignalHandler
@@ -22,9 +21,8 @@ KeyAbort(SIGNAL_ARG)
 }
 
 struct frameset *
-newFrameSet(struct parsed_tagarg *arg)
+newFrameSet(struct parsed_tag *tag)
 {
-    struct parsed_tagarg *t;
     struct frameset *f;
     int i;
     char *cols = NULL, *rows = NULL, *p, *q;
@@ -34,12 +32,8 @@ newFrameSet(struct parsed_tagarg *arg)
     f->attr = F_FRAMESET;
     f->name = NULL;
     f->currentURL = NULL;
-    for (t = arg; t; t = t->next) {
-	if (!strcasecmp(t->arg, "cols") && t->value)
-	    cols = t->value;
-	else if (!strcasecmp(t->arg, "rows") && t->value)
-	    rows = t->value;
-    }
+    parsedtag_get_value(tag, ATTR_COLS, &cols);
+    parsedtag_get_value(tag, ATTR_ROWS, &rows);
     i = 0;
     if (cols) {
 	length[i] = p = cols;
@@ -111,36 +105,39 @@ newFrameSet(struct parsed_tagarg *arg)
 }
 
 struct frame_body *
-newFrame(struct parsed_tagarg *arg, ParsedURL * baseURL)
+newFrame(struct parsed_tag *tag, ParsedURL * baseURL)
 {
-    struct parsed_tagarg *t;
     struct frame_body *body;
+    char *p;
 
     body = New(struct frame_body);
     bzero((void *) body, sizeof(*body));
     body->attr = F_UNLOADED;
     body->flags = 0;
     body->baseURL = baseURL;
-    for (t = arg; t; t = t->next) {
-	if (!strcasecmp(t->arg, "src") && t->value) {
-	    body->url = t->value;
-	}
-	else if (!strcasecmp(t->arg, "name") && t->value && *t->value != '_')
-	    body->name = t->value;
+    if (tag) {
+	parsedtag_get_value(tag, ATTR_SRC, &body->url);
+	if (parsedtag_get_value(tag, ATTR_NAME, &p) && *p != '_')
+	    body->name = p;
     }
     return body;
 }
 
+static void
+unloadFrame(struct frame_body *b)
+{
+    if (b->source && b->flags & FB_TODELETE)
+	pushText(fileToDelete, b->source);
+    b->attr = F_UNLOADED;
+}
+
 void
 deleteFrame(struct frame_body *b)
 {
     if (b == NULL)
 	return;
-    if (b->source && b->flags & FB_TODELETE)
-	pushText(fileToDelete, b->source);
+    unloadFrame(b);
     bzero((void *) b, sizeof(*b));
-    b->attr = F_UNLOADED;
-    return;
 }
 
 void
@@ -215,11 +212,11 @@ copyFrameSet(struct frameset *of)
     rf->width = New_N(char *, rf->col);
     bcopy((const void *) of->width,
 	  (void *) rf->width,
-	  sizeof(char *) * n);
+	  sizeof(char *) * rf->col);
     rf->height = New_N(char *, rf->row);
     bcopy((const void *) of->height,
 	  (void *) rf->height,
-	  sizeof(char *) * n);
+	  sizeof(char *) * rf->row);
     rf->frame = New_N(union frameset_element, n);
     while (n) {
 	n--;
@@ -344,7 +341,7 @@ resetFrameElement(union frameset_element *f_element,
 	f_element->set->name = f_name;
     }
     else {
-	f_body = newFrame(NULL, buf->baseURL);
+	f_body = newFrame(NULL, baseURL(buf));
 	f_body->attr = F_BODY;
 	f_body->name = f_name;
 	f_body->url = parsedURL2Str(&buf->currentURL)->ptr;
@@ -352,8 +349,7 @@ resetFrameElement(union frameset_element *f_element,
 	    f_body->source = buf->sourcefile;
 	}
 	else {
-	    Str tmp = Sprintf("%s/w3mframe%d.%d",
-			      rc_dir, getpid(), frameFileSeq++);
+	    Str tmp = tmpfname(TMPF_FRAME, NULL);
 	    rename(buf->sourcefile, tmp->ptr);
 	    f_body->source = tmp->ptr;
 	    f_body->flags |= FB_TODELETE;
@@ -367,13 +363,13 @@ resetFrameElement(union frameset_element *f_element,
 }
 
 static struct frameset *
-frame_download_source(struct frame_body *b, ParsedURL * currentURL, ParsedURL * baseURL)
+frame_download_source(struct frame_body *b, ParsedURL * currentURL,
+		      ParsedURL * baseURL, int flag)
 {
     Buffer *buf;
     struct frameset *ret_frameset = NULL;
     Str tmp;
     ParsedURL url;
-    int flag = 0;
 
     if (b == NULL || b->url == NULL || b->url[0] == '\0')
 	return NULL;
@@ -385,11 +381,13 @@ frame_download_source(struct frame_body *b, ParsedURL * currentURL, ParsedURL *
 	b->source = url.file;
 	b->flags = 0;
     default:
+	is_redisplay = TRUE;
 	buf = loadGeneralFile(b->url,
 			      baseURL ? baseURL : currentURL,
 			      b->referer,
 			      flag,
 			      b->request);
+	is_redisplay = FALSE;
 	break;
     }
 
@@ -398,13 +396,9 @@ frame_download_source(struct frame_body *b, ParsedURL * currentURL, ParsedURL *
 	b->flags = (buf == NO_BUFFER) ? FB_NO_BUFFER : 0;
 	return NULL;
     }
-    if (buf->firstLine == NULL) {
-	discardBuffer(buf);
-	return NULL;
-    }
     b->url = parsedURL2Str(&buf->currentURL)->ptr;
     if (buf->real_scheme != SCM_LOCAL) {
-	tmp = Sprintf("%s/w3mframe%d.%d", rc_dir, getpid(), frameFileSeq++);
+	tmp = tmpfname(TMPF_FRAME, NULL);
 	rename(buf->sourcefile, tmp->ptr);
 	b->source = tmp->ptr;
 	b->flags |= FB_TODELETE;
@@ -423,17 +417,17 @@ frame_download_source(struct frame_body *b, ParsedURL * currentURL, ParsedURL *
 }
 
 static int
-createFrameFile(struct frameset *f, FILE * f1, Buffer * current, int level)
+createFrameFile(struct frameset *f, FILE * f1, Buffer * current, int level, int force_reload)
 {
-    int r, c, status, t_stack;
-    FILE *f2;
-    Str tmp, tok;
+    int r, c, t_stack;
+    InputStream f2;
 #ifdef JP_CHARSET
-    char code, ic;
+    char code, ic, charset[2];
 #endif				/* JP_CHARSET */
     char *d_target, *p_target, *s_target, *t_target;
     ParsedURL *currentURL, base;
     MySignalHandler(*prevtrap) (SIGNAL_ARG) = NULL;
+    int flag;
 
     if (f == NULL)
 	return -1;
@@ -459,22 +453,25 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer * current, int level)
 
     if (level == 0) {
 	fprintf(f1, "<html><head><title>%s</title></head><body>\n",
-		current->buffername);
+		htmlquote_str(current->buffername));
 	fputs("<table hborder width=\"100%\">\n", f1);
     }
     else
 	fputs("<table hborder>\n", f1);
 
     currentURL = f->currentURL ? f->currentURL : &current->currentURL;
-    tok = Strnew();
-    status = R_ST_NORMAL;
+#ifdef JP_CHARSET
+    charset[1] = '\0';
+#endif
     for (r = 0; r < f->row; r++) {
-	fputs("<tr>\n", f1);
+	fputs("<tr valign=top>\n", f1);
 	for (c = 0; c < f->col; c++) {
 	    union frameset_element frame;
 	    struct frameset *f_frameset;
 	    int i = c + r * f->col;
 	    char *p = "";
+	    Str tok = Strnew();
+	    int status;
 
 	    frame = f->frame[i];
 
@@ -485,15 +482,21 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer * current, int level)
 
 	    fputs("<td", f1);
 	    if (frame.element->name)
-		fprintf(f1, " id=\"_%s\"", frame.element->name);
+		fprintf(f1, " id=\"_%s\"", htmlquote_str(frame.element->name));
 	    if (!r)
 		fprintf(f1, " width=\"%s\"", f->width[c]);
 	    fputs(">\n", f1);
 
+	    flag = 0;
+	    if (force_reload) {
+		flag |= RG_NOCACHE;
+		if (frame.element->attr == F_BODY)
+		    unloadFrame(frame.body);
+	    }
 	    switch (frame.element->attr) {
 	    default:
 		fprintf(f1, "Frameset \"%s\" frame %d: type unrecognized",
-			f->name, i + 1);
+			htmlquote_str(f->name), i + 1);
 		break;
 	    case F_UNLOADED:
 		if (!frame.body->name && f->name) {
@@ -501,7 +504,8 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer * current, int level)
 		}
 		f_frameset = frame_download_source(frame.body,
 						   currentURL,
-						   current->baseURL);
+						   current->baseURL,
+						   flag);
 		if (f_frameset) {
 		    deleteFrame(frame.body);
 		    f->frame[i].set = frame.set = f_frameset;
@@ -511,7 +515,7 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer * current, int level)
 	    case F_BODY:
 		f2 = NULL;
 		if (frame.body->source == NULL ||
-		    (f2 = fopen(frame.body->source, "r")) == NULL) {
+		    (f2 = openIS(frame.body->source)) == NULL) {
 		    frame.body->attr = F_UNLOADED;
 		    if (frame.body->flags & FB_NO_BUFFER)
 			fprintf(f1, "Open %s with other method", frame.body->url);
@@ -528,45 +532,42 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer * current, int level)
 		code = '\0';
 #endif				/* JP_CHARSET */
 		t_stack = 0;
-		while (1) {
+		do {
+		    status = R_ST_NORMAL;
 		    do {
 			if (*p == '\0') {
-			    tmp = Strmyfgets(f2);
+ 			    Str tmp = StrmyISgets(f2);
+			    if (tmp->length == 0 && status != R_ST_NORMAL)
+				tmp = correct_irrtag(status);
 			    if (tmp->length == 0)
 				break;
 #ifdef JP_CHARSET
-			    if ((ic = checkShiftCode(tmp->ptr, code)) != '\0')
-				tmp = conv(tmp->ptr, (code = ic), InnerCode);
+			    if ((ic = checkShiftCode(tmp, code)) != '\0')
+				tmp = conv_str(tmp, (code = ic), InnerCode);
 
 #endif				/* JP_CHARSET */
+			    cleanup_line(tmp, HTML_MODE);
 			    p = tmp->ptr;
 			}
-			if (status == R_ST_NORMAL &&
-			    !strncmp(p, "<!", 2)) {
+			if (status == R_ST_NORMAL || ST_IS_COMMENT(status))
+			    read_token(tok, &p, &status, 1, 0);
+			else
 			    read_token(tok, &p, &status, 1, 1);
-			    Strclear(tok);
-			    continue;
-			}
-			read_token(tok, &p, &status, 1, 1);
 		    } while (status != R_ST_NORMAL);
 
-		    if (tok->length > 0 && tok->ptr[0] == '<') {
+		    if (tok->length == 0)
+			continue;
+
+		    if (tok->ptr[0] == '<') {
 			char *q = tok->ptr;
-			int cmd, j, is_anchor = 0, a_target = 0;
-			struct parsed_tagarg *targ, *t;
-			Str tagname = Strnew();
+			int j, a_target = 0;
+			struct parsed_tag *tag;
 			ParsedURL url;
 
-			cmd = gethtmlcmd(&q, &j);
-			targ = parse_tag(tok->ptr);
+			if (!(tag = parse_tag(&q, FALSE)))
+			    goto token_end;
 
-			for (j = 1;
-			     j < tok->length && !IS_SPACE(tok->ptr[j]) && tok->ptr[j] != '>';
-			     j++) {
-			    Strcat_char(tagname, tok->ptr[j]);
-			}
-
-			switch (cmd) {
+			switch (tag->tagid) {
 			case HTML_TITLE:
 			    fputs("<!-- title:", f1);
 			    goto token_end;
@@ -574,20 +575,15 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer * current, int level)
 			    fputs("-->", f1);
 			    goto token_end;
 			case HTML_BASE:
-			    for (t = targ; t; t = t->next) {
-				if (!strcasecmp(t->arg, "href") &&
-				    t->value)
-				    parseURL(t->value, &base, NULL);
-				else if (!strcasecmp(t->arg, "target") &&
-					 t->value) {
-				    if (!strcasecmp(t->value, "_self"))
-					d_target = s_target;
-				    else if (!strcasecmp(t->value,
-							 "_parent"))
-					d_target = p_target;
-				    else
-					d_target = t->value;
-				}
+			    if (parsedtag_get_value(tag, ATTR_HREF, &q))
+				parseURL(q, &base, NULL);
+			    if (parsedtag_get_value(tag, ATTR_TARGET, &q)) {
+				if (!strcasecmp(q, "_self"))
+				    d_target = s_target;
+				else if (!strcasecmp(q, "_parent"))
+				    d_target = p_target;
+				else
+				    d_target = q;
 			    }
 			    /* fall thru, "BASE" is prohibit tag */
 			case HTML_HEAD:
@@ -637,94 +633,90 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer * current, int level)
 
 			    }
 			    break;
-			case HTML_A:
-			    is_anchor = 1;
-			    a_target = 2;
-			    for (t = targ; t; t = t->next) {
-				if (a_target == 2 &&
-				    !strcasecmp(t->arg, "href") &&
-				    t->value)
-				    a_target = 1;
-				else if (!strcasecmp(t->arg, "target") &&
-					 t->value) {
-				    a_target = 0;
-				    if (!strcasecmp(t->value, "_self"))
-					t->value = s_target;
-				    else if (!strcasecmp(t->value, "_parent"))
-					t->value = p_target;
-				}
-			    }
-			    break;
-			case HTML_FORM:
-			    a_target = 1;
-			    for (t = targ; t; t = t->next) {
-				if (!strcasecmp(t->arg, "target") &&
-				    t->value) {
-				    a_target = 0;
-				    if (!strcasecmp(t->value, "_self"))
-					t->value = s_target;
-				    else if (!strcasecmp(t->value, "_parent"))
-					t->value = p_target;
-				    break;
-				}
-			    }
-			    break;
 			default:
 			    break;
 			}
-			fputs("<", f1);
-			fputs(tagname->ptr, f1);
-			for (t = targ; t; t = t->next) {
-			    if ((!strcasecmp(t->arg, "src") ||
-				 !strcasecmp(t->arg, "href") ||
-				 !strcasecmp(t->arg, "action")) &&
-				t->value) {
-				if (strncasecmp(t->value, "mailto:", 7) != 0) {
-				    parseURL2(t->value, &url, &base);
-				    t->value = htmlquote_str(parsedURL2Str(&url)->ptr);
+			for (j = 0; j < TagMAP[tag->tagid].max_attribute; j++) {
+			    switch (tag->attrid[j]) {
+			    case ATTR_SRC:
+			    case ATTR_HREF:
+			    case ATTR_ACTION:
+				if (!tag->value[j])
+				    break;
+				parseURL2(tag->value[j], &url, &base);
+				if (url.scheme == SCM_MAILTO ||
+				    url.scheme == SCM_UNKNOWN ||
+				    url.scheme == SCM_MISSING)
+				    break;
+				a_target |= 1;
+				tag->value[j] = parsedURL2Str(&url)->ptr;
+				tag->need_reconstruct = TRUE;
+				parsedtag_set_value(tag,
+						    ATTR_REFERER,
+						    parsedURL2Str(&base)->ptr);
 #ifdef JP_CHARSET
-				    if (ic != '\0')
-					fprintf(f1, " CHARSET=\"%c\"", ic);
-#endif				/* JP_CHARSET */
+				if (code != '\0') {
+				    charset[0] = code;
+				    parsedtag_set_value(tag,
+							ATTR_CHARSET,
+							charset);
 				}
+#endif
+				break;
+			    case ATTR_TARGET:
+				if (!tag->value[j])
+				    break;
+				a_target |= 2;
+				if (!strcasecmp(tag->value[j], "_self")) {
+				    parsedtag_set_value(tag,
+							ATTR_TARGET,
+							s_target);
+				}
+				else if (!strcasecmp(tag->value[j],
+						     "_parent")) {
+				    parsedtag_set_value(tag,
+							ATTR_TARGET,
+							p_target);
+				}
+				break;
+			    case ATTR_NAME:
+			    case ATTR_ID:
+				if (!tag->value[j])
+				    break;
+				parsedtag_set_value(tag,
+						    ATTR_FRAMENAME,
+						    s_target);
+				break;
 			    }
-			    else if ((!strcasecmp(t->arg, "name") ||
-				      !strcasecmp(t->arg, "id")) &&
-				     t->value) {
-				fprintf(f1, " FRAMENAME=\"%s\"", s_target);
-			    }
-			    fprintf(f1, " %s", t->arg);
-			    if (t->value)
-				fprintf(f1, "=\"%s\"", t->value);
 			}
 			if (a_target == 1) {
-			    /* there is HREF attribute and no * TARGET
+			    /* there is HREF attribute and no TARGET
 			     * attribute */
-			    fprintf(f1, " TARGET=\"%s\"", d_target);
+			    parsedtag_set_value(tag,
+						ATTR_TARGET,
+						d_target);
 			}
-			if (is_anchor)
-			    fprintf(f1, " REFERER=\"%s\"",
-			       htmlquote_str(parsedURL2Str(&base)->ptr));
-			fputs(">", f1);
+			if (parsedtag_need_reconstruct(tag))
+			    Strfputs(parsedtag2str(tag), f1);
+			else
+			    Strfputs(tok, f1);
 		    }
 		    else {
-			fputs(tok->ptr, f1);
+			Strfputs(tok, f1);
 		    }
 		  token_end:
-		    if (tmp->length == 0)
-			break;
 		    Strclear(tok);
-		}
+		} while (*p != '\0' || !iseos(f2));
 		while (t_stack--)
 		    fputs("</TABLE>\n", f1);
-		fclose(f2);
+		ISclose(f2);
 		break;
 	    case F_FRAMESET:
 	      render_frameset:
 		if (!frame.set->name && f->name) {
 		    frame.set->name = Sprintf("%s_%d", f->name, i)->ptr;
 		}
-		createFrameFile(frame.set, f1, current, level + 1);
+		createFrameFile(frame.set, f1, current, level + 1, force_reload);
 		break;
 	    }
 	    fputs("</td>\n", f1);
@@ -743,7 +735,7 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer * current, int level)
 }
 
 Buffer *
-renderFrame(Buffer * Cbuf)
+renderFrame(Buffer * Cbuf, int force_reload)
 {
     Str tmp;
     FILE *f;
@@ -751,18 +743,17 @@ renderFrame(Buffer * Cbuf)
     int flag;
     struct frameset *fset;
 
-    tmp = Sprintf("%s/w3mframe%d.%lx.html", rc_dir, getpid(),
-		  (unsigned long) Cbuf->frameset);
+    tmp = tmpfname(TMPF_FRAME, ".html");
     f = fopen(tmp->ptr, "w");
     if (f == NULL)
 	return NULL;
     /* 
      * if (Cbuf->frameQ != NULL) fset = Cbuf->frameQ->frameset; else */
     fset = Cbuf->frameset;
-    if (fset == NULL || createFrameFile(fset, f, Cbuf, 0) < 0)
+    if (fset == NULL || createFrameFile(fset, f, Cbuf, 0, force_reload) < 0)
 	return NULL;
     fclose(f);
-    flag = 0;
+    flag = RG_FRAME;
     if ((Cbuf->currentURL).is_nocache)
 	flag |= RG_NOCACHE;
     renderFrameSet = Cbuf->frameset;
@@ -772,7 +763,6 @@ renderFrame(Buffer * Cbuf)
     if (buf == NULL || buf == NO_BUFFER)
 	return NULL;
     buf->sourcefile = tmp->ptr;
-    buf->bufferprop |= BP_FRAME;
     copyParsedURL(&buf->currentURL, &Cbuf->currentURL);
     return buf;
 }
