@@ -1,10 +1,21 @@
-/* $Id: html.h,v 1.10 2000/06/06 05:39:13 aito Exp $ */
+/* $Id: html.h,v 1.4.2.6 2000/12/26 06:17:30 okabe Exp $ */
 #ifndef _HTML_H
 #define _HTML_H
 #ifdef USE_SSL
+#include <bio.h>
+#include <x509.h>
 #include <ssl.h>
 #endif				/* USE_SSL */
 
+#include "istream.h"
+
+#define StrUFgets(f) StrISgets((f)->stream)
+#define StrmyUFgets(f) StrmyISgets((f)->stream)
+#define UFgetc(f) ISgetc((f)->stream)
+#define UFundogetc(f) ISundogetc((f)->stream)
+#define UFread(f,buf,len) ISread((f)->stream,buf,len)
+#define UFclose(f) (ISclose((f)->stream), (f)->stream = NULL)
+
 struct cmdtable {
     char *cmdname;
     int cmd;
@@ -13,8 +24,19 @@ struct cmdtable {
 struct mailcap {
     char *type;
     char *viewer;
+    int flags;
+    char *test;
+    char *nametemplate;
+    char *edit;
 };
 
+#define MAILCAP_NEEDSTERMINAL	0x01
+#define MAILCAP_COPIOUSOUTPUT	0x02
+#define MAILCAP_HTMLOUTPUT      0x04
+
+#define MCSTAT_REPNAME          0x01
+#define MCSTAT_REPTYPE          0x02
+
 struct table2 {
     char *item1;
     char *item2;
@@ -36,49 +58,21 @@ typedef struct _ParsedURL {
     int is_nocache;
 } ParsedURL;
 
-#ifdef USE_SSL
-#define SSL_BUF_SIZE	1536
-struct ssl_stream {
-    char *buf;
-    int bufsize;
-    int pos, epos;
-    SSL *handle;
-    int sock;
-    int status;
-};
-typedef struct ssl_stream *SSLStream;
-#endif				/* USE_SSL */
-
-typedef union {
-    FILE *f;
-    char *is;
-#ifdef USE_SSL
-    SSLStream ss;
-#endif				/* USE_SSL */
-} Stream;
-
 typedef struct {
     unsigned char scheme;
-    char iseof;
     char is_cgi;
     char encoding;
-    Stream stream;
-    char stream_type;
-    void (*close) ();
+    InputStream stream;
     char *ext;
     int compression;
+    char *guess_type;
 } URLFile;
 
 #define CMP_NOCOMPRESS   0
 #define CMP_COMPRESS     1
 #define CMP_GZIP         2
 #define CMP_BZIP2        3
-
-#define SMT_FILE         0
-#define SMT_STRING       1
-#ifdef USE_SSL
-#define SMT_SSL          2
-#endif				/* USE_SSL */
+#define CMP_DEFLATE      4
 
 #define ENC_7BIT	0
 #define ENC_BASE64	1
@@ -185,7 +179,7 @@ typedef struct {
 
 
    /* pseudo tag */
-#define HTML_DUMMY_TABLE    113
+#define HTML_TABLE_ALT      113
 #define HTML_RULE           114
 #define HTML_N_RULE         115
 #define HTML_PRE_INT        116
@@ -201,6 +195,105 @@ typedef struct {
 #define HTML_EOL            126
 #define HTML_NOP	    127
 
+#define MAX_HTMLTAG	    128
+
+/* Tag attribute */
+
+#define ATTR_UNKNOWN		0
+#define ATTR_ACCEPT		1
+#define ATTR_ACCEPT_CHARSET	2
+#define ATTR_ACTION		3
+#define ATTR_ALIGN		4
+#define ATTR_ALT		5
+#define ATTR_ARCHIVE		6
+#define ATTR_BACKGROUND		7
+#define ATTR_BORDER		8
+#define ATTR_CELLPADDING	9
+#define ATTR_CELLSPACING	10
+#define ATTR_CHARSET		11
+#define ATTR_CHECKED		12
+#define ATTR_COLS		13
+#define ATTR_COLSPAN		14
+#define ATTR_CONTENT		15
+#define ATTR_ENCTYPE		16
+#define ATTR_HEIGHT		17
+#define ATTR_HREF		18
+#define ATTR_HTTP_EQUIV		19
+#define ATTR_ID			20
+#define ATTR_LINK		21
+#define ATTR_MAXLENGTH		22
+#define ATTR_METHOD		23
+#define ATTR_MULTIPLE		24
+#define ATTR_NAME		25
+#define ATTR_NOWRAP		26
+#define ATTR_PROMPT		27
+#define ATTR_ROWS		28
+#define ATTR_ROWSPAN		29
+#define ATTR_SIZE		30
+#define ATTR_SRC		31
+#define ATTR_TARGET		32
+#define ATTR_TYPE		33
+#define ATTR_USEMAP		34
+#define ATTR_VALIGN		35
+#define ATTR_VALUE		36
+#define ATTR_VSPACE		37
+#define ATTR_WIDTH		38
+#define ATTR_COMPACT		39
+#define ATTR_START		40
+#define ATTR_SELECTED		41
+#define ATTR_LABEL		42
+
+/* Internal attribute */
+#define ATTR_TID		53
+#define ATTR_FID		54
+#define ATTR_FOR_TABLE		55
+#define ATTR_FRAMENAME		56
+#define ATTR_HBORDER		57
+#define ATTR_HSEQ		58
+#define ATTR_NO_EFFECT		59
+#define ATTR_REFERER		60
+#define ATTR_SELECTNUMBER	61
+#define ATTR_TEXTAREANUMBER	62
+#define ATTR_TITLE		63
+
+#define MAX_TAGATTR		64
+
+/* HTML Tag Information Table */
+
+typedef struct html_tag_info {
+    unsigned char 	*accept_attribute;
+    unsigned char 	max_attribute;
+    unsigned char 	flag;
+} TagInfo;
+
+#define TFLG_END	1
+#define TFLG_INT	2
+
+/* HTML Tag Attribute Information Table */
+
+typedef struct tag_attribute_info {
+    char 		*name;
+    unsigned char 	vtype;
+    unsigned char 	flag;
+} TagAttrInfo;
+
+#define AFLG_INT	1
+
+#define VTYPE_NONE	0
+#define VTYPE_STR	1
+#define VTYPE_NUMBER	2
+#define VTYPE_LENGTH	3
+#define VTYPE_ALIGN     4
+#define VTYPE_VALIGN    5
+#define VTYPE_ACTION    6
+#define VTYPE_ENCTYPE   7
+#define VTYPE_METHOD    8
+#define VTYPE_MLENGTH   9
+#define VTYPE_TYPE      10
+
+extern TagInfo TagMAP[];
+extern TagAttrInfo AttrMAP[];
+
 struct environment {
     char env;
     int type;
