@@ -8,6 +8,7 @@
 #include "menu.h"
 #include "func.h"
 #include "myctype.h"
+#include "regex.h"
 
 #ifdef MOUSE
 #ifdef USE_GPM
@@ -91,119 +92,188 @@ static int mCancel(char c);
 static int mClose(char c);
 static int mSusp(char c);
 static int mMouse(char c);
+static int mSrchF(char c);
+static int mSrchB(char c);
+static int mSrchN(char c);
+static int mSrchP(char c);
+#ifdef __EMX__
+static int	mPc(char c);
+#endif
 
 static int (*MenuKeymap[128]) (char c) = {
-/* C-@    C-a     C-b     C-c     C-d     C-e     C-f     C-g      */
-    mNull, mTop, mPrev, mClose, mNull, mLast, mNext, mNull,
-/* C-h    C-i     C-j     C-k     C-l     C-m     C-n     C-o      */
-	mCancel, mNull, mOk, mNull, mNull, mOk, mDown, mNull,
-/* C-p    C-q     C-r     C-s     C-t     C-u     C-v     C-w      */
-	mUp, mNull, mNull, mNull, mNull, mNull, mNext, mNull,
-/* C-x    C-y     C-z     C-[     C-\     C-]     C-^     C-_      */
-	mNull, mNull, mSusp, mEsc, mNull, mNull, mNull, mNull,
-/* SPC    !       "       #       $       %       &       '        */
-	mOk, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-/* (      )       *       +       ,       -       .       /        */
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-/* 0      1       2       3       4       5       6       7        */
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-/* 8      9       :       ;       <       =       >       ?        */
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-/* @      A       B       C       D       E       F       G        */
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-/* H      I       J       K       L       M       N       O        */
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-/* P      Q       R       S       T       U       V       W        */
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-/* X      Y       Z       [       \       ]       ^       _        */
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-/* `      a       b       c       d       e       f       g        */
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-/* h      i       j       k       l       m       n       o        */
-	mNull, mNull, mDown, mUp, mNull, mNull, mNull, mNull,
-/* p      q       r       s       t       u       v       w        */
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-/* x      y       z       {       |       }       ~       DEL      */
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mCancel,
+/*  C-@     C-a     C-b     C-c     C-d     C-e     C-f     C-g      */
+#ifdef __EMX__
+    mPc,    mTop,   mPrev,  mClose, mNull,  mLast,  mNext,  mNull,
+#else
+    mNull,  mTop,   mPrev,  mClose, mNull,  mLast,  mNext,  mNull,
+#endif
+/*  C-h     C-i     C-j     C-k     C-l     C-m     C-n     C-o      */
+    mCancel,mNull,  mOk,    mNull,  mNull,  mOk,    mDown,  mNull,
+/*  C-p     C-q     C-r     C-s     C-t     C-u     C-v     C-w      */
+    mUp,    mNull,  mNull,  mNull,  mNull,  mNull,  mNext,  mNull,
+/*  C-x     C-y     C-z     C-[     C-\     C-]     C-^     C-_      */
+    mNull,  mNull,  mSusp,  mEsc,   mNull,  mNull,  mNull,  mNull,
+/*  SPC     !       "       #       $       %       &       '        */
+    mOk,    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+/*  (       )       *       +       ,       -       .       /        */
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mSrchF,
+/*  0       1       2       3       4       5       6       7        */
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull , mNull,  mNull,
+/*  8       9       :       ;       <       =       >       ?        */
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mSrchB,
+/*  @       A       B       C       D       E       F       G        */
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+/*  H       I       J       K       L       M       N       O        */
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mSrchP, mNull,
+/*  P       Q       R       S       T       U       V       W        */
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+/*  X       Y       Z       [       \       ]       ^       _        */
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+/*  `       a       b       c       d       e       f       g        */
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+/*  h       i       j       k       l       m       n       o        */
+    mCancel,mNull,  mDown,  mUp,    mOk,    mNull,  mSrchN, mNull,
+/*  p       q       r       s       t       u       v       w        */
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+/*  x       y       z       {       |       }       ~       DEL      */
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mCancel,
 };
 static int (*MenuEscKeymap[128]) (char c) = {
-    mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
 
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
 
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-/* O     */
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mEscB,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-/* [                                     */
-	mNull, mNull, mNull, mEscB, mNull, mNull, mNull, mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+/*  O     */
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mEscB,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+/*  [                                     */
+    mNull,  mNull,  mNull,  mEscB,  mNull,  mNull,  mNull,  mNull,
 
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-/* v             */
-	mNull, mNull, mNull, mNull, mNull, mNull, mPrev, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+/*  v             */
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mPrev,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
 };
 static int (*MenuEscBKeymap[128]) (char c) = {
-    mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
 
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-/* A       B       C       D       E                     */
-	mNull, mUp, mDown, mOk, mCancel, mClose, mNull, mNull,
-/* L       M                     */
-	mNull, mNull, mNull, mNull, mClose, mMouse, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+/*  A       B       C       D       E                     */
+    mNull,  mUp,    mDown,  mOk,    mCancel,mClose, mNull, mNull,
+/*  L       M                     */
+    mNull,  mNull,  mNull,  mNull,  mClose, mMouse, mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
 
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
 };
 static int (*MenuEscDKeymap[128]) (char c) = {
-/* 0      1       INS     3       4       PgUp,   PgDn    7     */
-    mNull, mNull, mClose, mNull, mNull, mBack, mFore, mNull,
-/* 8      9       10      F1      F2      F3      F4      F5       */
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-/* 16     F6      F7      F8      F9      F10     22      23       */
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-/* 24     25      26      27      HELP    29      30      31       */
-	mNull, mNull, mNull, mNull, mClose, mNull, mNull, mNull,
+/*  0       1       INS     3       4       PgUp,   PgDn    7     */
+    mNull,  mNull,  mClose, mNull,  mNull,  mBack,  mFore,  mNull,
+/*  8       9       10      F1      F2      F3      F4      F5       */
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+/*  16      F6      F7      F8      F9      F10     22      23       */
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+/*  24      25      26      27      HELP    29      30      31       */
+    mNull,  mNull,  mNull,  mNull,  mClose, mNull,  mNull,  mNull,
 
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
 
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
 
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
-	mNull, mNull, mNull, mNull, mNull, mNull, mNull, mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
 };
 
+#ifdef __EMX__
+static int (*MenuPcKeymap[256])(char c)={
+//			  Null
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+//							  S-Tab
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+// A-q	  A-w	  A-E	  A-r	  A-t	  A-y	  A-u	  A-i
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+// A-o	  A-p	  A-[	  A-]			  A-a	  A-s
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+// A-d	  A-f	  A-g	  A-h	  A-j	  A-k	  A-l	  A-;
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+// A-'    A-'		  A-\		  A-x	  A-c	  A-v
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mPrev,
+// A-b	  A-n	  A-m	  A-,	  A-.	  A-/		  A-+
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+//			  F1	  F2	  F3	  F4	  F5
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+// F6	  F7	  F8	  F9	  F10			  Home
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mTop,
+// Up	  PgUp	  A-/	  Left	  5	  Right	  C-*	  End
+  mUp,	  mUp,	  mNull,  mCancel,mNull,  mOk,	  mNull,  mLast,
+// Down	  PgDn	  Ins	  Del	  S-F1	  S-F2	  S-F3	  S-F4
+  mDown,  mDown,  mClose, mCancel,mNull,  mNull,  mNull,  mNull,
+// S-F5	  S-F6	  S-F7	  S-F8	  S-F9	  S-F10	  C-F1	  C-F2
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+// C-F3	  C-F4	  C-F5	  C-F6	  C-F7	  C-F8	  C-F9	  C-F10
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+// A-F1	  A-F2	  A-F3	  A-F4	  A-F5	  A-F6	  A-F7	  A-F8
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+// A-F9	  A-F10	  PrtSc	  C-Left  C-Right C-End	  C-PgDn  C-Home
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+// A-1	  A-2	  A-3	  A-4	  A-5	  A-6	  A-7/8	  A-9
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+// A-0	  A -	  A-=		  C-PgUp  F11	  F12	  S-F11
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+// S-F12  C-F11	  C-F12	  A-F11	  A-F12	  C-Up	  C-/	  C-5
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+// S-*	  C-Down  C-Ins	  C-Del	  C-Tab	  C -	  C-+
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
+//				  A -	  A-Tab	  A-Enter
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,   // 160
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,   // 168
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,   // 176
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,   // 184
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,   // 192
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,   // 200
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,   // 208
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,   // 216
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,   // 224
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,   // 232
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,   // 240
+  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull	   // 248
+};
+#endif
+
 /* --- SelectMenu --- */
 
 static Menu SelectMenu;
 static int SelectV = 0;
 static void initSelectMenu(void);
 static void smChBuf(void);
+static int smDelBuf(char c);
 
 /* --- SelectMenu (END) --- */
 
@@ -213,50 +283,50 @@ static Menu MainMenu;
 #if LANG == JA
 static MenuItem MainMenuItem[] =
 {
-/* type        label         variabel value func     popup keys  */
-    {MENU_FUNC, "         (b)", NULL, 0, backBf, NULL, "b"},
-    {MENU_FUNC, "Хåե (s)", NULL, 0, selBuf, NULL, "s"},
-    {MENU_FUNC, "ɽ (v)", NULL, 0, vwSrc, NULL, "vV"},
-    {MENU_FUNC, "Խ (e)", NULL, 0, editBf, NULL, "eE"},
-    {MENU_FUNC, "¸ (S)", NULL, 0, svSrc, NULL, "S"},
-    {MENU_FUNC, "ɤ߹   (r)", NULL, 0, reload, NULL, "rR"},
-    {MENU_NOP, "", NULL, 0, nulcmd, NULL, ""},
-    {MENU_FUNC, "󥯤ɽ (a)", NULL, 0, followA, NULL, "a"},
-    {MENU_FUNC, "󥯤¸ (A)", NULL, 0, svA, NULL, "A"},
-    {MENU_FUNC, "ɽ   (i)", NULL, 0, followI, NULL, "i"},
-    {MENU_FUNC, "¸   (I)", NULL, 0, svI, NULL, "I"},
-    {MENU_FUNC, "ե졼ɽ (f)", NULL, 0, rFrame, NULL, "fF"},
-    {MENU_NOP, "", NULL, 0, nulcmd, NULL, ""},
-    {MENU_FUNC, "֥åޡ (B)", NULL, 0, ldBmark, NULL, "B"},
-    {MENU_FUNC, "إ       (h)", NULL, 0, ldhelp, NULL, "hH"},
-    {MENU_FUNC, "ץ   (o)", NULL, 0, ldOpt, NULL, "oO"},
-    {MENU_NOP, "", NULL, 0, nulcmd, NULL, ""},
-    {MENU_FUNC, "λ         (q)", NULL, 0, qquitfm, NULL, "qQ"},
-    {MENU_END, "", NULL, 0, nulcmd, NULL, ""},
+/* type        label         variabel value func     popup keys data  */
+    {MENU_FUNC, "         (b)", NULL, 0, backBf, NULL, "b", NULL},
+    {MENU_FUNC, "Хåե (s)", NULL, 0, selBuf, NULL, "s", NULL},
+    {MENU_FUNC, "ɽ (v)", NULL, 0, vwSrc, NULL, "vV", NULL},
+    {MENU_FUNC, "Խ (e)", NULL, 0, editBf, NULL, "eE", NULL},
+    {MENU_FUNC, "¸ (S)", NULL, 0, svSrc, NULL, "S", NULL},
+    {MENU_FUNC, "ɤ߹   (r)", NULL, 0, reload, NULL, "rR", NULL},
+    {MENU_NOP, "", NULL, 0, nulcmd, NULL, "", NULL},
+    {MENU_FUNC, "󥯤ɽ (a)", NULL, 0, followA, NULL, "a", NULL},
+    {MENU_FUNC, "󥯤¸ (A)", NULL, 0, svA, NULL, "A", NULL},
+    {MENU_FUNC, "ɽ   (i)", NULL, 0, followI, NULL, "i", NULL},
+    {MENU_FUNC, "¸   (I)", NULL, 0, svI, NULL, "I", NULL},
+    {MENU_FUNC, "ե졼ɽ (f)", NULL, 0, rFrame, NULL, "fF", NULL},
+    {MENU_NOP, "", NULL, 0, nulcmd, NULL, "", NULL},
+    {MENU_FUNC, "֥åޡ (B)", NULL, 0, ldBmark, NULL, "B", NULL},
+    {MENU_FUNC, "إ       (h)", NULL, 0, ldhelp, NULL, "hH", NULL},
+    {MENU_FUNC, "ץ   (o)", NULL, 0, ldOpt, NULL, "oO", NULL},
+    {MENU_NOP, "", NULL, 0, nulcmd, NULL, "", NULL},
+    {MENU_FUNC, "λ         (q)", NULL, 0, qquitfm, NULL, "qQ", NULL},
+    {MENU_END, "", NULL, 0, nulcmd, NULL, "", NULL},
 };
 #else				/* LANG != JA */
 static MenuItem MainMenuItem[] =
 {
-/* type        label           variable value func     popup keys  */
-    {MENU_FUNC, " Back         (b) ", NULL, 0, backBf, NULL, "b"},
-    {MENU_FUNC, " Select Buffer(s) ", NULL, 0, selBuf, NULL, "s"},
-    {MENU_FUNC, " View Source  (v) ", NULL, 0, vwSrc, NULL, "vV"},
-    {MENU_FUNC, " Edit Source  (e) ", NULL, 0, editBf, NULL, "eE"},
-    {MENU_FUNC, " Save Source  (S) ", NULL, 0, svSrc, NULL, "S"},
-    {MENU_FUNC, " Reload       (r) ", NULL, 0, reload, NULL, "rR"},
-    {MENU_NOP, " ---------------- ", NULL, 0, nulcmd, NULL, ""},
-    {MENU_FUNC, " Go Link      (a) ", NULL, 0, followA, NULL, "a"},
-    {MENU_FUNC, " Save Link    (A) ", NULL, 0, svA, NULL, "A"},
-    {MENU_FUNC, " View Image   (i) ", NULL, 0, followI, NULL, "i"},
-    {MENU_FUNC, " Save Image   (I) ", NULL, 0, svI, NULL, "I"},
-    {MENU_FUNC, " View Frame   (f) ", NULL, 0, rFrame, NULL, "fF"},
-    {MENU_NOP, " ---------------- ", NULL, 0, nulcmd, NULL, ""},
-    {MENU_FUNC, " Bookmark     (B) ", NULL, 0, ldBmark, NULL, "B"},
-    {MENU_FUNC, " Help         (h) ", NULL, 0, ldhelp, NULL, "hH"},
-    {MENU_FUNC, " Option       (o) ", NULL, 0, ldOpt, NULL, "oO"},
-    {MENU_NOP, " ---------------- ", NULL, 0, nulcmd, NULL, ""},
-    {MENU_FUNC, " Quit         (q) ", NULL, 0, qquitfm, NULL, "qQ"},
-    {MENU_END, "", NULL, 0, nulcmd, NULL, ""},
+/* type        label           variable value func     popup keys data  */
+    {MENU_FUNC, " Back         (b) ", NULL, 0, backBf, NULL, "b", NULL},
+    {MENU_FUNC, " Select Buffer(s) ", NULL, 0, selBuf, NULL, "s", NULL},
+    {MENU_FUNC, " View Source  (v) ", NULL, 0, vwSrc, NULL, "vV", NULL},
+    {MENU_FUNC, " Edit Source  (e) ", NULL, 0, editBf, NULL, "eE", NULL},
+    {MENU_FUNC, " Save Source  (S) ", NULL, 0, svSrc, NULL, "S", NULL},
+    {MENU_FUNC, " Reload       (r) ", NULL, 0, reload, NULL, "rR", NULL},
+    {MENU_NOP, " ---------------- ", NULL, 0, nulcmd, NULL, "", NULL},
+    {MENU_FUNC, " Go Link      (a) ", NULL, 0, followA, NULL, "a", NULL},
+    {MENU_FUNC, " Save Link    (A) ", NULL, 0, svA, NULL, "A", NULL},
+    {MENU_FUNC, " View Image   (i) ", NULL, 0, followI, NULL, "i", NULL},
+    {MENU_FUNC, " Save Image   (I) ", NULL, 0, svI, NULL, "I", NULL},
+    {MENU_FUNC, " View Frame   (f) ", NULL, 0, rFrame, NULL, "fF", NULL},
+    {MENU_NOP, " ---------------- ", NULL, 0, nulcmd, NULL, "", NULL},
+    {MENU_FUNC, " Bookmark     (B) ", NULL, 0, ldBmark, NULL, "B", NULL},
+    {MENU_FUNC, " Help         (h) ", NULL, 0, ldhelp, NULL, "hH", NULL},
+    {MENU_FUNC, " Option       (o) ", NULL, 0, ldOpt, NULL, "oO", NULL},
+    {MENU_NOP, " ---------------- ", NULL, 0, nulcmd, NULL, "", NULL},
+    {MENU_FUNC, " Quit         (q) ", NULL, 0, qquitfm, NULL, "qQ", NULL},
+    {MENU_END, "", NULL, 0, nulcmd, NULL, "", NULL},
 };
 #endif				/* LANG != JA  */
 
@@ -301,7 +371,7 @@ new_menu(Menu * menu, MenuItem * item)
     for (i = 0; i < menu->nitem; i++) {
 	if ((p = item[i].keys) != NULL) {
 	    while (*p) {
-		if ((*p & 0x80) == 0) {
+		if (IS_ASCII(*p)) {
 		    menu->keymap[(int) *p] = mSelect;
 		    menu->keyselect[(int) *p] = i;
 		}
@@ -473,18 +543,28 @@ select_menu(Menu * menu, int select)
 void
 goto_menu(Menu * menu, int select, int down)
 {
+    int select_in;
     if (select >= menu->nitem)
 	select = menu->nitem - 1;
     else if (select < 0)
 	select = 0;
+    select_in = select;
     while (menu->item[select].type == MENU_NOP) {
 	if (down > 0) {
 	    if (++select >= menu->nitem)
-		return;
+	    {
+		down_menu(menu, select_in - menu->select);
+		select = menu->select;
+		break;
+	    }
 	}
 	else if (down < 0) {
 	    if (--select < 0)
-		return;
+	    {
+		up_menu(menu, menu->select - select_in);
+		select = menu->select;
+		break;
+	    }
 	}
 	else {
 	    return;
@@ -538,16 +618,6 @@ action_menu(Menu * menu)
 	    mouse_active();
 #endif				/* MOUSE */
 	c = getch();
-#ifdef __EMX__
-	if (!c)
-	    switch (getch()) {
-	    case 72:
-		c = 'k';
-		break;		/* down arrow */
-	    case 80:
-		c = 'j';	/* up arrow   */
-	    }
-#endif				/* __EMX__ */
 #ifdef MOUSE
 	if (use_mouse)
 	    mouse_inactive();
@@ -560,7 +630,7 @@ action_menu(Menu * menu)
 #endif				/* defined(USE_GPM) || * * * * * *
 				 * defined(USE_SYSMOUSE) */
 #endif				/* MOUSE */
-	if ((c & 0x80) == 0) {	/* Ascii */
+	if (IS_ASCII(c)) {	/* Ascii */
 	    select = (*menu->keymap[(int) c]) (c);
 	    if (select != MENU_NOTHING)
 		break;
@@ -576,8 +646,13 @@ action_menu(Menu * menu)
 	    menu->parent->active = 0;
 	if (item.type & MENU_VALUE)
 	    *item.variable = item.value;
-	if (item.type & MENU_FUNC)
+	if (item.type & MENU_FUNC) {
+	    CurrentKey = -1;
+	    CurrentKeyData = NULL;
+	    CurrentMenuData = item.data;
 	    (*item.func) ();
+	    CurrentMenuData = NULL;
+	}
     }
     else if (select == MENU_CLOSE) {
 	if (menu->parent != NULL)
@@ -681,6 +756,15 @@ new_option_menu(Menu * menu, char **label, int *variable, void (*func) ())
 
 /* --- MenuFunctions --- */
 
+#ifdef __EMX__
+static int
+mPc(char c)
+{
+  c = getch();
+  return(MenuPcKeymap[(int)c](c));
+}
+#endif
+
 static int
 mEsc(char c)
 {
@@ -724,7 +808,7 @@ mNull(char c)
 static int
 mSelect(char c)
 {
-    if ((c & 0x80) == 0)
+    if (IS_ASCII(c))
 	return (select_menu(CurrentMenu, CurrentMenu->keyselect[(int) c]));
     else
 	return (MENU_NOTHING);
@@ -835,18 +919,167 @@ mSusp(char c)
     return (MENU_NOTHING);
 }
 
+static char *SearchString = NULL;
+
+int (*menuSearchRoutine) (Menu *, char *, int);
+
+static int
+menuForwardSearch (Menu* menu, char* str, int from)
+{
+    int i;
+    char* p;
+    if ((p = regexCompile (str, IgnoreCase)) != NULL) {
+	message (p, 0, 0);
+	return -1;
+    }
+    for (i = from; i < menu->nitem; i++)
+	if (regexMatch (menu->item[i].label, 0, 0) == 1)
+            return i;
+    return -1;
+}
+
+int
+menu_search_forward (Menu* menu, int from)
+{
+    char *str;
+    int found;
+    str = inputStrHist("Forward: ", NULL, TextHist);
+    if (str != NULL && *str == '\0')
+	str = SearchString;
+    if (str == NULL || *str == '\0')
+	return (MENU_NOTHING);
+    SearchString = str;
+    found = menuForwardSearch (menu, SearchString, from);
+    if (WrapSearch && found == -1)
+        found = menuForwardSearch (menu, SearchString, 0);
+    menuSearchRoutine = menuForwardSearch;
+    if (found >= 0)
+        return found;
+    return from;
+}
+
+static int
+mSrchF (char c)
+{
+    int select;
+    select = menu_search_forward (CurrentMenu, CurrentMenu->select);
+    goto_menu (CurrentMenu, select, 1);
+    return (MENU_NOTHING);
+}
+
+static int
+menuBackwardSearch (Menu* menu, char* str, int from)
+{
+    int i;
+    char* p;
+    if ((p = regexCompile (str, IgnoreCase)) != NULL) {
+	message (p, 0, 0);
+	return -1;
+    }
+    for (i = from; i >= 0 ; i--)
+	if (regexMatch (menu->item[i].label, 0, 0) == 1)
+            return i;
+    return -1;
+}
+
+int
+menu_search_backward (Menu* menu, int from)
+{
+    char *str;
+    int found;
+    str = inputStrHist("Backward: ", NULL, TextHist);
+    if (str != NULL && *str == '\0')
+	str = SearchString;
+    if (str == NULL || *str == '\0')
+	return (MENU_NOTHING);
+    SearchString = str;
+    found = menuBackwardSearch (menu, SearchString, from);
+    if (WrapSearch && found == -1)
+        found = menuBackwardSearch (menu, SearchString, 0);
+    menuSearchRoutine = menuBackwardSearch;
+    if (found >= 0)
+        return found;
+    return from;
+}
+
+static int
+mSrchB (char c)
+{
+    int select;
+    select = menu_search_backward (CurrentMenu, CurrentMenu->select);
+    goto_menu (CurrentMenu, select, -1);
+    return (MENU_NOTHING);
+}
+
+static int
+menu_search_next_previous (Menu* menu, int from, int reverse)
+{
+    int found;
+    int new_from;
+    static int (*routine[2]) (Menu *, char *, int) =
+    {
+	menuForwardSearch, menuBackwardSearch
+    };
+
+    if (menuSearchRoutine == NULL) {
+	disp_message ("No previous regular expression", TRUE);
+	return from;
+    }
+    addstr(menuSearchRoutine == menuForwardSearch ? "Forward: " : "Backward: ");
+    addstr(SearchString);
+    if (reverse != 0)
+	reverse = 1;
+    if (menuSearchRoutine == menuBackwardSearch)
+	reverse ^= 1;
+    new_from = from - reverse * 2 + 1;
+    if (new_from >=0 && new_from < menu->nitem)
+	found = (*routine[reverse]) (menu, SearchString, new_from);
+    else
+	found = (*routine[reverse]) (menu, SearchString, from);
+    if (WrapSearch && found == -1) {
+        found = (*routine[reverse]) (menu, SearchString, reverse * menu->nitem);
+    }
+    if (found >= 0)
+        return found;
+    return from;
+}
+
+static int
+mSrchN (char c)
+{
+    int select;
+    select = menu_search_next_previous (CurrentMenu, CurrentMenu->select, 0);
+    goto_menu (CurrentMenu, select, 1);
+    return (MENU_NOTHING);
+}
+
+static int
+mSrchP (char c)
+{
+    int select;
+    select = menu_search_next_previous (CurrentMenu, CurrentMenu->select, 1);
+    goto_menu (CurrentMenu, select, -1);
+    return (MENU_NOTHING);
+}
+
 #ifdef MOUSE
 #define MOUSE_BTN1_DOWN 0
 #define MOUSE_BTN2_DOWN 1
 #define MOUSE_BTN3_DOWN 2
+#define MOUSE_BTN4_DOWN_RXVT 3
+#define MOUSE_BTN5_DOWN_RXVT 4
+#define MOUSE_BTN4_DOWN_XTERM 64
+#define MOUSE_BTN5_DOWN_XTERM 65
 #define MOUSE_BTN_UP 3
+#define MOUSE_BTN_RESET -1
+#define MOUSE_SCROLL_LINE 5
 
 static int
 process_mMouse(int btn, int x, int y)
 {
     Menu *menu;
     int select;
-    static int press_btn, press_x, press_y;
+    static int press_btn = MOUSE_BTN_RESET, press_x, press_y;
     char c = ' ';
 
     menu = CurrentMenu;
@@ -976,22 +1209,39 @@ mMouse(char c)
 /* --- MainMenu --- */
 
 void
-mainMenu(int x, int y)
+popupMenu(int x, int y, Menu *menu)
 {
-    MainMenu.cursorX = Currentbuf->cursorX;
-    MainMenu.cursorY = Currentbuf->cursorY;
-    MainMenu.x = x + FRAME_WIDTH + 1;
-    MainMenu.y = y + 2;
-
     initSelectMenu();
 
-    popup_menu(NULL, &MainMenu);
+    menu->cursorX = Currentbuf->cursorX;
+    menu->cursorY = Currentbuf->cursorY;
+    menu->x = x + FRAME_WIDTH + 1;
+    menu->y = y + 2;
+
+    popup_menu(NULL, menu);
+}
+
+void
+mainMenu(int x, int y)
+{
+    popupMenu(x, y, &MainMenu);
 }
 
 void
 mainMn(void)
 {
-    mainMenu(Currentbuf->cursorX, Currentbuf->cursorY);
+    Menu *menu = &MainMenu;
+    char *data;
+    int n;
+
+    data = searchKeyData();
+    if (data != NULL) {
+	n = getMenuN(w3mMenuList, data);
+	if (n < 0)
+	    return;
+	menu = w3mMenuList[n].menu;
+    }
+    popupMenu(Currentbuf->cursorX, Currentbuf->cursorY, menu);
 }
 
 /* --- MainMenu (END) --- */
@@ -1001,10 +1251,11 @@ mainMn(void)
 static void
 initSelectMenu(void)
 {
-    int i, nitem;
+    int i, nitem, len = 0, l;
     Buffer *buf;
     Str str;
     char **label;
+    static char *comment = " SPC for select / D for delete buffer ";
 
     SelectV = -1;
     for (i = 0, buf = Firstbuf; buf != NULL; i++, buf = buf->nextBuffer) {
@@ -1013,9 +1264,9 @@ initSelectMenu(void)
     }
     nitem = i;
 
-    label = New_N(char *, nitem + 1);
+    label = New_N(char *, nitem + 2);
     for (i = 0, buf = Firstbuf; i < nitem; i++, buf = buf->nextBuffer) {
-	str = Sprintf("<%s>", cleanup_str(buf->buffername));
+	str = Sprintf("<%s>", buf->buffername);
 	if (buf->filename != NULL) {
 	    switch (buf->currentURL.scheme) {
 	    case SCM_LOCAL:
@@ -1035,12 +1286,30 @@ initSelectMenu(void)
 	    }
 	}
 	label[i] = str->ptr;
+	if (len < str->length)
+	    len = str->length;
     }
+    l = strlen(comment);
+    if (len < l + 4)
+	len = l + 4;
+    if (len > COLS - 2 * FRAME_WIDTH)
+	len = COLS - 2 * FRAME_WIDTH;
+    len = (len > 1) ? ((len - l + 1) / 2) : 0;
+    str = Strnew();
+    for (i = 0; i < len; i++)
+	Strcat_char(str, '-');
+    Strcat_charp(str, comment);
+    for (i = 0; i < len; i++)
+	Strcat_char(str, '-');
+    label[nitem] = str->ptr;
+    label[nitem + 1] = NULL;
 
     new_option_menu(&SelectMenu, label, &SelectV, smChBuf);
     SelectMenu.initial = SelectV;
     SelectMenu.cursorX = Currentbuf->cursorX;
     SelectMenu.cursorY = Currentbuf->cursorY;
+    SelectMenu.keymap['D'] = smDelBuf;
+    SelectMenu.item[nitem].type = MENU_NOP;
 }
 
 static void
@@ -1052,8 +1321,49 @@ smChBuf(void)
     if (SelectV < 0 || SelectV >= SelectMenu.nitem)
 	return;
     for (i = 0, buf = Firstbuf; i < SelectV; i++, buf = buf->nextBuffer);
-
     Currentbuf = buf;
+    if (clear_buffer) {
+	for (buf = Firstbuf; buf != NULL; buf = buf->nextBuffer)
+	    tmpClearBuffer(buf);
+    }
+}
+
+static int
+smDelBuf(char c) {
+    int i, x, y, select;
+    Buffer *buf;
+
+    if (CurrentMenu->select < 0 || CurrentMenu->select >= SelectMenu.nitem)
+	return (MENU_NOTHING);
+    for (i = 0, buf = Firstbuf; i < CurrentMenu->select; i++, buf = buf->nextBuffer);
+    if (Currentbuf == buf)
+	Currentbuf = buf->nextBuffer;
+    Firstbuf = deleteBuffer(Firstbuf, buf);
+    if (!Currentbuf)
+	Currentbuf = nthBuffer(Firstbuf, i - 1);;
+    if (Firstbuf == NULL) {
+	Firstbuf = nullBuffer();
+	Currentbuf = Firstbuf;
+    }
+
+    x = CurrentMenu->x;
+    y = CurrentMenu->y;
+    select = CurrentMenu->select;
+
+    initSelectMenu();
+
+    CurrentMenu->x = x;
+    CurrentMenu->y = y;
+
+    geom_menu(CurrentMenu, x, y, 0);
+
+    CurrentMenu->select = (select <= CurrentMenu->nitem - 2) ? select
+	: (CurrentMenu->nitem - 2);
+
+    displayBuffer(Currentbuf, B_FORCE_REDRAW);
+    draw_all_menu(CurrentMenu);
+    select_menu(CurrentMenu, CurrentMenu->select);
+    return (MENU_NOTHING);
 }
 
 /* --- SelectMenu (END) --- */
@@ -1098,7 +1408,7 @@ initMenu(void)
     w3mMenuList[1].item = NULL;
     w3mMenuList[2].id = NULL;
 
-    if ((mf = fopen(rcFile(MENU_FILE), "r")) == NULL)
+    if ((mf = fopen(rcFile(MENU_FILE), "rt")) == NULL)
 	goto create_menu;
 
     if (!w3mNFuncList)
@@ -1157,7 +1467,7 @@ initMenu(void)
 int
 setMenuItem(MenuItem * item, char *type, char *line)
 {
-    char *label, *func, *popup, *keys;
+    char *label, *func, *popup, *keys, *data;
     int f;
     int n;
 
@@ -1176,6 +1486,7 @@ setMenuItem(MenuItem * item, char *type, char *line)
 	label = getQWord(&line);
 	func = getWord(&line);
 	keys = getQWord(&line);
+	data = getQWord(&line);
 	if (*func == '\0')	/* error */
 	    return -1;
 	item->type = MENU_FUNC;
@@ -1183,11 +1494,12 @@ setMenuItem(MenuItem * item, char *type, char *line)
 	f = getFuncList(func, w3mFuncList, w3mNFuncList);
 	item->func = w3mFuncList[(f >= 0) ? f : FUNCNAME_nulcmd].func;
 	item->keys = keys;
+	item->data = data;
 	return MENU_FUNC;
     }
     else if (strcmp(type, "popup") == 0) {
 	label = getQWord(&line);
-	popup = getWord(&line);
+	popup = getQWord(&line);
 	keys = getQWord(&line);
 	if (*popup == '\0')	/* error */
 	    return -1;
