@@ -8,11 +8,7 @@
 #include <errno.h>
 #include "parsetag.h"
 #include "local.h"
-
-#ifdef __EMX__
-#include "os2emx.h"
-#endif				/* __EMX__ */
-
+#include "terms.h"
 #include <stdlib.h>
 
 struct param_ptr {
@@ -38,11 +34,6 @@ static struct rc_search_table *RC_search_table;
 static int RC_table_size;
 
 static int rc_initialized = 0;
-char *rc_dir;
-
-#ifdef USE_HISTORY
-extern Hist *URLHist;
-#endif				/* USE_HISTORY */
 
 #define P_INT      0
 #define P_SHORT    1
@@ -55,10 +46,15 @@ extern Hist *URLHist;
 #ifdef COLOR
 #define P_COLOR    6
 #endif
+#ifdef JP_CHARSET
+#define P_CODE     7
+#endif
+#define P_PIXELS   8
 
 #if LANG == JA
 #define CMT_HELPER	 "ӥ塼Խ"
 #define CMT_TABSTOP      ""
+#define CMT_PIXEL_PER_CHAR      "ʸ (4.0...32.0)"
 #define CMT_PAGERLINE    "ڡȤѤ¸Կ"
 #define CMT_HISTSIZE     "ݻURLο"
 #define CMT_SAVEHIST     "URL¸"
@@ -75,8 +71,12 @@ extern Hist *URLHist;
 #define CMT_BG_COLOR     "طʤο"
 #define CMT_ACTIVE_STYLE "򤵤Ƥ󥯤οꤹ"
 #define CMT_C_COLOR	 "򤵤Ƥ󥯤ο"
+#define CMT_VISITED_ANCHOR	"ˬ줿Ȥ󥯤ϿѤ"
+#define CMT_V_COLOR	 "ˬ줿Ȥ󥯤ο"
 #define CMT_HTTP_PROXY   "HTTPץ(URL)"
+#ifdef USE_GOPHER
 #define CMT_GOPHER_PROXY "GOPHERץ(URL)"
+#endif				/* USE_GOPHER */
 #define CMT_FTP_PROXY    "FTPץ(URL)"
 #define CMT_NO_PROXY     "ץɥᥤ"
 #define CMT_NOPROXY_NETADDR	"ͥåȥɥ쥹ǥץΥå"
@@ -86,6 +86,8 @@ extern Hist *URLHist;
 #define CMT_CGIBIN       "/cgi-bin ɽǥ쥯ȥ"
 #define CMT_CONFIRM_QQ   "q Ǥνλ˳ǧ"
 #define CMT_SHOW_NUM     "ֹɽ"
+#define CMT_MIMETYPES    "Ѥmime.types"
+#define CMT_MAILCAP      "Ѥmailcap"
 #define CMT_EDITOR       "Ѥ륨ǥ"
 #define CMT_MAILER       "Ѥ᡼"
 #define CMT_EXTBRZ       "֥饦"
@@ -94,11 +96,14 @@ extern Hist *URLHist;
 #define CMT_FTPPASS      "FTPΥѥ(̤ϼʬmail addressȤ)"
 #define CMT_USERAGENT    "User-Agent"
 #define CMT_ACCEPTLANG   "Ĥ(Accept-Language:)"
+#define CMT_DOCUMENTCODE "ʸʸ"
 #define CMT_WRAP         "ޤ֤"
-#define CMT_USECOOKIE "åѤ"
-#define CMT_ACCEPTCOOKIE "åդ"
 #define CMT_VIEW_UNSEENOBJECTS "طʲؤΥ󥯤"
+#ifdef __EMX__
+#define CMT_BGEXTVIEW	 "ӥ塼̥åư"
+#else
 #define CMT_BGEXTVIEW    "ӥ塼Хå饦ɤư"
+#endif
 #define CMT_EXT_DIRLIST  "ǥ쥯ȥꥹȤ˳ޥɤȤ"
 #define CMT_DIRLIST_CMD  "ǥ쥯ȥꥹѥޥ"
 #define CMT_IGNORE_NULL_IMG_ALT "IMG ALT°λ˥̾ɽ"
@@ -110,6 +115,8 @@ extern Hist *URLHist;
 #endif				/* MOUSE */
 #define CMT_CLEAR_BUF     "ɽƤʤХåեΥ"
 #define CMT_NOSENDREFERER "Referer: ʤ褦ˤ"
+#define CMT_IGNORE_CASE "ʸʸζ̤򤷤ʤ"
+#define CMT_USE_LESSOPEN "LESSOPEN"
 #if defined(USE_SSL) && defined(USE_SSL_VERIFY)
 #define CMT_SSL_VERIFY_SERVER "SSLΥǧڤԤ"
 #define CMT_SSL_CERT_FILE "SSLΥ饤PEMե"
@@ -118,6 +125,16 @@ extern Hist *URLHist;
 #define CMT_SSL_CA_FILE "SSLǧڶɤPEM񷲤Υե"
 #endif				/* defined(USE_SSL) &&
 				 * defined(USE_SSL_VERIFY) */
+#ifdef USE_SSL
+#define CMT_SSL_FORBID_METHOD "ȤʤSSL᥽åɤΥꥹ(2: SSLv2, 3: SSLv3, t:TLSv1)"
+#endif
+#ifdef USE_COOKIE
+#define CMT_USECOOKIE "åѤ"
+#define CMT_ACCEPTCOOKIE "åդ"
+#define CMT_ACCEPTBADCOOKIE "Τ륯åǤդ"
+#define CMT_COOKIE_REJECT_DOMAINS "åդʤɥᥤ"
+#define CMT_COOKIE_ACCEPT_DOMAINS "åդɥᥤ"
+#endif
 
 
 #else				/* LANG != JA */
@@ -125,6 +142,7 @@ extern Hist *URLHist;
 
 #define CMT_HELPER	 "External Viewer Setup"
 #define CMT_TABSTOP      "Tab width"
+#define CMT_PIXEL_PER_CHAR      "# of pixels per character (4.0...32.0)"
 #define CMT_PAGERLINE    "# of reserved line when w3m is used as a pager"
 #define CMT_HISTSIZE     "# of reserved URL"
 #define CMT_SAVEHIST     "Save URL history"
@@ -140,9 +158,13 @@ extern Hist *URLHist;
 #define CMT_F_COLOR      "Color of form"
 #define CMT_ACTIVE_STYLE "Use active link color"
 #define CMT_C_COLOR	 "Color of currently active link"
+#define CMT_VISITED_ANCHOR "Use visited link color"
+#define CMT_V_COLOR	 "Color of visited link"
 #define CMT_BG_COLOR     "Color of background"
 #define CMT_HTTP_PROXY   "URL of HTTP proxy host"
+#ifdef USE_GOPHER
 #define CMT_GOPHER_PROXY "URL of GOPHER proxy host"
+#endif				/* USE_GOPHER */
 #define CMT_FTP_PROXY    "URL of FTP proxy host"
 #define CMT_NO_PROXY     "Domains for direct access (no proxy)"
 #define CMT_NOPROXY_NETADDR	"Check noproxy by network address"
@@ -152,6 +174,8 @@ extern Hist *URLHist;
 #define CMT_CGIBIN       "Directory corresponds to /cgi-bin"
 #define CMT_CONFIRM_QQ   "Confirm when quitting with q"
 #define CMT_SHOW_NUM     "Show line number"
+#define CMT_MIMETYPES    "mime.types files"
+#define CMT_MAILCAP      "mailcap files"
 #define CMT_EDITOR       "Editor"
 #define CMT_MAILER       "Mailer"
 #define CMT_EXTBRZ       "External Browser"
@@ -160,11 +184,14 @@ extern Hist *URLHist;
 #define CMT_FTPPASS      "Password for FTP(use your mail address)"
 #define CMT_USERAGENT    "User-Agent"
 #define CMT_ACCEPTLANG   "Accept-Language"
+/* #define CMT_DOCUMENTCODE "Document Charset" */
 #define CMT_WRAP         "Wrap search"
-#define CMT_USECOOKIE   "Use Cookie"
-#define CMT_ACCEPTCOOKIE "Accept Cookie"
 #define CMT_VIEW_UNSEENOBJECTS "Display unseenobjects (e.g. bgimage) tag"
+#ifdef __EMX__
+#define CMT_BGEXTVIEW	 "Another session for an external viewer"
+#else
 #define CMT_BGEXTVIEW    "Background an external viewer"
+#endif
 #define CMT_EXT_DIRLIST  "Use external program for directory listing"
 #define CMT_DIRLIST_CMD  "Directory listing command"
 #define CMT_IGNORE_NULL_IMG_ALT	"Ignore IMG ALT=\"\" (display link name)"
@@ -176,6 +203,8 @@ extern Hist *URLHist;
 #endif				/* MOUSE */
 #define CMT_CLEAR_BUF     "Free memory of the undisplayed buffers"
 #define CMT_NOSENDREFERER "Don't send header `Referer:'"
+#define CMT_IGNORE_CASE "Ignore case when search"
+#define CMT_USE_LESSOPEN "Use LESSOPEN"
 #if defined(USE_SSL) && defined(USE_SSL_VERIFY)
 #define CMT_SSL_VERIFY_SERVER "Perform SSL server verification"
 #define CMT_SSL_CERT_FILE "PEM encoded certificate file of client"
@@ -184,6 +213,16 @@ extern Hist *URLHist;
 #define CMT_SSL_CA_FILE "File consisting of PEM encoded certificates of CAs"
 #endif				/* defined(USE_SSL) &&
 				 * defined(USE_SSL_VERIFY) */
+#ifdef USE_SSL
+#define CMT_SSL_FORBID_METHOD "List of forbidden SSL method (2: SSLv2, 3: SSLv3, t:TLSv1)"
+#endif
+#ifdef USE_COOKIE
+#define CMT_USECOOKIE   "Use Cookie"
+#define CMT_ACCEPTCOOKIE "Accept Cookie"
+#define CMT_ACCEPTBADCOOKIE "Invalid Cookie"
+#define CMT_COOKIE_REJECT_DOMAINS "Domains from which should reject cookies"
+#define CMT_COOKIE_ACCEPT_DOMAINS "Domains from which should accept cookies"
+#endif
 #endif				/* LANG != JA */
 
 #define PI_TEXT    0
@@ -199,12 +238,21 @@ struct sel_c {
 #ifdef JP_CHARSET
 static struct sel_c kcodestr[] =
 {
-    {'E', "E", "EUC-JP"},
-    {'S', "S", "Shift_JIS"},
-    {'j', "j", "JIS X0208-1978 + JIS X0201-1976"},
-    {'N', "N", "JIS X0208-1983 + JIS X0201-1976"},
-    {'m', "m", "JIS X0208-1978 + US_ASCII"},
-    {'n', "n", "JIS X0208-1983 + US_ASCII"},
+    {CODE_EUC,   "E", STR_EUC},
+    {CODE_SJIS,  "S", STR_SJIS},
+    {CODE_JIS_j, "j", STR_JIS_j},
+    {CODE_JIS_N, "N", STR_JIS_N},
+    {CODE_JIS_m, "m", STR_JIS_m},
+    {CODE_JIS_n, "n", STR_JIS_n},
+    {0, NULL, NULL}
+};
+
+static struct sel_c dcodestr[] =
+{
+    {'\0',           "0", "auto detect"},
+    {CODE_EUC,       "E", STR_EUC},
+    {CODE_SJIS,      "S", STR_SJIS},
+    {CODE_INNER_EUC, "I", STR_INNER_EUC},
     {0, NULL, NULL}
 };
 #endif				/* JP_CHARSET */
@@ -212,7 +260,7 @@ static struct sel_c kcodestr[] =
 #ifdef COLOR
 static struct sel_c colorstr[] =
 {
-#if LANG== JA
+#if LANG == JA
     {0, "black", ""},
     {1, "red", ""},
     {2, "green", ""},
@@ -248,11 +296,23 @@ static struct sel_c dnsorders[] =
 };
 #endif				/* INET6 */
 
+#ifdef USE_COOKIE
+static struct sel_c badcookiestr[] = {
+  {0, "0", "discard"},
+#if 0
+  {1, "1", "accept"},
+#endif
+  {2, "2", "ask"},
+  {0, NULL, NULL}
+};
+#endif				/* USE_COOKIE */
+
 struct param_ptr params1[] =
 {
     {"tabstop", P_INT, PI_TEXT, (void *) &Tabstop, CMT_TABSTOP, NULL},
+    {"pixel_per_char", P_PIXELS, PI_TEXT, (void *) &pixel_per_char, CMT_PIXEL_PER_CHAR, NULL},
 #ifdef JP_CHARSET
-    {"kanjicode", P_CHAR, PI_SEL_C, (void *) &DisplayCode, CMT_KANJICODE, kcodestr},
+    {"kanjicode", P_CODE, PI_SEL_C, (void *) &DisplayCode, CMT_KANJICODE, kcodestr},
 #endif				/* JP_CHARSET */
   {"frame", P_CHARINT, PI_ONOFF, (void *) &RenderFrame, CMT_FRAME, NULL},
     {"target_self", P_CHARINT, PI_ONOFF, (void *) &TargetSelf, CMT_TSELF, NULL},
@@ -277,6 +337,8 @@ struct param_ptr params2[] =
 #endif				/* BG_COLOR */
     {"active_style", P_INT, PI_ONOFF, (void *) &useActiveColor, CMT_ACTIVE_STYLE, NULL},
     {"active_color", P_COLOR, PI_SEL_C, (void *) &active_color, CMT_C_COLOR, colorstr},
+    {"visited_anchor", P_INT, PI_ONOFF, (void *) &useVisitedColor, CMT_VISITED_ANCHOR, NULL},
+    {"visited_color", P_COLOR, PI_SEL_C, (void *) &visited_color, CMT_V_COLOR, colorstr},
     {NULL, 0, 0, NULL, NULL, NULL},
 };
 #endif				/* COLOR */
@@ -287,7 +349,7 @@ struct param_ptr params3[] =
   {"pagerline", P_INT, PI_TEXT, (void *) &PagerMax, CMT_PAGERLINE, NULL},
 #ifdef USE_HISTORY
   {"history", P_INT, PI_TEXT, (void *) &URLHistSize, CMT_HISTSIZE, NULL},
-{"save_hist", P_INT, PI_ONOFF, (void *) &URLSaveHist, CMT_SAVEHIST, NULL},
+{"save_hist", P_INT, PI_ONOFF, (void *) &SaveURLHist, CMT_SAVEHIST, NULL},
 #endif				/* USE_HISTORY */
     {"confirm_qq", P_INT, PI_ONOFF, (void *) &confirm_on_quit, CMT_CONFIRM_QQ, NULL},
 {"show_lnum", P_INT, PI_ONOFF, (void *) &showLineNum, CMT_SHOW_NUM, NULL},
@@ -295,29 +357,34 @@ struct param_ptr params3[] =
     {"user_agent", P_STRING, PI_TEXT, (void *) &UserAgent, CMT_USERAGENT, NULL},
     {"no_referer", P_INT, PI_ONOFF, (void *) &NoSendReferer, CMT_NOSENDREFERER, NULL},
     {"accept_language", P_STRING, PI_TEXT, (void *) &AcceptLang, CMT_ACCEPTLANG, NULL},
+#ifdef JP_CHARSET
+    {"document_code", P_CODE, PI_SEL_C, (void *) &DocumentCode, CMT_DOCUMENTCODE, dcodestr},
+#endif
  {"wrap_search", P_INT, PI_ONOFF, (void *) &WrapDefault, CMT_WRAP, NULL},
+ {"ignorecase_search", P_INT, PI_ONOFF, (void *) &IgnoreCase, CMT_IGNORE_CASE, NULL},
 #ifdef VIEW_UNSEENOBJECTS
     {"view_unseenobject", P_INT, PI_ONOFF, (void *) &view_unseenobject, CMT_VIEW_UNSEENOBJECTS, NULL},
 #endif				/* VIEW_UNSEENOBJECTS */
-#ifdef USE_COOKIE
-    {"use_cookie", P_INT, PI_ONOFF, (void *) &use_cookie, CMT_USECOOKIE, NULL},
-    {"accept_cookie", P_INT, PI_ONOFF, (void *) &accept_cookie, CMT_ACCEPTCOOKIE, NULL},
-#endif				/* USE_COOKIE */
 #ifdef MOUSE
     {"use_mouse", P_INT, PI_ONOFF, (void *) &use_mouse, CMT_MOUSE, NULL},
     {"reverse_mouse", P_INT, PI_ONOFF, (void *) &reverse_mouse, CMT_REVERSE_MOUSE, NULL},
 #endif				/* MOUSE */
     {"retry_http", P_INT, PI_ONOFF, (void *) &retryAsHttp, CMT_RETRY_HTTP, NULL},
     {"clear_buffer", P_INT, PI_ONOFF, (void *) &clear_buffer, CMT_CLEAR_BUF, NULL},
+#ifdef USE_SSL
+    {"ssl_forbid_method", P_STRING, PI_TEXT, (void *) &ssl_forbid_method, CMT_SSL_FORBID_METHOD, NULL},
+#endif				/* USE_SSL */
     {NULL, 0, 0, NULL, NULL, NULL},
 };
 
 struct param_ptr params4[] =
 {
     {"http_proxy", P_STRING, PI_TEXT, (void *) &HTTP_proxy, CMT_HTTP_PROXY, NULL},
+#ifdef USE_GOPHER
     {"gopher_proxy", P_STRING, PI_TEXT, (void *) &GOPHER_proxy, CMT_GOPHER_PROXY, NULL},
+#endif				/* USE_GOPHER */
     {"ftp_proxy", P_STRING, PI_TEXT, (void *) &FTP_proxy, CMT_FTP_PROXY, NULL},
- {"no_proxy", P_STRING, PI_TEXT, (void *) &NO_proxy, CMT_NO_PROXY, NULL},
+    {"no_proxy", P_STRING, PI_TEXT, (void *) &NO_proxy, CMT_NO_PROXY, NULL},
     {"noproxy_netaddr", P_INT, PI_ONOFF, (void *) &NOproxy_netaddr, CMT_NOPROXY_NETADDR, NULL},
 #ifdef INET6
     {"dns_order", P_INT, PI_SEL_C, (void *) &DNS_order, CMT_DNS_ORDER, dnsorders},
@@ -336,14 +403,15 @@ struct param_ptr params5[] =
 
 struct param_ptr params6[] =
 {
+    {"mime_types", P_STRING, PI_TEXT, (void *) &mimetypes_files, CMT_MIMETYPES, NULL},
+    {"mailcap", P_STRING, PI_TEXT, (void *) &mailcap_files, CMT_MAILCAP, NULL},
     {"editor", P_STRING, PI_TEXT, (void *) &Editor, CMT_EDITOR, NULL},
     {"mailer", P_STRING, PI_TEXT, (void *) &Mailer, CMT_MAILER, NULL},
 {"extbrowser", P_STRING, PI_TEXT, (void *) &ExtBrowser, CMT_EXTBRZ, NULL},
     {"extbrowser2", P_STRING, PI_TEXT, (void *) &ExtBrowser2, CMT_EXTBRZ2, NULL},
     {"extbrowser3", P_STRING, PI_TEXT, (void *) &ExtBrowser3, CMT_EXTBRZ3, NULL},
-#ifndef __EMX__
     {"bgextviewer", P_INT, PI_ONOFF, (void *) &BackgroundExtViewer, CMT_BGEXTVIEW, NULL},
-#endif				/* __EMX__ */
+    {"use_lessopen", P_INT, PI_ONOFF, (void *) &use_lessopen, CMT_USE_LESSOPEN, NULL},
     {NULL, 0, 0, NULL, NULL, NULL},
 };
 
@@ -360,6 +428,18 @@ struct param_ptr params7[] =
 #endif				/* defined(USE_SSL) &&
 				 * defined(USE_SSL_VERIFY) */
 
+#ifdef USE_COOKIE
+struct param_ptr params8[] =
+{
+    {"use_cookie", P_INT, PI_ONOFF, (void *) &use_cookie, CMT_USECOOKIE, NULL},
+    {"accept_cookie", P_INT, PI_ONOFF, (void *) &accept_cookie, CMT_ACCEPTCOOKIE, NULL},
+    {"accept_bad_cookie", P_INT,  PI_SEL_C, (void *) &accept_bad_cookie, CMT_ACCEPTBADCOOKIE, badcookiestr},
+    {"cookie_reject_domains", P_STRING, PI_TEXT, (void *) &cookie_reject_domains, CMT_COOKIE_REJECT_DOMAINS, NULL},
+    {"cookie_accept_domains", P_STRING, PI_TEXT, (void *) &cookie_accept_domains, CMT_COOKIE_ACCEPT_DOMAINS, NULL},
+    {NULL, 0, 0, NULL, NULL, NULL},
+};
+#endif
+
 struct param_section sections[] =
 {
 #if LANG == JA
@@ -375,6 +455,12 @@ struct param_section sections[] =
     {"SSLǧ", params7},
 #endif				/* defined(USE_SSL) &&
 				 * defined(USE_SSL_VERIFY) */
+#ifdef USE_SSL
+#define CMT_SSL_FORBID_METHOD "ȤʤSSL᥽åɤΥꥹ(2: SSLv2, 3: SSLv3, t:TLSv1)"
+#endif
+#ifdef USE_COOKIE
+    {"å", params8},
+#endif
 #else				/* LANG != JA */
     {"Display", params1},
 #ifdef COLOR
@@ -388,6 +474,12 @@ struct param_section sections[] =
     {"SSL Verification Setting", params7},
 #endif				/* defined(USE_SSL) &&
 				 * defined(USE_SSL_VERIFY) */
+#ifdef USE_SSL
+#define CMT_SSL_FORBID_METHOD N_("List of forbidden SSL method (2: SSLv2, 3: SSLv3, t:TLSv1)")
+#endif
+#ifdef USE_COOKIE
+    {"Cookie Setting", params8},
+#endif
 #endif				/* LANG != JA */
     {NULL, NULL}
 };
@@ -463,10 +555,11 @@ search_param(char *name)
 		return RC_search_table[i].param;
 	    }
 	    else {
-		while (i >= 0 &&
-		(cmp = strcmp(name, RC_search_table[i].param->name)) <= 0)
+		while ((cmp = strcmp(name, RC_search_table[i].param->name)) <= 0)
 		    if (!cmp)
 			return RC_search_table[i].param;
+		    else if (i == 0)
+			return NULL;
 		    else
 			i--;
 		/* ambiguous */
@@ -510,7 +603,7 @@ show_params(FILE * fp)
 		t = (sections[j].params[i].inputtype == PI_ONOFF) ? "bool" : "number";
 		break;
 	    case P_CHAR:
-		t = "E|S|j|N|m|n";	/* only kanjicode */
+		t = "char";
 		break;
 	    case P_STRING:
 		t = "string";
@@ -525,6 +618,14 @@ show_params(FILE * fp)
 		t = "color";
 		break;
 #endif
+#ifdef JP_CHARSET
+	    case P_CODE:
+		t = "E|S|j|N|m|n";
+		break;
+#endif
+	    case P_PIXELS:
+		t = "number";
+		break;
 	    }
 #ifdef JP_CHARSET
 	    if (InnerCode != DisplayCode)
@@ -602,10 +703,69 @@ str_to_color(char *value)
 }
 #endif
 
+#ifdef JP_CHARSET
+char
+str_to_code(char *str)
+{
+    if (str == NULL)
+	return CODE_ASCII;
+    switch (*str) {
+    case CODE_ASCII:
+	return CODE_ASCII;
+    case CODE_EUC:
+    case 'e':
+	return CODE_EUC;
+    case CODE_SJIS:
+    case 's':
+	return CODE_SJIS;
+    case CODE_JIS_n:
+	return CODE_JIS_n;
+    case CODE_JIS_m:
+	return CODE_JIS_m;
+    case CODE_JIS_N:
+	return CODE_JIS_N;
+    case CODE_JIS_j:
+	return CODE_JIS_j;
+    case CODE_JIS_J:
+	return CODE_JIS_J;
+    case CODE_INNER_EUC:
+	return CODE_INNER_EUC;
+    }
+    return CODE_ASCII;
+}
+
+char *
+code_to_str(char code)
+{
+    switch (code) {
+    case CODE_ASCII:
+	return STR_ASCII;
+    case CODE_EUC:
+	return STR_EUC;
+    case CODE_SJIS:
+	return STR_SJIS;
+    case CODE_JIS_n:
+	return STR_JIS_n;
+    case CODE_JIS_m:
+	return STR_JIS_m;
+    case CODE_JIS_N:
+	return STR_JIS_N;
+    case CODE_JIS_j:
+	return STR_JIS_j;
+    case CODE_JIS_J:
+	return STR_JIS_J;
+    case CODE_INNER_EUC:
+	return STR_INNER_EUC;
+    }
+    return "unknown";
+}
+#endif
+
 static int
 set_param(char *name, char *value)
 {
     struct param_ptr *p;
+    double ppc;
 
     if (value == NULL)
 	return 0;
@@ -642,10 +802,23 @@ set_param(char *name, char *value)
 	*(int *) p->varptr = str_to_color(value);
 	break;
 #endif
+#ifdef JP_CHARSET
+    case P_CODE:
+	*(char *) p->varptr = str_to_code(value);
+	break;
+#endif
+    case P_PIXELS:
+	ppc = atof(value);
+	if (ppc >= MINIMUM_PIXEL_PER_CHAR &&
+	    ppc <= MAXIMUM_PIXEL_PER_CHAR)
+	    *(double *) p->varptr = ppc;
+	break;
     }
     return 1;
 }
 
+static void sync_with_option(void);
+
 int
 set_param_option(char *option)
 {
@@ -663,7 +836,7 @@ set_param_option(char *option)
     }
     Strlower(tmp);
     if (set_param(tmp->ptr, p))
-	return 1;
+	goto option_assigned;
     q = tmp->ptr;
     if (!strncmp(q, "no", 2)) {	/* -o noxxx, -o no-xxx, -o no_xxx */
 	q += 2;
@@ -674,7 +847,12 @@ set_param_option(char *option)
 	q++;
     else
 	return 0;
-    return set_param(q, "0");
+    if (set_param(q, "0"))
+	goto option_assigned;
+    return 0;
+ option_assigned:
+    sync_with_option();
+    return 1;
 }
 
 static void
@@ -708,24 +886,42 @@ parse_proxy()
 {
     if (non_null(HTTP_proxy))
 	parseURL(HTTP_proxy, &HTTP_proxy_parsed, NULL);
+#ifdef USE_GOPHER
     if (non_null(GOPHER_proxy))
 	parseURL(GOPHER_proxy, &GOPHER_proxy_parsed, NULL);
+#endif				/* USE_GOPHER */
     if (non_null(FTP_proxy))
 	parseURL(FTP_proxy, &FTP_proxy_parsed, NULL);
     if (non_null(NO_proxy))
 	set_no_proxy(NO_proxy);
 }
 
+#ifdef USE_COOKIE
+void
+parse_cookie()
+{
+    if (non_null(cookie_reject_domains))
+	Cookie_reject_domains = make_domain_list(cookie_reject_domains);
+    if (non_null(cookie_accept_domains))
+	Cookie_accept_domains = make_domain_list(cookie_accept_domains);
+}
+#endif
+
 #ifdef __EMX__
 static int
 do_mkdir(const char *dir, long mode)
 {
     char *r, abs[_MAX_PATH];
+    size_t n;
 
     _abspath(abs, rc_dir, _MAX_PATH);	/* Translate '\\' to '/' */
-    r = strrchr(abs, '/');
-    if (r && r > abs + 2)
+
+    if(!(n=strlen(abs)))
+	return -1;
+
+    if(*(r=abs+n-1)=='/')	/* Ignore tailing slash if it is */
 	*r = 0;
+
     return mkdir(abs, mode);
 }
 #else				/* not __EMX__ */
@@ -765,6 +961,8 @@ loadMimeTypes(char *filename)
 	if (d[0] == '#')
 	    continue;
 	type = strtok(d, " \t\n\r");
+	if (type == NULL)
+	    continue;
 	while (1) {
 	    d = strtok(NULL, " \t\n\r");
 	    if (d == NULL)
@@ -783,16 +981,47 @@ loadMimeTypes(char *filename)
 void
 initMimeTypes()
 {
-    UserMimeTypes = loadMimeTypes(USER_MIMETYPES);
-    SysMimeTypes = loadMimeTypes(SYS_MIMETYPES);
+    int i;
+    TextListItem *tl;
+
+    if (non_null(mimetypes_files))
+	mimetypes_list = make_domain_list(mimetypes_files);
+    else
+	mimetypes_list = NULL;
+    if (mimetypes_list == NULL)
+	return;
+    UserMimeTypes = New_N(struct table2 *, mimetypes_list->nitem);
+    for (i = 0, tl = mimetypes_list->first; tl; i++, tl = tl->next)
+	UserMimeTypes[i] = loadMimeTypes(tl->ptr);
 }
 
+static void
+sync_with_option(void)
+{
+    WrapSearch = WrapDefault;
+    parse_proxy();
+#ifdef USE_COOKIE
+    parse_cookie();
+#endif
+    initMailcap();
+    initMimeTypes();
+}
+    
 void
 init_rc(char *config_file)
 {
     struct stat st;
     FILE *f;
-    char *tmpdir = "/tmp";
+#ifndef __EMX__ /* jsawa */
+    char           *tmpdir = "/tmp";
+#else
+    char	   *tmpdir;
+
+    if ( (tmpdir = getenv("TMP")) == NULL || *tmpdir == '\0' )
+	if ( (tmpdir = getenv("TEMP")) == NULL || *tmpdir == '\0' )
+	    if ( (tmpdir = getenv("TMPDIR")) == NULL || *tmpdir == '\0' )
+		*tmpdir = "/tmp";
+#endif /* __EMX__ */
 
     if (rc_initialized)
 	return;
@@ -803,6 +1032,7 @@ init_rc(char *config_file)
 	    if (do_mkdir(rc_dir, 0700) < 0) {
 		fprintf(stderr, "Can't create config directory (%s)!", rc_dir);
 		rc_dir = tmpdir;
+		rc_dir_is_tmp = TRUE;
 		return;
 	    }
 	    else {
@@ -812,6 +1042,7 @@ init_rc(char *config_file)
 	else {
 	    fprintf(stderr, "Can't open config directory (%s)!", rc_dir);
 	    rc_dir = tmpdir;
+	    rc_dir_is_tmp = TRUE;
 	    return;
 	}
     }
@@ -819,24 +1050,22 @@ init_rc(char *config_file)
 	/* not a directory */
 	fprintf(stderr, "%s is not a directory!", rc_dir);
 	rc_dir = tmpdir;
+	rc_dir_is_tmp = TRUE;
 	return;
     }
 
     /* open config file */
-    if ((f = fopen(libFile(W3MCONFIG), "r")) != NULL) {
+    if ((f = fopen(libFile(W3MCONFIG), "rt")) != NULL) {
 	interpret_rc(f);
 	fclose(f);
     }
     if (config_file == NULL)
 	config_file = rcFile(CONFIG_FILE);
-    if ((f = fopen(config_file, "r")) != NULL) {
+    if ((f = fopen(config_file, "rt")) != NULL) {
 	interpret_rc(f);
 	fclose(f);
     }
-    parse_proxy();
-    WrapSearch = WrapDefault;
-    initMailcap();
-    initMimeTypes();
+    sync_with_option();
 }
 
 
@@ -864,12 +1093,17 @@ to_str(struct param_ptr *p)
     case P_CHARINT:
 	return Sprintf("%d", *(char *) p->varptr);
     case P_CHAR:
+#ifdef JP_CHARSET
+    case P_CODE:
+#endif
 	return Sprintf("%c", *(char *) p->varptr);
     case P_STRING:
 #if defined(USE_SSL) && defined(USE_SSL_VERIFY)
     case P_SSLPATH:
 #endif
 	return Strnew_charp(*(char **) p->varptr);
+    case P_PIXELS:
+	return Sprintf("%g", *(double *) p->varptr);
     }
     /* not reached */
     return NULL;
@@ -884,22 +1118,23 @@ load_option_panel(void)
     int x, i;
     Str tmp;
 
+    Strcat_charp(src, "<table><tr><td>");
     for (i = 0; sections[i].name != NULL; i++) {
 	Strcat_m_charp(src, "<h1>",
 		       sections[i].name,
 		       "</h1>", NULL);
 	p = sections[i].params;
-	Strcat_charp(src, "<table width=100%>");
+	Strcat_charp(src, "<table width=100% cellpadding=0>");
 	while (p->name) {
-	    Strcat_m_charp(src, "<tr><td>",
-			   p->comment,
-			   "</td><td>", NULL);
+	    Strcat_m_charp(src, "<tr><td>", p->comment, NULL);
+	    Strcat(src, Sprintf("</td><td width=%d>",
+		(int)(28 * pixel_per_char)));
 	    switch (p->inputtype) {
 	    case PI_TEXT:
 		Strcat_m_charp(src, "<input type=text name=",
 			       p->name,
 			       " value=\"",
-			       to_str(p)->ptr,
+			       htmlquote_str( to_str(p)->ptr ),
 			       "\">", NULL);
 		break;
 	    case PI_ONOFF:
@@ -922,9 +1157,16 @@ load_option_panel(void)
 		for (s = p->select; s->text != NULL; s++) {
 		    Strcat_charp(src, "<option value=");
 		    Strcat(src, Sprintf("%s\n", s->cvalue));
-		    if ((p->type != P_CHAR && s->value == atoi(tmp->ptr)) ||
-			(p->type == P_CHAR &&
-			 (char) (s->value) == *(tmp->ptr)))
+		    if ((p->type != P_CHAR &&
+#ifdef JP_CHARSET
+			 p->type != P_CODE &&
+#endif
+			 s->value == atoi(tmp->ptr)) ||
+			((p->type == P_CHAR
+#ifdef JP_CHARSET
+			  || p->type == P_CODE
+#endif
+			    ) && (char) (s->value) == *(tmp->ptr)))
 			Strcat_charp(src, " selected");
 		    Strcat_char(src, '>');
 		    Strcat_charp(src, s->text);
@@ -937,44 +1179,53 @@ load_option_panel(void)
 	Strcat_charp(src, "<tr><td></td><td><p><input type=submit value=\"OK\"></td></tr>");
 	Strcat_charp(src, "</table><hr width=50%>");
     }
-    Strcat_charp(src, "</form></body></html>");
-    return loadHTMLString(src->ptr);
+    Strcat_charp(src, "</table></form></body></html>");
+    return loadHTMLString(src);
 }
 
 void
 panel_set_option(struct parsed_tagarg *arg)
 {
-    FILE *f;
+    FILE *f = NULL;
 
-    f = fopen(config_file, "w");
-    if (f == NULL) {
-	disp_message("Can't write option!", FALSE);
-	backBf();
+    if (rc_dir_is_tmp) {
+	disp_message("There's no ~/.w3m directory... config not saved", FALSE);
+    }
+    else {
+	f = fopen(config_file, "wt");
+	if (f == NULL) {
+	    disp_message("Can't write option!", FALSE);
+	}
     }
     while (arg) {
 	if (set_param(arg->arg, arg->value)) {
-	    fprintf(f, "%s %s\n", arg->arg, arg->value);
+	    if (f)
+		fprintf(f, "%s %s\n", arg->arg, arg->value);
 	}
 	arg = arg->next;
     }
-    fclose(f);
-    parse_proxy();
-#ifdef USE_HISTORY
-    resizeHist(URLHist, URLHistSize + 1);
-#endif				/* USE_HISTORY */
-    WrapSearch = WrapDefault;
+    if (f)
+	fclose(f);
+    sync_with_option();
     backBf();
 }
 
 char *
 rcFile(char *base)
 {
-    Str file = Strnew_charp(rc_dir);
+    if (base &&
+	(base[0] == '/' ||
+	 (base[0] == '.' && (base[1] == '/' || (base[1] == '.' && base[2] == '/'))) ||
+	 (base[0] == '~' && base[1] == '/')))
+	return expandName(base);
+    else {
+	Str file = Strnew_charp(rc_dir);
 
-    if (Strlastchar(file) != '/')
-	Strcat_char(file, '/');
-    Strcat_charp(file, base);
-    return expandName(file->ptr);
+	if (Strlastchar(file) != '/')
+	    Strcat_char(file, '/');
+	Strcat_charp(file, base);
+	return expandName(file->ptr);
+    }
 }
 
 char *
