@@ -1,4 +1,4 @@
-/* $Id: terms.c,v 1.19 2000/06/06 05:39:13 aito Exp $ */
+/* $Id: terms.c,v 1.2.4.3 2000/12/02 11:33:18 okabe Exp $ */
 /* 
  * An original curses library for EUC-kanji by Akinori ITO,     December 1989
  * revised by Akinori ITO, January 1995
@@ -32,14 +32,20 @@ int mouseActive = 0;
 #include <sys/select.h>
 #endif				/* AIX */
 
-#ifdef __EMX__
-#include <sys/select.h>
-#endif				/* __EMX__ */
-
 #include "terms.h"
 #include "fm.h"
 #include "myctype.h"
 
+#ifdef __EMX__
+#define INCL_DOSNLS
+#include <os2.h>
+#include <sys/select.h>
+
+#ifndef JP_CHARSET
+extern int	CodePage;
+#endif				/* !JP_CHARSET */
+#endif				/* __EMX__ */
+
 char *getenv(const char *);
 MySignalHandler reset_exit(SIGNAL_ARG), error_dump(SIGNAL_ARG);
 void setlinescols(void);
@@ -194,7 +200,7 @@ void clear(), wrap(), touch_line(), touch_column(int);
 #ifdef JP_CHARSET
 void switch_wchar(FILE *);
 void switch_ascii(FILE *);
-#endif				/* JP_CHARSET */
+#endif
 void need_clrtoeol(void), clrtoeol(void);
 
 int write1(char);
@@ -298,10 +304,19 @@ set_cc(int spec, int val)
 }
 #endif				/* not SGTTY */
 
+void
+close_tty(void)
+{
+    close(tty);
+}
+
 void
 reset_tty(void)
 {
-    if (DisplayCode != 'E' && DisplayCode != 'S')
+    if (DisplayCode != CODE_EUC && DisplayCode != CODE_SJIS)
+#if defined(__EMX__)&&!defined(JP_CHARSET)
+      if(!CodePage)
+#endif
 	writestr("\033(B");	/* designate US_ASCII */
     writestr(T_op);		/* turn off */
     writestr(T_me);
@@ -472,6 +487,14 @@ setlinescols(void)
 		pclose(fd);
 	    }
 	}
+#ifndef JP_CHARSET
+	else{
+	    ULONG CpList[8],CpSize;
+	    CodePage=-1;
+	    if(!DosQueryCp(sizeof(CpList),CpList,&CpSize))
+		CodePage=*CpList;
+	}
+#endif
     }
 #elif defined(TERMIOS) && defined(TIOCGWINSZ)
     struct winsize wins;
@@ -560,7 +583,7 @@ write1(char c)
     usleep(50);
 #endif				/* SCREEN_DEBUG */
 #ifdef JP_CHARSET
-    if (c & 0x80) {
+    if (IS_KANJI(c)) {
 	switch (wmode) {
 	case C_ASCII:
 	    switch_wchar(ttyf);
@@ -630,35 +653,34 @@ switch_wchar(FILE * f)
 void
 putchars(unsigned char c1, unsigned char c2, FILE * f)
 {
+    Str s;
+    char *p;
+
     switch (DisplayCode) {
     case CODE_EUC:
 	putc(c1, f);
 	putc(c2, f);
-	break;
-    case CODE_JIS1:
-    case CODE_JIS2:
-    case CODE_JIS3:
-    case CODE_JIS4:
+	return;
+    case CODE_JIS_n:
+    case CODE_JIS_m:
+    case CODE_JIS_N:
+    case CODE_JIS_j:
+    case CODE_JIS_J:
 	putc(c1 & 0x7f, f);
 	putc(c2 & 0x7f, f);
-	break;
+	return;
     case CODE_SJIS:
-	c1 &= 0x7f;
-	c2 &= 0x7f;
-	c1 -= 0x20;
-	c2 -= 0x20;
-	if ((c1 & 1) == 0)
-	    c2 += 94;
-	c1 = ((c1 - 1) >> 1) + 0x81;
-	c2 += 0x3f;
-	if (c1 > 0x9f)
-	    c1 += 0x40;
-	if (c2 > 0x7e)
-	    c2++;
-	putc(c1, f);
-	putc(c2, f);
+	s = Strnew_size(3);
+	put_sjis(s, c1 & 0x7f, c2 & 0x7f);
 	break;
     }
+    if (! s->length) {
+	putc('?', f);
+	putc('?', f);
+	return;
+    }
+    for (p = s->ptr; *p != '\0'; p++)
+	putc(*p, f);
 }
 #endif				/* JP_CHARSET */
 
@@ -700,6 +722,9 @@ addch(char c)
     l_prop *pr;
     int dest, i;
     short *dirty;
+#ifdef __EMX__
+    extern int CodePage;
+#endif
 
     if (CurColumn == COLS)
 	wrap();
@@ -709,6 +734,11 @@ addch(char c)
     pr = ScreenImage[CurLine]->lineprop;
     dirty = &ScreenImage[CurLine]->isdirty;
 
+    /* Eliminate unprintables according to * iso-8859-*.
+     * Particularly 0x96 messes up T.Dickey's * (xfree-)xterm */
+    if (IS_INTERNAL(c))
+	c = ' ';
+
     if (pr[CurColumn] & S_EOL) {
 	if (c == ' ' &&
 #ifdef JP_CHARSET
@@ -729,10 +759,10 @@ addch(char c)
 #ifdef JP_CHARSET
     else if (CHMODE(CurrentMode) == C_WCHAR1)
 	SETCHMODE(CurrentMode, C_WCHAR2);
-    else if (c & 0x80)
+    else if (IS_KANJI1(c))
 	SETCHMODE(CurrentMode, C_WCHAR1);
 #endif				/* JP_CHARSET */
-    else if (c & ~0x1f)
+    else if (!IS_CNTRL(c))
 	SETCHMODE(CurrentMode, C_ASCII);
     else
 	return;
@@ -1377,7 +1407,7 @@ addnstr(char *s, int n)
     for (i = 0; i < n - 1 && *s != '\0'; i++)
 	addch(*(s++));
     if (*s != '\0') {
-	if (*s & 0x80) {
+	if (IS_KANJI2(*s)) {
 	    /* WCHAR */
 	    if (CHMODE(CurrentMode) == C_WCHAR1) {
 		/* WCHAR second byte */
@@ -1404,7 +1434,7 @@ addnstr_sup(char *s, int n)
     for (i = 0; i < n - 1 && *s != '\0'; i++)
 	addch(*(s++));
     if (*s != '\0') {
-	if (*s & 0x80) {
+	if (IS_KANJI2(*s)) {
 	    /* WCHAR */
 	    if (CHMODE(CurrentMode) == C_WCHAR1) {
 		/* WCHAR second byte */
@@ -1499,7 +1529,14 @@ void
 term_cooked(void)
 #ifndef SGTTY
 {
+#ifdef __EMX__
+    /* On XFree86/OS2, some scrambled characters
+     * will appear when asserting IEXTEN flag.
+     */
+    ttymode_set((TTY_MODE)&~IEXTEN,0);
+#else
     ttymode_set(TTY_MODE, 0);
+#endif
 #ifdef TERMIOS
     set_cc(VMIN, 4);
 #else				/* not TERMIOS */
