@@ -20,6 +20,13 @@
 #include "myctype.h"
 #include "regex.h"
 
+#ifdef USE_SSL
+#ifndef SSLEAY_VERSION_NUMBER
+#include <crypto.h>		/* SSLEAY_VERSION_NUMBER may be here */
+#endif
+#include <err.h>
+#endif
+
 #ifdef	__WATT32__
 #define	write(a,b,c)	write_s(a,b,c)
 #endif				/* __WATT32__ */
@@ -107,7 +114,7 @@ static struct table2 DefaultGuess[] =
     {NULL, NULL}
 };
 
-static FILE *add_index_file(ParsedURL * pu, void (**close_rout) ());
+static void add_index_file(ParsedURL * pu, URLFile *uf);
 
 /* #define HTTP_DEFAULT_FILE    "/index.html" */
 
@@ -142,8 +149,10 @@ DefaultFile(int scheme)
     case SCM_HTTPS:
 #endif				/* USE_SSL */
 	return allocStr(HTTP_DEFAULT_FILE, 0);
+#ifdef USE_GOPHER
     case SCM_GOPHER:
 	return allocStr("1", 0);
+#endif				/* USE_GOPHER */
     case SCM_LOCAL:
     case SCM_LOCAL_CGI:
     case SCM_FTP:
@@ -168,11 +177,61 @@ free_ssl_ctx()
 	SSL_CTX_free(ssl_ctx);
 }
 
+#if SSLEAY_VERSION_NUMBER >= 0x00905100
+#include <rand.h>
+void
+init_PRNG()
+{
+    char buffer[256];
+    const char *file;
+    long l;
+    if (RAND_status())
+	return;
+    if (file = RAND_file_name(buffer, sizeof(buffer))) {
+#ifdef USE_EGD
+	if (RAND_egd(file) > 0)
+	    return;
+#endif
+	RAND_load_file(file, -1);
+    }
+    if (RAND_status())
+	goto seeded;
+    srand48((long) time(NULL));
+    while (!RAND_status()) {
+	l = lrand48();
+	RAND_seed((unsigned char *)&l, sizeof(long));
+    }
+ seeded:
+    if (file)
+	RAND_write_file(file);
+}
+#endif				/* SSLEAY_VERSION_NUMBER >= 0x00905100 */
+
 SSL *
 openSSLHandle(int sock)
 {
     SSL *handle;
+    char *emsg;
+    static char *old_ssl_forbid_method = NULL;
 #ifdef USE_SSL_VERIFY
+    static int old_ssl_verify_server = -1;
+#endif
+
+    if (! old_ssl_forbid_method || ! ssl_forbid_method ||
+	strcmp(old_ssl_forbid_method, ssl_forbid_method)) {
+	old_ssl_forbid_method = ssl_forbid_method;
+#ifdef USE_SSL_VERIFY
+	ssl_path_modified = 1;
+#else
+	free_ssl_ctx();
+	ssl_ctx = NULL;
+#endif
+    }
+#ifdef USE_SSL_VERIFY
+    if (old_ssl_verify_server != ssl_verify_server) {
+	old_ssl_verify_server = ssl_verify_server;
+	ssl_path_modified = 1;
+    }
     if (ssl_path_modified) {
 	free_ssl_ctx();
 	ssl_ctx = NULL;
@@ -180,13 +239,23 @@ openSSLHandle(int sock)
     }
 #endif				/* defined(USE_SSL_VERIFY) */
     if (ssl_ctx == NULL) {
+	int option;
 #if SSLEAY_VERSION_NUMBER < 0x0800
 	ssl_ctx = SSL_CTX_new();
 	X509_set_default_verify_paths(ssl_ctx->cert);
 #else				/* SSLEAY_VERSION_NUMBER >= 0x0800 */
 	SSLeay_add_ssl_algorithms();
-	ssl_ctx = SSL_CTX_new(SSLv23_client_method());
-	SSL_CTX_set_options(ssl_ctx, SSL_OP_ALL);
+	SSL_load_error_strings();
+	if (!(ssl_ctx = SSL_CTX_new(SSLv23_client_method())))
+	  goto eend;
+	option = SSL_OP_ALL;
+	if (ssl_forbid_method) {
+	  if (strchr(ssl_forbid_method, '2')) option |= SSL_OP_NO_SSLv2;
+	  if (strchr(ssl_forbid_method, '3')) option |= SSL_OP_NO_SSLv3;
+	  if (strchr(ssl_forbid_method, 't')) option |= SSL_OP_NO_TLSv1;
+	  if (strchr(ssl_forbid_method, 'T')) option |= SSL_OP_NO_TLSv1;
+	}
+	SSL_CTX_set_options(ssl_ctx, option);
 #ifdef USE_SSL_VERIFY
 	/* derived from openssl-0.9.5/apps/s_{client,cb}.c */
 	SSL_CTX_set_verify(ssl_ctx, ssl_verify_server ? SSL_VERIFY_PEER : SSL_VERIFY_NONE, NULL);
@@ -200,10 +269,10 @@ openSSLHandle(int sock)
 	    }
 	    if (ng) {
 		free_ssl_ctx();
-		return NULL;
+		goto eend;
 	    }
 	}
-	SSL_CTX_load_verify_locations(ssl_ctx, ssl_ca_file, ssl_ca_path) &&
+	if (SSL_CTX_load_verify_locations(ssl_ctx, ssl_ca_file, ssl_ca_path))
 #endif				/* defined(USE_SSL_VERIFY) */
 	    SSL_CTX_set_default_verify_paths(ssl_ctx);
 #endif				/* SSLEAY_VERSION_NUMBER >= 0x0800 */
@@ -211,9 +280,16 @@ openSSLHandle(int sock)
     }
     handle = SSL_new(ssl_ctx);
     SSL_set_fd(handle, sock);
+#if SSLEAY_VERSION_NUMBER >= 0x00905100
+    init_PRNG();
+#endif				/* SSLEAY_VERSION_NUMBER >= 0x00905100 */
     if (SSL_connect(handle) <= 0)
-	return NULL;
+      goto eend;
     return handle;
+eend:
+    emsg = Sprintf("SSL error: %s", ERR_error_string(ERR_get_error(), NULL))->ptr;
+    disp_err_message(emsg, FALSE);
+    return NULL;
 }
 
 static void
@@ -270,7 +346,7 @@ openSocket(char *hostname,
 	   char *remoteport_name,
 	   unsigned short remoteport_num)
 {
-    int sock;
+    int sock = -1;
 #ifdef INET6
     int *af;
     struct addrinfo hints, *res0, *res;
@@ -285,10 +361,16 @@ openSocket(char *hostname,
 #endif				/* not INET6 */
     MySignalHandler(*trap) ();
 
+    if (fmInitialized) {
+        message(Sprintf("Opening socket...\n")->ptr, 0, 0);
+        refresh();
+    }
     if (SETJMP(AbortLoading) != 0) {
 #ifdef SOCK_DEBUG
 	sock_log("openSocket() failed. reason: user abort\n");
 #endif
+	if (sock >= 0)
+	    close(sock);
 	goto error;
     }
     trap = signal(SIGINT, KeyAbort);
@@ -363,12 +445,16 @@ openSocket(char *hostname,
 	goto error;
     }
     regexCompile("[0-9][0-9]*\\.[0-9][0-9]*\\.[0-9][0-9]*\\.[0-9][0-9]*", 0);
-    if (regexMatch(hostname, 1)) {
+    if (regexMatch(hostname, 0, 1)) {
 	sscanf(hostname, "%d.%d.%d.%d", &a1, &a2, &a3, &a4);
 	adr = htonl((a1 << 24) | (a2 << 16) | (a3 << 8) | a4);
 	bcopy((void *) &adr, (void *) &hostaddr.sin_addr, sizeof(long));
 	hostaddr.sin_family = AF_INET;
 	hostaddr.sin_port = s_port;
+	if (fmInitialized) {
+	    message(Sprintf("Connecting to %s\n", hostname)->ptr, 0, 0);
+	    refresh();
+	}
 	if (connect(sock, (struct sockaddr *) &hostaddr,
 		    sizeof(struct sockaddr_in)) < 0) {
 #ifdef SOCK_DEBUG
@@ -470,8 +556,10 @@ void
 parseURL(char *url, ParsedURL * p_url, ParsedURL * current)
 {
     char *p, *q;
-    char tmp[LINELEN];
+    Str tmp;
+#ifdef __EMX__
     int i;
+#endif
 
     p = url;
     p_url->scheme = SCM_MISSING;
@@ -489,27 +577,29 @@ parseURL(char *url, ParsedURL * p_url, ParsedURL * current)
 	goto do_label;
     }
 #ifdef __EMX__
-    if (isalpha(*url) && url[1] == ':') {	/* DRIVE:foo */
-	p_url->scheme = SCM_LOCAL;
-	p_url->file = copyPath(url, -1, COPYPATH_SPC_ALLOW);
-	goto analyze_file;
+    for (i = 17; i > 0; i -= 10)
+	if (!strncmp(url, "file://localhost/", i)) {
+	    p_url->scheme = SCM_LOCAL;
+	    p += i;
+	    url += i;
+	    break;
+	}
+    if (*url == '/' && IS_ALPHA(url[1]) && (url[2] == ':' || url[2] == '|')) {
+	++p;
+	++url;
     }
+    if (IS_ALPHA(*url) && (url[1] == ':' || url[1] == '|')) {
+	p_url->scheme = SCM_LOCAL;
+	p_url->file = p = allocStr(url, 0);
+	if (p[1] == '|')
+	    p[1] = ':';
+    }
+    if (p_url->scheme == SCM_LOCAL)
+	goto analyze_file;
 #endif
     /* search for scheme */
-    while (*p && (isalpha(*p) || *p == '.' || *p == '+' || *p == '-'))
-	p++;
-    if (*p == ':') {		/* scheme found */
-	p_url->scheme = SCM_UNKNOWN;
-	for (i = 0; (q = schemetable[i].cmdname) != NULL; i++) {
-	    int len = strlen(q);
-	    if (!strncasecmp(q, url, len) && url[len] == ':') {
-		p_url->scheme = schemetable[i].cmd;
-		break;
-	    }
-	}
-	p++;
-    }
-    else {			/* scheme not found */
+    p_url->scheme = getURLScheme(&p);
+    if (p_url->scheme == SCM_MISSING) {		/* scheme not found */
 	if (current)
 	    copyParsedURL(p_url, current);
 	else
@@ -544,8 +634,27 @@ parseURL(char *url, ParsedURL * p_url, ParsedURL * current)
     p += 2;			/* scheme://foo         */
   analyze_url:
     q = p;
-    while (*p && *p != ':' && *p != '/' && *p != '@')
+    while (*p && *p != ':' && *p != '/' && *p != '@') {
+#ifdef INET6
+	if (*p == '[') {        /* rfc2732 compliance */
+	    char *p_colon = NULL;
+	    do {
+		p++;
+		if ((p_colon == NULL) && (*p == ':'))
+		    p_colon = p;
+	    } while (*p && (IS_ALNUM(*p) || *p == ':' || *p == '.'));
+	    if (*p == ']') {
+		p++;
+		break;
+	    }
+	    else if (p_colon) {
+		p = p_colon;
+		break;
+	    }
+	}
+#endif
 	p++;
+    }
     switch (*p) {
     case '\0':			/* scheme://host                */
 	/* scheme://user@host           */
@@ -565,9 +674,8 @@ parseURL(char *url, ParsedURL * p_url, ParsedURL * current)
 	    p_url->host = NULL;
 	    goto analyze_url;
 	}
-	strncpy(tmp, q, p - q);
-	tmp[p - q] = '\0';
-	p_url->port = atoi(tmp);
+	tmp = Strnew_charp_n(q, p - q);
+	p_url->port = atoi(tmp->ptr);
 	if (*p == '\0') {	/* scheme://host:port           */
 	    /* scheme://user@host:port      */
 	    /* scheme://user:pass@host:port */
@@ -587,6 +695,14 @@ parseURL(char *url, ParsedURL * p_url, ParsedURL * current)
 	p_url->port = DefaultPort[p_url->scheme];
 	break;
     }
+#ifdef INET6
+    /* rfc2732 compliance */
+    if (p_url->host != NULL && p_url->host[0] == '[' &&
+	p_url->host[strlen(p_url->host)-1] == ']' ) {
+	p_url->host[strlen(p_url->host)-1] = '\0';
+	++(p_url->host);
+    }
+#endif
   analyze_file:
     if (p_url->scheme == SCM_LOCAL && p_url->user == NULL &&
 	p_url->host != NULL && strcmp(p_url->host, "localhost")) {
@@ -596,12 +712,14 @@ parseURL(char *url, ParsedURL * p_url, ParsedURL * current)
     }
 
     q = p;
+#ifdef USE_GOPHER
     if (p_url->scheme == SCM_GOPHER) {
 	if (*q == '/')
 	    q++;
 	if (*q && q[0] != '/' && q[1] != '/' && q[2] == '/')
 	    q++;
     }
+#endif				/* USE_GOPHER */
     if (*p == '/')
 	p++;
     if (*p == '\0') {		/* scheme://host[:port]/ */
@@ -609,18 +727,21 @@ parseURL(char *url, ParsedURL * p_url, ParsedURL * current)
 	p_url->label = NULL;
 	return;
     }
+#ifdef USE_GOPHER
     if (p_url->scheme == SCM_GOPHER && *p == 'R') {
 	p++;
-	q = tmp;
-	*q++ = *p++;
+	tmp = Strnew();
+	Strcat_char(tmp, *(p++));
 	while (*p && *p != '/')
 	    p++;
+	Strcat_charp(tmp, p);
 	while (*p)
-	    *q++ = *p++;
-	*q = '\0';
-	p_url->file = copyPath(tmp, -1, COPYPATH_SPC_IGNORE);
+	    p++;
+	p_url->file = copyPath(tmp->ptr, -1, COPYPATH_SPC_IGNORE);
     }
-    else {
+    else
+#endif				/* USE_GOPHER */
+    {
 	char *cgi = strchr(p, '?');
       again:
 	while (*p && *p != '#')
@@ -721,11 +842,16 @@ parseURL2(char *url, ParsedURL * pu, ParsedURL * current)
 	    pu->host = current->host;
 	}
 	if (pu->file) {
-	    if (pu->scheme != SCM_GOPHER &&
+	    if (
+#ifdef USE_GOPHER
+		pu->scheme != SCM_GOPHER &&
+#endif				/* USE_GOPHER */
+#ifdef USE_NNTP
 		pu->scheme != SCM_NEWS &&
+#endif				/* USE_NNTP */
 		pu->file[0] != '/'
 #ifdef __EMX__
-		&& !(pu->scheme == SCM_LOCAL && isalpha(pu->file[0]) && pu->file[1] == ':')
+		&& !(pu->scheme == SCM_LOCAL && IS_ALPHA(pu->file[0]) && pu->file[1] == ':')
 #endif
 		) {
 		p = pu->file;
@@ -742,11 +868,13 @@ parseURL2(char *url, ParsedURL * pu, ParsedURL * current)
 		    pu->file = allocStr(tmp->ptr, 0);
 		}
 	    }
+#ifdef USE_GOPHER
 	    else if (pu->scheme == SCM_GOPHER &&
 		     pu->file[0] == '/') {
 		p = pu->file;
 		pu->file = allocStr(p + 1, 0);
 	    }
+#endif				/* USE_GOPHER */
 	}
 	else if (pu->label) {
 	    /* pu has only label */
@@ -772,23 +900,39 @@ parseURL2(char *url, ParsedURL * pu, ParsedURL * current)
 	    }
 	}
 #else
-	if (pu->scheme == SCM_LOCAL && pu->file[0] != '/') {
-	    tmp = Strnew_charp(currentdir());
+	if (pu->scheme == SCM_LOCAL && pu->file[0] != '/' &&
+	    strcmp(pu->file, "-")) {
+	    tmp = Strnew_charp(CurrentDir);
 	    if (Strlastchar(tmp) != '/')
 		Strcat_char(tmp, '/');
 	    Strcat_charp(tmp, pu->file);
 	    pu->file = cleanupName(tmp->ptr);
 	}
 #endif
-	else if (pu->scheme != SCM_GOPHER && pu->scheme != SCM_NEWS &&
+	else if ((pu->scheme == SCM_HTTP
+#ifdef USE_SSL
+		|| pu->scheme == SCM_HTTPS
+#endif
+		) && pu->file[0] == '/') {
+		char *p = &pu->file[1];
+		int s = getURLScheme(&p);
+		if (s == SCM_MISSING || s == SCM_UNKNOWN)
+		    pu->file = cleanupName(pu->file);
+	} else if (
+#ifdef USE_GOPHER
+		 pu->scheme != SCM_GOPHER &&
+#endif				/* USE_GOPHER */
+#ifdef USE_NNTP
+		 pu->scheme != SCM_NEWS &&
+#endif				/* USE_NNTP */
 		 pu->file[0] == '/') {
 	    pu->file = cleanupName(pu->file);
 	}
     }
 }
 
-Str
-parsedURL2Str(ParsedURL * pu)
+static Str
+_parsedURL2Str(ParsedURL * pu, int pass)
 {
     Str tmp = Strnew();
     static char *scheme_str[] =
@@ -806,19 +950,32 @@ parsedURL2Str(ParsedURL * pu)
 	/* local label */
 	return Sprintf("#%s", pu->label);
     }
+    if (pu->scheme == SCM_LOCAL && ! strcmp(pu->file, "-")) {
+	tmp = Strnew_charp("-");
+	if (pu->label) {
+	    Strcat_char(tmp, '#');
+	    Strcat_charp(tmp, pu->label);
+	}
+	return tmp;
+    }
     tmp = Strnew_charp(scheme_str[pu->scheme]);
     Strcat_char(tmp, ':');
     if (pu->scheme == SCM_MAILTO) {
 	Strcat_charp(tmp, pu->file);
 	return tmp;
     }
-    if (pu->scheme != SCM_NEWS) {
+#ifdef USE_NNTP
+    if (pu->scheme != SCM_NEWS)
+#endif				/* USE_NNTP */
+    {
 	Strcat_charp(tmp, "//");
     }
     if (pu->user) {
 	Strcat_charp(tmp, pu->user);
-/* 
- * if (pu->pass) { Strcat_char(tmp,':'); Strcat_charp(tmp,pu->pass); } */
+	if (pass && pu->pass) {
+	    Strcat_char(tmp, ':');
+	    Strcat_charp(tmp, pu->pass);
+	}
 	Strcat_char(tmp, '@');
     }
     if (pu->host) {
@@ -828,10 +985,14 @@ parsedURL2Str(ParsedURL * pu)
 	    Strcat(tmp, Sprintf("%d", pu->port));
 	}
     }
+    if (
+#ifdef USE_NNTP
+	pu->scheme != SCM_NEWS &&
+#endif				/* USE_NNTP */
 #ifdef __EMX__
-    if (pu->file == NULL || (pu->file[0] != '/' && pu->file[1] != ':'))
+	(pu->file == NULL || (pu->file[0] != '/' && pu->file[1] != ':')))
 #else
-    if (pu->file == NULL || pu->file[0] != '/')
+	(pu->file == NULL || pu->file[0] != '/'))
 #endif
 	Strcat_char(tmp, '/');
     Strcat_charp(tmp, pu->file);
@@ -842,6 +1003,35 @@ parsedURL2Str(ParsedURL * pu)
     return tmp;
 }
 
+Str
+parsedURL2Str(ParsedURL * pu)
+{
+    return _parsedURL2Str(pu, FALSE);
+}
+
+int
+getURLScheme(char **url)
+{
+    char *p = *url, *q;
+    int i;
+    int scheme = SCM_MISSING;
+
+    while (*p && (IS_ALPHA(*p) || *p == '.' || *p == '+' || *p == '-'))
+	p++;
+    if (*p == ':') {		/* scheme found */
+	scheme = SCM_UNKNOWN;
+	for (i = 0; (q = schemetable[i].cmdname) != NULL; i++) {
+	    int len = strlen(q);
+	    if (!strncasecmp(q, *url, len) && (*url)[len] == ':') {
+		scheme = schemetable[i].cmd;
+		*url = p + 1;
+		break;
+	    }
+	}
+    }
+    return scheme;
+}
+
 static char *
 otherinfo(ParsedURL * target, ParsedURL * current, char *referer)
 {
@@ -854,6 +1044,7 @@ otherinfo(ParsedURL * target, ParsedURL * current, char *referer)
 	Strcat_charp(s, UserAgent);
     Strcat_charp(s, "\r\n");
     Strcat_charp(s, "Accept: text/*, image/*, audio/*, application/*\r\n");
+    Strcat_charp(s, "Accept-Encoding: gzip, compress, bzip, bzip2, deflate\r\n");
     Strcat_charp(s, "Accept-Language: ");
     if (AcceptLang != NULL && *AcceptLang != '\0') {
 	Strcat_charp(s, AcceptLang);
@@ -924,7 +1115,7 @@ HTTPrequest(ParsedURL * pu, ParsedURL * current, HRequest * hr, TextList * extra
 	Strcat_charp(tmp, pu->file);
     }
     else {
-	Strcat(tmp, parsedURL2Str(pu));
+	Strcat(tmp, _parsedURL2Str(pu, TRUE));
     }
     Strcat_charp(tmp, " HTTP/1.0\r\n");
     if (hr->referer == NO_REFERER)
@@ -940,6 +1131,9 @@ HTTPrequest(ParsedURL * pu, ParsedURL * current, HRequest * hr, TextList * extra
 	Strcat_charp(tmp, "Cookie: ");
 	Strcat(tmp, cookie);
 	Strcat_charp(tmp, "\r\n");
+	/* [DRAFT 12] s. 10.1 */
+	if (cookie->ptr[0] != '$')
+	    Strcat_charp(tmp, "Cookie2: $Version=\"1\"\r\n");
     }
 #endif				/* USE_COOKIE */
     if (hr->command == HR_COMMAND_POST) {
@@ -966,34 +1160,49 @@ HTTPrequest(ParsedURL * pu, ParsedURL * current, HRequest * hr, TextList * extra
     return tmp;
 }
 
+void
+init_stream(URLFile *uf, int scheme, InputStream stream)
+{
+    uf->stream = stream;
+    uf->scheme = scheme;
+    uf->encoding = ENC_7BIT;
+    uf->is_cgi = FALSE;
+    uf->compression = 0;
+    uf->guess_type = NULL;
+    uf->ext = NULL;
+}
+
+static InputStream
+openFTPStream(ParsedURL * pu)
+{
+    return newFileStream(openFTP(pu), closeFTP);
+}
+
 URLFile
 openURL(char *url, ParsedURL * pu, ParsedURL * current,
 	URLOption * option, FormList * request, TextList * extra_header,
-	URLFile * ouf, char *status)
+	URLFile * ouf, unsigned char *status)
 {
     Str tmp = Strnew();
-    FILE *f, *fw;
-    int i, sock;
-    char *p, *q, *r;
+    int sock;
+    char *p, *q;
     URLFile uf;
-    char *last_dot;
     HRequest hr;
 #ifdef USE_SSL
     SSL *sslh;
 #endif				/* USE_SSL */
+#ifdef USE_NNTP
+    FILE *fw;
+    int i;
+    char *r;
+    InputStream stream;
+#endif				/* USE_NNTP */
 
     if (ouf) {
 	uf = *ouf;
     }
     else {
-	uf.scheme = SCM_MISSING;
-	uf.stream.f = NULL;
-	uf.stream_type = SMT_FILE;
-	uf.iseof = FALSE;
-	uf.close = (void (*)()) fclose;
-	uf.encoding = ENC_7BIT;
-	uf.is_cgi = FALSE;
-	uf.compression = 0;
+	init_stream(&uf, SCM_MISSING, NULL);
     }
 
   retry:
@@ -1017,20 +1226,7 @@ openURL(char *url, ParsedURL * pu, ParsedURL * current,
 
     uf.scheme = pu->scheme;
     pu->is_nocache = (option->flag & RG_NOCACHE);
-    last_dot = "";
-    for (p = pu->file; p && *p; p++) {
-	if (*p == '.') {
-	    last_dot = p;
-	}
-	else if (*p == '?')
-	    break;
-    }
-    if (*last_dot == '.') {
-	for (i = 1; last_dot[i] && IS_ALNUM(last_dot[i]) && i < 8; i++);
-	uf.ext = allocStr(last_dot, i);
-    }
-    else
-	uf.ext = last_dot;
+    uf.ext = filename_extension(pu->file, 1);
 
     hr.command = HR_COMMAND_GET;
     hr.flag = 0;
@@ -1042,10 +1238,10 @@ openURL(char *url, ParsedURL * pu, ParsedURL * current,
     case SCM_LOCAL_CGI:
 	if (request && request->body) {
 	    /* local CGI: POST */
-	    uf.stream.f = localcgi_post(pu->file, request);
-	    if (uf.stream.f == NULL)
+	    uf.stream = newFileStream(localcgi_post(pu->file, request, option->referer),
+				      (void (*)()) pclose);
+	    if (uf.stream == NULL)
 		goto ordinary_local_file;
-	    uf.close = (void (*)()) pclose;
 	    uf.is_cgi = TRUE;
 	    uf.scheme = SCM_LOCAL_CGI;
 	}
@@ -1053,12 +1249,12 @@ openURL(char *url, ParsedURL * pu, ParsedURL * current,
 	    /* lodal CGI: GET */
 	    for (q = pu->file; *q && *q != '?'; q++)
 		Strcat_char(tmp, *q);
-	    uf.stream.f = localcgi_get(tmp->ptr, p + 1);
-	    if (uf.stream.f == NULL) {
+	    uf.stream = newFileStream(localcgi_get(tmp->ptr, p + 1, option->referer),
+					(void (*)()) pclose);
+	    if (uf.stream == NULL) {
 		pu->file = tmp->ptr;
 		goto ordinary_local_file;
 	    }
-	    uf.close = (void (*)()) pclose;
 	    uf.is_cgi = TRUE;
 	    uf.scheme = SCM_LOCAL_CGI;
 	}
@@ -1069,21 +1265,21 @@ openURL(char *url, ParsedURL * pu, ParsedURL * current,
 #endif
 	{
 	    /* lodal CGI: GET */
-	    uf.stream.f = localcgi_get(pu->file, "");
-	    if (uf.stream.f == NULL)
+	    uf.stream = newFileStream(localcgi_get(pu->file, "", option->referer),
+				      (void (*)()) pclose);
+	    if (uf.stream == NULL)
 		goto ordinary_local_file;
-	    uf.close = (void (*)()) pclose;
 	    uf.is_cgi = TRUE;
 	    uf.scheme = SCM_LOCAL_CGI;
 	}
 	else {
 	  ordinary_local_file:
-	    uf.stream.f = examineFile(pu->file, &uf.close);
+	    examineFile(pu->file, &uf);
 	}
-	if (uf.stream.f == NULL) {
+	if (uf.stream == NULL) {
 	    if (dir_exist(pu->file)) {
-		uf.stream.f = add_index_file(pu, &uf.close);
-		if (uf.stream.f == NULL)
+		add_index_file(pu, &uf);
+		if (uf.stream == NULL)
 		    return uf;
 	    }
 	    else if (document_root != NULL) {
@@ -1094,23 +1290,22 @@ openURL(char *url, ParsedURL * pu, ParsedURL * current,
 		tmp = Strnew_charp(cleanupName(tmp->ptr));
 		if (dir_exist(tmp->ptr)) {
 		    pu->file = tmp->ptr;
-		    uf.stream.f = add_index_file(pu, &uf.close);
-		    if (uf.stream.f == NULL) {
+		    add_index_file(pu, &uf);
+		    if (uf.stream == NULL) {
 			return uf;
 		    }
 		}
 		else {
-		    uf.stream.f = examineFile(tmp->ptr, &uf.close);
-		    if (uf.stream.f)
+		    examineFile(tmp->ptr, &uf);
+		    if (uf.stream)
 			pu->file = tmp->ptr;
 		}
 	    }
 	}
-	if (uf.stream.f == NULL) {
-	    if (retryAsHttp &&
-		url[0] != '/' &&
-		strncasecmp(url, "http://", 7) != 0 &&
-		strncasecmp(url, "file://", 7) != 0) {
+	if (uf.stream == NULL && retryAsHttp && url[0] != '/') {
+	    char *u = url;
+	    int scheme = getURLScheme(&u);
+	    if (scheme == SCM_MISSING || scheme == SCM_UNKNOWN) {
 		/* retry it as "http://" */
 		url = Strnew_m_charp("http://", url, NULL)->ptr;
 		goto retry;
@@ -1122,6 +1317,7 @@ openURL(char *url, ParsedURL * pu, ParsedURL * current,
 	    pu->file = allocStr("/", 0);
 	if (non_null(FTP_proxy) &&
 	    !Do_not_use_proxy &&
+	    pu->host != NULL &&
 	    !check_no_proxy(pu->host)) {
 	    sock = openSocket(FTP_proxy_parsed.host,
 			    schemetable[FTP_proxy_parsed.scheme].cmdname,
@@ -1133,9 +1329,8 @@ openURL(char *url, ParsedURL * pu, ParsedURL * current,
 	    write(sock, tmp->ptr, tmp->length);
 	}
 	else {
-	    uf.stream.f = openFTP(pu);
+	    uf.stream = openFTPStream(pu);
 	    uf.scheme = pu->scheme;
-	    uf.close = (void (*)()) closeFTP;
 	    return uf;
 	}
 	break;
@@ -1151,11 +1346,12 @@ openURL(char *url, ParsedURL * pu, ParsedURL * current,
 	    hr.command = HR_COMMAND_HEAD;
 	if (non_null(HTTP_proxy) &&
 	    !Do_not_use_proxy &&
+	    pu->host != NULL &&
 	    !check_no_proxy(pu->host)) {
 	    char *save_label;
 #ifdef USE_SSL
 	    if (pu->scheme == SCM_HTTPS && *status == HTST_CONNECT) {
-		sock = ouf->stream.ss->sock;
+		sock = ssl_socket_of(ouf->stream);
 		if (!(sslh = openSSLHandle(sock))) {
 		    *status = HTST_MISSING;
 		    return uf;
@@ -1223,8 +1419,7 @@ openURL(char *url, ParsedURL * pu, ParsedURL * current,
 	}
 #ifdef USE_SSL
 	if (pu->scheme == SCM_HTTPS) {
-	    uf.stream_type = SMT_SSL;
-	    uf.stream.ss = newSSLStream(sslh, sock);
+	    uf.stream = newSSLStream(sslh, sock);
 	    if (sslh)
 		SSL_write(sslh, tmp->ptr, tmp->length);
 	    else
@@ -1254,9 +1449,11 @@ openURL(char *url, ParsedURL * pu, ParsedURL * current,
 		write_from_file(sock, request->body);
 	}
 	break;
+#ifdef USE_GOPHER
     case SCM_GOPHER:
 	if (non_null(GOPHER_proxy) &&
 	    !Do_not_use_proxy &&
+	    pu->host != NULL &&
 	    !check_no_proxy(pu->host)) {
 	    sock = openSocket(GOPHER_proxy_parsed.host,
 			 schemetable[GOPHER_proxy_parsed.scheme].cmdname,
@@ -1279,19 +1476,27 @@ openURL(char *url, ParsedURL * pu, ParsedURL * current,
 	}
 	write(sock, tmp->ptr, tmp->length);
 	break;
+#endif				/* USE_GOPHER */
+#ifdef USE_NNTP
+    case SCM_NNTP:
+	/* nntp://<host>:<port>/<newsgroup-name>/<article-number> */
     case SCM_NEWS:
-	p = getenv("NNTPSERVER");
+	if (pu->scheme == SCM_NNTP) {
+	    p = pu->host;
+	} else {
+	    p = getenv("NNTPSERVER");
+	}
 	r = getenv("NNTPMODE");
 	if (p == NULL)
 	    return uf;
 	sock = openSocket(p, "nntp", pu->port);
 	if (sock < 0)
 	    return uf;
-	f = fdopen(sock, "rb");
+	stream = newInputStream(sock);
 	fw = fdopen(sock, "wb");
-	if (f == NULL || fw == NULL)
+	if (stream == NULL || fw == NULL)
 	    return uf;
-	tmp = Strfgets(f);
+	tmp = StrISgets(stream);
 	if (tmp->length == 0)
 	    goto nntp_error;
 	sscanf(tmp->ptr, "%d", &i);
@@ -1300,53 +1505,76 @@ openURL(char *url, ParsedURL * pu, ParsedURL * current,
 	if (r && *r != '\0') {
 	    fprintf(fw, "MODE %s\r\n", r);
 	    fflush(fw);
-	    tmp = Strfgets(f);
+	    tmp = StrISgets(stream);
 	    if (tmp->length == 0)
 		goto nntp_error;
 	    sscanf(tmp->ptr, "%d", &i);
 	    if (i != 200 && i != 201)
 		goto nntp_error;
 	}
-	fprintf(fw, "ARTICLE <%s>\r\n", pu->file);
+	if (pu->scheme == SCM_NNTP) {
+	    char *group;
+	    if (pu->file == NULL || *pu->file == '\0')
+		goto nntp_error;
+	    /* first char of pu->file is '/' */
+	    group = Strnew_charp(pu->file + 1)->ptr;
+	    p = strchr(group, '/');
+	    if (p == NULL)
+		goto nntp_error;
+	    *p++ = '\0';
+	    fprintf(fw, "GROUP %s\r\n", group);
+	    fflush(fw);
+	    tmp = StrISgets(stream);
+	    if (tmp->length == 0) {
+		goto nntp_error;
+	    }
+	    sscanf(tmp->ptr, "%d", &i);
+	    if (i != 211)
+		goto nntp_error;
+	    fprintf(fw, "ARTICLE %s\r\n", p);
+	} else {
+	    fprintf(fw, "ARTICLE <%s>\r\n", pu->file);
+	}
 	fflush(fw);
-	tmp = Strfgets(f);
+	tmp = StrISgets(stream);
 	if (tmp->length == 0)
 	    goto nntp_error;
 	sscanf(tmp->ptr, "%d", &i);
 	if (i != 220)
 	    goto nntp_error;
-	uf.stream.f = f;
-	uf.close = (void (*)()) fclose;
+	uf.scheme = SCM_NEWS; /* XXX */
+	uf.stream = stream;
 	return uf;
       nntp_error:
-	fclose(f);
+	ISclose(stream);
 	fclose(fw);
 	return uf;
+#endif				/* USE_NNTP */
     case SCM_UNKNOWN:
+    default:
 	return uf;
     }
-    uf.stream.f = fdopen(sock, "rb");
-    uf.stream_type = SMT_FILE;
-    uf.close = (void (*)()) fclose;
+    uf.stream = newInputStream(sock);
     return uf;
 }
 
 /* add index_file if exists */
-static FILE *
-add_index_file(ParsedURL * pu, void (**close_rout) ())
+static void
+add_index_file(ParsedURL * pu, URLFile *uf)
 {
     Str tmp = Strnew_charp(pu->file);
-    FILE *f;
 
-    if (index_file == NULL || index_file[0] == '\0')
-	return NULL;
+    if (index_file == NULL || index_file[0] == '\0') {
+	uf->stream = NULL;
+	return;
+    }
     Strcat_m_charp(tmp, "/", index_file, NULL);
     tmp->ptr = cleanupName(tmp->ptr);
-    f = examineFile(tmp->ptr, close_rout);
-    if (f == NULL)
-	return NULL;
+    examineFile(tmp->ptr, uf);
+    if (uf->stream == NULL)
+	return;
     pu->file = tmp->ptr;
-    return f;
+    return;
 }
 
 char *
@@ -1373,24 +1601,30 @@ char *
 guessContentType(char *filename)
 {
     char *ret;
+    int i;
 
     if (filename == NULL)
 	return NULL;
-    if ((ret = guessContentTypeFromTable(UserMimeTypes, filename)) != NULL)
-	return ret;
-    if ((ret = guessContentTypeFromTable(SysMimeTypes, filename)) != NULL)
-	return ret;
+    if (mimetypes_list == NULL)
+	goto no_user_mimetypes;
+
+    for (i = 0; i < mimetypes_list->nitem; i++) {
+	if ((ret = guessContentTypeFromTable(UserMimeTypes[i], filename)) != NULL)
+	    return ret;
+    }
+
+ no_user_mimetypes:
     return guessContentTypeFromTable(DefaultGuess, filename);
 }
 
-void
-set_no_proxy(char *no_proxy_list)
+TextList *
+make_domain_list(char *domain_list)
 {
     char *p;
     Str tmp;
+    TextList *domains = NULL;
 
-    NO_proxy_domains = newTextList();
-    p = no_proxy_list;
+    p = domain_list;
     tmp = Strnew_size(64);
     while (*p) {
 	while (*p && IS_SPACE(*p))
@@ -1398,18 +1632,24 @@ set_no_proxy(char *no_proxy_list)
 	Strclear(tmp);
 	while (*p && !IS_SPACE(*p) && *p != ',')
 	    Strcat_char(tmp, *p++);
-	if (tmp->length > 0)
-	    pushText(NO_proxy_domains, tmp->ptr);
+	if (tmp->length > 0) {
+	    if (domains == NULL)
+		domains = newTextList();
+	    pushText(domains, tmp->ptr);
+	}
 	while (*p && IS_SPACE(*p))
 	    p++;
 	if (*p == ',')
 	    p++;
     }
+    return domains;
 }
 
 static int
 domain_match(char *pat, char *domain)
 {
+    if (domain == NULL)
+	return 0;
     if (*pat == '.')
 	pat++;
     for (;;) {
@@ -1427,7 +1667,8 @@ check_no_proxy(char *domain)
 {
     TextListItem *tl;
 
-    if (NO_proxy_domains == NULL || NO_proxy_domains->nitem == 0)
+    if (NO_proxy_domains == NULL || NO_proxy_domains->nitem == 0 ||
+	domain == NULL)
 	return 0;
     for (tl = NO_proxy_domains->first; tl != NULL; tl = tl->next) {
 	if (domain_match(tl->ptr, domain))
@@ -1513,3 +1754,31 @@ check_no_proxy(char *domain)
 #endif				/* NOPROXY_NETADDR */
     return 0;
 }
+
+char *
+filename_extension(char *path, int is_url)
+{
+    char *last_dot = "", *p = path;
+    int i;
+
+    if (path == NULL)
+	return last_dot;
+    if (*p == '.')
+	p++;
+    for (; *p; p++) {
+        if (*p == '.') {
+            last_dot = p;
+        }
+        else if (is_url && *p == '?')
+            break;
+    }
+    if (*last_dot == '.') {
+        for (i = 1; last_dot[i] && i < 8; i++) {
+	    if (is_url && !IS_ALNUM(last_dot[i]))
+		break;
+	}
+        return allocStr(last_dot, i);
+    }
+    else
+        return last_dot;
+}
